/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

#include "server.h"

/* thread states */
#define THREAD_STOPPED 0
#define THREAD_RUNNING 1
#define THREAD_END     -1
#define THREAD_DEAD    -2

/* how many threads to try and start */
#define THREAD_COUNT   16

/* globals */

u32 __shutdown_flag;   /* interupt flag */
v3_LOG __log_work;     /* for work results */
v3_LOG __log_activity; /* for network activity, firewall triggers, etc */
v3_NET_ACL __net_acl;  /* for evil hosts */

/* thread data */
v3_MUTEX __thread_mutex[THREAD_COUNT]; /* locks for threads */
v3_NET __thread_net[THREAD_COUNT];    /* network handles for threads */
s32 __thread_state[THREAD_COUNT];      /* thread states */
u32 __thread_num[THREAD_COUNT];        /* tell the threads who they are */
u64 __thread_id[THREAD_COUNT];         /* thread id's */

/* mutexes for assign and accept of work */
v3_MUTEX __mutex_assign;
v3_MUTEX __mutex_accept;

/* work tracking data */
u32 __count;
u32 __count_end;
u64 __total;

void CatchSignal( int arg )
{
  __shutdown_flag = 1;
  v3SignalRegister( V3_PROCESS_SIGINT, CatchSignal );
  v3SignalRegister( V3_PROCESS_SIGTERM, CatchSignal );

  /*
    do shutdown immediately (this may cause problems) but since we're in
    a ...Accept(...WAIT). We have no choice till we finish the
    multi-threaded example, or two-phase commit the work.
  */
  Shutdown();
  v3ProcessEnd( 0 );
}

void Thread( void * arg )
{
  u32 me;
  u32 type;

  /* do thread setup */
  me = *( (u32 *) arg );
  v3ThreadPriority( 192 ); /* a little less then control thread */

  for ( ; ; )
  {
    /* wait for us to be told to run */
    v3MutexLock( &__thread_mutex[me], V3_MUTEX_WAIT );
    v3MutexUnlock( &__thread_mutex[me] );

    /* check if we should exit now */
    if ( __thread_state[me] == THREAD_END )
    {
      v3PrintA( (ascii *) "Thread %u ending\n", me );
      v3ThreadEnd();
    }

    /* do the work we need to do */
    v3PrintA( (ascii *) "Hello from thread %u\n", me );

    while ( v3NetRecv( &type, (u32) sizeof( type ), &__thread_net[me], 5000 )
      == sizeof( type ) )
    {
      /* send off to subroutines */
      v3u32Load( &type, (u8 *) &type );
      switch ( type )
      {
        case PACKET_TYPE_REQUEST:
          Assign( &__thread_net[me] );
          break;
        case PACKET_TYPE_RESULTS:
          Accept( &__thread_net[me] );
          break;
        default:
          /* error, close connection*/
          v3NetClose( &__thread_net[me] );
          break;
      }
    }
    v3NetClose( &__thread_net[me] );

    /* go back to sleep */
    v3MutexLock( &__thread_mutex[me], V3_MUTEX_WAIT );
    __thread_state[me] = THREAD_STOPPED;
  }
}

s32 Initialize( v3_NET * net )
{
  u32 ip;
  u32 port;

  /* load unassigned work data */

  /* load assigned/finished work data */

  /* setup network listener */
  if ( v3NetDNS( &ip, 1, NETPACKET_SERVER ) != 1 )
  {
    v3PrintA( (ascii *) "Hostname lookup error.\n" );
    return( V3_FAIL );
  }

  if ( v3u32A( &port, NULL, NETPACKET_PORT, 10 ) != V3_PASS )
  {
    v3PrintA( (ascii *) "Port error.\n" );
    return( V3_FAIL );
  }

  if ( v3NetListen( net, ip, (u16) port, V3_NET_MODE_TCP, 16 )
    != V3_PASS )
  {
    v3PrintA( (ascii *) "Network already in use.\n" );
    return( V3_FAIL );
  }

  v3Log( &__log_activity, 0, V3_LOG_ECHO,
    (ascii *) "Listening on address %u.%u.%u.%u port %u\n",
    ( net->my_host >> 24 ) & 0xFF, ( net->my_host >> 16 ) & 0xFF,
    ( net->my_host >> 8 ) & 0xFF, net->my_host & 0xFF, net->my_port );

  return( V3_PASS );
}

s32 Shutdown( void )
{
  /* close down network */

  /* save unassigned work data */

  /* save assigned/finished work data */

  return( V3_PASS );
}

void Assign( v3_NET * net )
{
  PACKET_REQUEST request;
  PACKET_ASSIGNMENT assign;
  u32 tmp_count;
  u32 length;
  u32 start;

  /* read the rest of the request, assign the work from the unassigned work */
  length = (u32) ( sizeof( PACKET_REQUEST ) - sizeof( request.type ) );
  /* make sure to wait long enough, about 1sec/KB + 5sec */
  if ( v3NetRecv( &request.speed, length, net, length + 5000 ) == length )
  {
    v3u64Load( &request.speed, (u8 *) &request.speed );

    assign.type = PACKET_TYPE_ASSIGNMENT;
    v3u32Save( (u8 *) &assign.type, &assign.type );

    /* work with the global variables */
    v3MutexLock( &__mutex_assign, V3_MUTEX_WAIT );
    v3u32Save( (u8 *) &assign.start, &__count );
    start = __count;
    /* already done? */
    if ( start >= __count_end )
    {
      tmp_count = 0;
    }
    else
    {
      tmp_count = __count + v3u32u64( request.speed );
      if ( tmp_count >= __count_end )
      {
        __count = __count_end;
        tmp_count = __count_end - 1;
      }
      else
      {
        __count = tmp_count;
        tmp_count--;
      }
    }
    v3MutexUnlock( &__mutex_assign );

    if ( tmp_count != 0 )
    {
      v3PrintA( (ascii *) "Assigned work %08X-%08X\n", start, tmp_count );
    }

    v3u32Save( (u8 *) &assign.end, &tmp_count );
    if ( v3NetSend( net, &assign, (u32) sizeof( PACKET_ASSIGNMENT ) )
      != V3_PASS )
    {
      v3PrintA( (ascii *) "Network Send Failed\n" );
      return;
    }

  }

  /* mark what we've handed out as assigned work */
   

}

void Accept( v3_NET * net )
{
  PACKET_RESULTS results;
  PACKET_ACCEPT work_accept;
  const ascii * os_types[] = V3_OS_TYPES;
  const ascii * cpu_types[] = V3_CPU_TYPES;
  u32 length;
  u64 tmp;

  /* accept and record results as finished */
  length = (u32) ( sizeof( PACKET_RESULTS ) - sizeof( results.type ) );
  /* make sure to wait long enough, about 1sec/KB + 5sec */
  if ( v3NetRecv( &results.start, length, net, length +5000 ) == length )
  {
    v3u32Load( &results.start, (u8 *) &results.start );
    v3u32Load( &results.end, (u8 *) &results.end );
    v3u64Load( &results.total, (u8 *) &results.total );
    v3u32Load( &results.cpu, (u8 *) &results.cpu );
    v3u32Load( &results.os, (u8 *) &results.os );
    /* ... */
    v3PrintA( (ascii *)
      "Got one - Sum %016Y - OS = %.20s, CPU = %.20s [%.63s]\n",
      results.total, os_types[results.os], cpu_types[results.cpu],
      results.email );

    /* work with the global variables */
    v3MutexLock( &__mutex_accept, V3_MUTEX_WAIT );
    tmp = __total = v3u64Add( __total, results.total );
    v3MutexUnlock( &__mutex_accept );

    v3PrintA( (ascii *) "New total: %016Y\n", tmp );

    work_accept.type = PACKET_TYPE_ACCEPT;
    v3u32Save( (u8 *) &work_accept.type, &work_accept.type );
    work_accept.result = V3_PASS;
    v3u32Save( (u8 *) &work_accept.result, &work_accept.result );


    if ( v3NetSend( net, &work_accept, (u32) sizeof( PACKET_ACCEPT ) )
      != V3_PASS )
    {
      v3PrintA( (ascii *) "Network Send Failed\n" );
      return;
    }

    /* log work done */
  }

}

int main( int argc, char * argv[] )
{
  s32 result;
  v3_NET net, new_net;
  u32 type;
  v3_NET_HOSTNAME name;
  u32 i;
  u32 threads_started;
  u32 thread_seek, thread_tries, found;
  s32 err, err2;

  /* run system tests */
  v3PrintA( (ascii *) "Running system tests... " );
  if ( v3Test( &err, &err2, 0 ) != V3_PASS )
  {
    v3PrintA( (ascii *) "Error: O=%u, C=%u, M=%i, E=%i\n",
      OS_TYPE, CPU_TYPE, err, err2 );
    v3ProcessEnd( -1 );
  }
  v3PrintA( (ascii *) "all passed.\n" );

  /* initialize the globals */
  v3MemSet( &__log_work, v3u64u32( sizeof( v3_LOG ) ), 0 );
  v3MemSet( &__log_activity, v3u64u32( sizeof( v3_LOG ) ), 0 );
  v3MemSet( &__net_acl, v3u64u32( sizeof( v3_NET_ACL ) ), 0 );
  __count = 0;
  __count_end = 0x1000000;
  _V3_SET64( __total, 00000000, 00000000 )
  v3MemSet( &__mutex_assign, v3u64u32( sizeof( v3_MUTEX ) ), 0 );
  v3MemSet( &__mutex_accept, v3u64u32( sizeof( v3_MUTEX ) ), 0 );
  for ( i = 0 ; i < THREAD_COUNT ; i ++ )
  {
    v3MemSet( &__thread_mutex[i], v3u64u32( sizeof( v3_MUTEX ) ), 0 );
    _V3_SET64( __thread_id[i], 00000000, 00000000 );
    __thread_state[i] = THREAD_STOPPED;
    __thread_num[i] = i;
    v3MutexLock( &__thread_mutex[i], V3_MUTEX_WAIT );
  }

  /* register signals, after this we have to check the flag for shutdown */
  __shutdown_flag = 0;
  if ( ( v3SignalRegister( V3_PROCESS_SIGINT, CatchSignal ) != V3_PASS ) ||
    ( v3SignalRegister( V3_PROCESS_SIGTERM, CatchSignal ) != V3_PASS ) )
  {
    v3PrintA( (ascii *) "Signal registration failed\n" );
    v3ProcessEnd( -1 );
  }

  /* open logs */
  if ( v3LogOpen( &__log_work, WORK_LOG, 100, V3_LOG_MODE_NUMBER )
    != V3_PASS )
  {
    v3PrintA( (ascii *) "Unable to open work log.\n" );
    v3ProcessEnd( -1 );
  }

  if ( v3LogOpen( &__log_activity, ACTIVITY_LOG, 10, V3_LOG_MODE_NUMBER )
    != V3_PASS )
  {
    v3PrintA( (ascii *) "Unable to open activity log.\n" );
    v3ProcessEnd( -1 );
  }

  /* start the threads */
  threads_started = 0;
  for ( i = 0 ; i < THREAD_COUNT ; i ++ )
  {
    /* note we cant just pass &i as a thread number */
    /* Make sure that the stack is large enough for your application! */
    __thread_id[i] = v3ThreadBegin( Thread, &__thread_num[i], 16 * 1024 );

    /* Test for bad thread creation */
    if ( _V3_EQ64( __thread_id[i], FFFFFFFF, FFFFFFFF ) )
    {
      v3PrintA( (ascii *) "%u was one thread too many\n", i + 1 );
      continue;
    }
    else
    {
      v3PrintA( (ascii *) "Thread %Y started\n", __thread_id[i] );
      threads_started++;
    }
  }

  /* initialize network */
  v3MemSet( &net, v3u64u32( sizeof( v3_NET ) ), 0 );
  v3MemSet( &new_net, v3u64u32( sizeof( v3_NET ) ), 0 );
  if ( Initialize( &net ) != V3_PASS )
  {
    v3PrintA( (ascii *) "Server Initialization failed\n" );
    v3ProcessEnd( -1 );
  }

  if ( threads_started > 0 )
  {
    /* multi-thread version */
    thread_seek = 0;
    for ( ; ; )
    {
      found = 0;
      while ( found == 0 )
      {
        thread_tries = 0;
        thread_seek = ( thread_seek + 1 ) % threads_started;
        if ( __thread_state[thread_seek] == THREAD_STOPPED )
        {
          /* found an available thread */
          if ( v3NetAccept( &__thread_net[thread_seek], &net, &__net_acl,
            V3_NET_ACCEPT_WAIT ) == V3_PASS )
          {
            v3PrintA( (ascii *) "Connection from %u.%u.%u.%u\n",
              ( new_net.host >> 24 ) & 0xFF, ( new_net.host >> 16 ) & 0xFF,
              ( new_net.host >> 8 ) & 0xFF, new_net.host & 0xFF );
            /* unleash thread */
            __thread_state[thread_seek] = THREAD_RUNNING;
            v3MutexUnlock( &__thread_mutex[thread_seek] );
            thread_tries = 0;
            found = 1;
          }
        }
        else
        {
          /* try the next one, but do not busy-wait */
          if ( ++thread_tries > threads_started )
          {
            thread_tries = 0;
            v3Sleep( 10 );
          }
        }
      }

    }
  }
  else
  {
    /* no threads */
    while ( ( result = v3NetAccept( &new_net, &net, &__net_acl,
      V3_NET_ACCEPT_WAIT ) ) == V3_PASS )
    {
      v3NetRevDNS( &name, new_net.host );
      v3PrintA( (ascii *) "Connection from %u.%u.%u.%u\n",
        ( new_net.host >> 24 ) & 0xFF, ( new_net.host >> 16 ) & 0xFF,
        ( new_net.host >> 8 ) & 0xFF, new_net.host & 0xFF );

      while ( v3NetRecv( &type, (u32) sizeof( type ), &new_net, 5000 )
        == sizeof( type ) )
      {
        /* send off to subroutines */
        v3u32Load( &type, (u8 *) &type );
        switch ( type )
        {
          case PACKET_TYPE_REQUEST:
            Assign( &new_net );
            break;
          case PACKET_TYPE_RESULTS:
            Accept( &new_net );
            break;
          default:
            /* error, close connection*/
            v3NetClose( &new_net );
            break;
        }
      }
      v3NetClose( &new_net );
    }
  }

  return( V3_PASS );
}
