/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

#include "buffer.h"
#include "cosmmath.h"
#include "cosmmem.h"

s32 v3BufferCreate( v3_BUFFER * buffer, u64 size, u32 mode, u64 grow,
  const void * const data, u64 length )
{
  s32 result;

  if ( buffer == NULL )
  {
    return( V3_BUFFER_ERROR_MEMORY ); /* Or what else to return? */
  }

  v3MemSet( buffer, v3u64u32( sizeof( v3_BUFFER ) ), 0 );

  v3MutexLock( &buffer->lock, V3_MUTEX_WAIT );

  if ( ( buffer->memory = v3MemAlloc( size, V3_MEM_NORMAL ) ) == NULL )
  {
    v3MutexUnlock( &buffer->lock );
    return( V3_BUFFER_ERROR_MEMORY );
  }

  if ( ( mode != V3_BUFFER_MODE_QUEUE ) && ( mode != V3_BUFFER_MODE_STACK ) )
  {
    v3MutexUnlock( &buffer->lock );
    return( V3_BUFFER_ERROR_MODE );
  }

  /* Simply set the contents of the struct acording to arguments. */
  buffer->mem_length = size;
  buffer->mode = mode;
  buffer->head = v3u64u32( 0 );
  buffer->tail = v3u64u32( 0 );
  buffer->data_length = v3u64u32( 0 );
  buffer->grow_size = grow;

  if ( data != NULL )
  {
    if ( ( result = v3_BufferPut( buffer, data, length ) ) != V3_PASS )
    {
      v3MutexUnlock( &buffer->lock );
      return( result ) ;
    }
  }
  /* What action should I take on buffer->lock ?? */

  v3MutexUnlock( &buffer->lock );

  return( V3_PASS ) ;
}

void v3BufferFree( v3_BUFFER * buffer )
{
  if ( buffer == NULL )
  {
    return;
  }

  v3MutexLock( &buffer->lock, V3_MUTEX_WAIT );

  v3MemFree( buffer->memory );

  buffer->memory = NULL;
  buffer->data_length = v3u64u32( 0 );
  buffer->mem_length  = v3u64u32( 0 );
  buffer->grow_size   = v3u64u32( 0 );

  v3MutexUnlock( &buffer->lock );
  v3MutexFree( &buffer->lock );
}

u64 v3BufferLength( v3_BUFFER * buffer )
{
  u64 length;

  if ( buffer == NULL )
  {
    return( v3u64u32( 0 ) );
  }
  v3MutexLock( &buffer->lock, V3_MUTEX_WAIT );
  length = buffer->data_length;
  v3MutexUnlock( &buffer->lock );

  return( length );
}

s32 v3BufferPut( v3_BUFFER * buffer, const void * data, u64 length )
{
  s32 result;

  if ( buffer == NULL )
  {
    return( V3_BUFFER_ERROR_MEMORY );
  }

  v3MutexLock( &buffer->lock, V3_MUTEX_WAIT );
  result = v3_BufferPut( buffer, data, length );
  v3MutexUnlock( &buffer->lock );

  return( result );
}

u64 v3BufferGet( void * data, u64 length, v3_BUFFER * buffer )
{
  u64 amount;
  u64 bot, top, ofs;
  u8 * memory, * data_u8;

  if ( buffer == NULL )
  {
    return( v3u64u32( 0 ) );
  }

  v3MutexLock( &buffer->lock, V3_MUTEX_WAIT );

  memory = (u8 *) buffer->memory;
  data_u8 = (u8 *) data;


  /*
    Choose how much data to grab. The smaller of wanted and available.
    This ensures we can always get the amount stored in amount.
  */
  if ( v3u64Gt( length, buffer->data_length ) )
  {
    amount = buffer->data_length;
  }
  else
  {
    amount = length;
  }

  switch ( buffer->mode )
  {
    case V3_BUFFER_MODE_QUEUE:
    {
      /* This is a FIFO, which means we should take from the beginning
         of the buffer. */

      if ( v3u64Gt ( amount, v3u64Sub( buffer->mem_length, buffer->tail ) ) )
      {
        /* We will have to wrap.*/
        top = v3u64Sub( buffer->mem_length, buffer->tail );
        bot = v3u64Sub( amount, top ); /* The rest. */
      }
      else
      {
        /* We wont have to wrap. */
        top = amount;
        bot = v3u64u32( 0 );
      }

      /* Start with what we can grab from the end.*/
      if ( v3u64Gt ( top, v3u64u32 ( 0 ) ) )
      {
        if ( v3MemCopy( data, v3MemOffset( memory, buffer->tail ), top )
          == V3_FAIL )
        {
          v3MutexUnlock( &buffer->lock );
          return( v3u64u32 ( 0 ) );
        }
        /* Move us forward. */
        buffer->tail = v3u64Add( buffer->tail, top );
      }

      /* Now start from the beginning. Now top is our offset into the
         target data, and the offset in the source is zero. */
      if ( v3u64Gt( bot, v3u64u32( 0 ) ) )
      {
        if ( v3MemCopy( v3MemOffset( data_u8, top ),
          buffer->memory, bot ) == V3_FAIL )
        {
          v3MutexUnlock( &buffer->lock );
          return( v3u64u32 ( 0 ) );
        }
        /* We are wrapping our tail. */
        buffer->tail = v3u64Add( buffer->tail, top );
      }

      buffer->data_length = v3u64Sub( buffer->data_length, amount );
      v3MutexUnlock( &buffer->lock );

      return( amount );
      /* MODE_QUEUE */
    }
    case V3_BUFFER_MODE_STACK:
    {
      /* This is LIFO which means we take from the back of the buffer. */
      if ( v3u64Gt( buffer->tail, buffer->head ) )
      {
        /* This means we have wrapped data.*/
        if ( v3u64Gt( amount, buffer->head ) )
        {
          /* This means we have more data then is on the wrapped side. */
          bot = buffer->head; /* All we can. */
          top = v3u64Sub( amount, bot ); /* The rest. */
        }
        else
        {
          /* We can take it all  from the 'wrapped' side.*/
          top = v3u64u32( 0 );
          bot = amount;
        }

      }
      else
      {
        /* We should take plainly from the end of the buffer. No wrap. */
        bot = v3u64u32( 0 );
        top = amount;
      }

      /* We should now have the amount of data to take from the wrapped
         part of the buffer in bot and the rest in top.
         if top is non zero that means we will not have anything wrapped
         once we are done taking, which means we can move head as below.
         this is really analogous to the FIFO case.*/
      if ( v3u64Gt( top, v3u64u32( 0 ) ) )
      {
        /* This is the 'back' of the buffer.*/
        if ( v3u64Gt( bot, v3u64u32( 0 ) ) )
        {
          ofs = v3u64Sub( buffer->data_length, top );
        }
        else
        {
          ofs = v3u64Sub( buffer->head, top );
        }
        buffer->head = ofs;

        if ( v3MemCopy( data, v3MemOffset( memory, ofs ), top )  == V3_FAIL )
        {
          v3MutexUnlock( &buffer->lock );
          return( v3u64u32( 0 ) );
        }
        /* Data from top.*/
      }
      else
      {
        /* No data from top.*/
        buffer->head = v3u64Sub( buffer->tail, bot );
      }

      if ( v3u64Gt( bot, v3u64u32( 0 ) ) )
      {
        /* Data from bottom. */
        if ( v3MemCopy( v3MemOffset( data_u8, top ), buffer->memory, bot )
          == V3_FAIL )
        {
          v3MutexUnlock( &buffer->lock );
          return( v3u64u32 ( 0 ) );
        }
      }

      buffer->data_length = v3u64Sub( buffer->data_length, amount );
      v3MutexUnlock( &buffer->lock );
      return( amount );
    }
    default:
      break;
  }

  v3MutexUnlock( &buffer->lock );
  return( v3u64u32( 0 ) );
} /* V3BufferGet */


s32 v3BufferUnget( v3_BUFFER * buffer, const void * data, u64 length )
{
  u64 factor, diff, amount;
  u64 bot, top;
  void * mem_temp;

  if ( buffer == NULL )
  {
    return( V3_BUFFER_ERROR_MEMORY );
  }

  v3MutexLock( &buffer->lock, V3_MUTEX_WAIT ); /* We need lock. */

  switch ( buffer->mode )
  {
    case V3_BUFFER_MODE_STACK:
    {
      v3MutexUnlock( &buffer->lock );
      return( v3BufferPut( buffer, data, length ) );
      /* NOREACH */
    }
    case V3_BUFFER_MODE_QUEUE:
    {
      /* Ungetting from a FIFO means to prepend data to the queue. */

      /* First determine how much more memory we need, if any. */
      diff = v3u64Sub( v3u64Add( length, buffer->data_length ),
        buffer->mem_length );

      if ( v3u64Gt( v3u64Add( length, buffer->data_length ),
        buffer->mem_length ) )
      {
        if ( v3u64Eq( buffer->grow_size, v3u64u32( 0 ) ) )
        {
          v3MutexUnlock( &buffer->lock );
          return( V3_BUFFER_ERROR_FULL );
        }

        /* We need more memory. Since we always allocate in even
           multiples of buffer->grow_size we figure that multiple
           out now. */
        factor = v3u64Div( diff, buffer->grow_size );
        if ( !v3u64Eq( v3u64Mod( diff, buffer->grow_size ),
          v3u64u32( 0 ) ) )
        {
          v3u64Inc( &factor );
        }

        amount = v3u64Add( buffer->mem_length,
          v3u64Mul( buffer->grow_size, factor ) );

        if ( (mem_temp = v3MemRealloc( buffer->memory, amount, 0 ) )
          == NULL )
        {
          v3MutexUnlock( &buffer->lock );
          return( V3_BUFFER_ERROR_MEMORY );
        }

        buffer->memory = mem_temp;
        buffer->mem_length = amount;
      }

      /* We now have suitable amounts of memory in buffer->memory. */

      /* Part of it goes before the current tail and the rest goes
       * down from the end. */

      if ( v3u64Gt( buffer->tail , buffer->head ) )
      {
        /* In this case everything should be prepended before
         * tail. */
        top = v3u64u32( 0 );
        bot = length;
      }
      else
      {
        if ( v3u64Gt( length, buffer->tail ) )
        {
          /* In this case everything should be prepended before
           * tail. */
          top = v3u64u32( 0 );
          bot = length;
        }
        else
        {
          /* We need to wrap. */
          bot = buffer->tail;
          top = v3u64Sub( length, bot );
        }
      }

      /* Now bot should contain the amount of data to put before tail
       * and top the amount to put before end of buffer.*/

      /* Start with prepending to tail.*/
      if ( v3u64Gt( bot , v3u64u32( 0 ) ) )
      {
        if ( v3MemCopy( v3MemOffset( buffer->memory ,
          v3u64Sub( buffer->tail, bot ) ) , data , bot ) == V3_FAIL )
        {
          v3MutexUnlock( &buffer->lock );
          return( V3_BUFFER_ERROR_MEMORY );
        }
        buffer->tail = v3u64Sub( buffer->tail , bot );
      }

      /* Now we prepend to the end of the buffer. */
      if ( v3u64Gt( top , v3u64u32( 0 ) ) )
      {
        if ( v3MemCopy( v3MemOffset( buffer->memory,
          v3u64Sub( buffer->mem_length, top ) ),
          v3MemOffset( data, bot ), top ) )
        {
          v3MutexUnlock( &buffer->lock );
          return( V3_BUFFER_ERROR_MEMORY );
        }
        buffer->tail = v3u64Sub( buffer->mem_length, top );
      }
      buffer->data_length = v3u64Add( buffer->data_length, length );
      v3MutexUnlock( &buffer->lock );
      return( V3_PASS );
      /* NOREACH */
    }
    default:
    {
      v3MutexUnlock( &buffer->lock );
      return( V3_BUFFER_ERROR_MODE );
    }
  }
}

/* low-level functions */

s32 v3_BufferPut( v3_BUFFER * buffer, const void * data, u64 length )
{
  u64 factor;
  u64 diff;
  u64 amount, space;
  u64 bot, top;
  void * mem_temp;
  void * memory;

  /* First determine how much more memory we need, if any. */
  diff = v3u64Sub( v3u64Add( length, buffer->data_length ),
    buffer->mem_length );

  if ( v3u64Gt( v3u64Add( length, buffer->data_length ),
    buffer->mem_length ) )
  {
    if ( v3u64Eq( buffer->grow_size, v3u64u32( 0 ) ) )
    {
      return( V3_BUFFER_ERROR_FULL );
    }

    /* We need more memory. Since we always allocate in even
       multiples of buffer->grow_size we figure that multiple
       out now. */
    factor = v3u64Div( diff, buffer->grow_size );
    if ( !v3u64Eq( v3u64Mod( diff, buffer->grow_size ),
      v3u64u32( 0 ) ) )
    {
      v3u64Inc( &factor );
    }

    amount = v3u64Add( buffer->mem_length,
      v3u64Mul( buffer->grow_size, factor ) );

    if ( (mem_temp = v3MemRealloc( buffer->memory, amount,
      V3_MEM_NORMAL ) ) == NULL )
    {
      return( V3_BUFFER_ERROR_MEMORY );
    }
    buffer->memory = mem_temp;
    buffer->mem_length = amount;
  }
  /* routines for increasing buffer space. */
  memory = buffer->memory;

  /* We now have suitable amounts of memory in buffer->memory. */
  /* Since the buffer is circular data we start with figuring out
     how much goes after the tail and how much goes before the
     head. */
  space = v3u64Sub( buffer->mem_length, buffer->head );

  diff = v3u64Sub( space, length );

  if ( v3u64Lt ( diff, v3u64u32( 0 ) ) )
  {
    /* We need to wrap. */
    top = space;
    bot = v3u64Sub( length, space ); /* Really -diff. */
  }
  else
  {
    /* We manage with a single copy to the end. */
    top = length;
    bot = v3u64u32( 0 );
  }

  /* Should we add at the end? */
  if ( v3u64Gt( top, v3u64u32( 0 ) ) )
  {
    if ( v3MemCopy( v3MemOffset( memory, buffer->head ), data, top )
      == V3_FAIL )
    {
      return( V3_BUFFER_ERROR_MEMORY );
    }
    buffer->head = v3u64Add( buffer->head, top );
  }

  /* Should we add down at the buffer start? */
  if ( v3u64Gt( bot, v3u64u32( 0 ) ) )
  {
    if ( v3MemCopy( buffer->memory, data, bot ) == V3_FAIL )
    {
      return( V3_BUFFER_ERROR_MEMORY );
    }
    /* We are wrapping our head. */
    buffer->head = bot;
  }

  buffer->data_length = v3u64Add( buffer->data_length, length );
  return( V3_PASS );
} /* v3_BufferPut */

/* testing */

s32 v3_TestBuffer( void )
{
  v3_BUFFER srcQ1, srcQ2, srcS3, dstQ1, dstS2;
  u32 item1, item2, item3, item4; /* Items for the queue */
  u64 large_item1, large_item2; /* Larger items. */
  u32 i;

  /* Tests totally rewritten.. */

  /* Philsophy of the tests.
   *   A Create some source buffers.
   *   B Fill those buffers.
   *   C Examine them.
   *   D Shuffle from source to destination
   *   E Examine them.
   *   F Get data out of them.
   *   G Check for sizes
   *   H Destruct it all.
   *   I Test some special cases, BufferCreate with item for one..
   *  ... And yes I know they are incomplete.. But they will always
   *   be..
   **/

  /* Assumptions :
   *  1) We will not run out of memory for any natural reasons.
   *   A machine that runs  out of memory during these tests is
   *   useless to us anyhow.
   *  2) The machine is sane. If the machine is insane there is no
   *   way of knowing if the code is.
   */

  /* Test section A - Create some source buffers.*/
  if ( ( v3BufferCreate( &srcQ1, v3u64u32( 10 ), V3_BUFFER_MODE_QUEUE,
    v3u64u32( 10 ), NULL, v3u64u32( 0 ) ) != V3_PASS ) )
  {
    return ( -1 );
  }


  if ( ( v3BufferCreate( &srcQ2, v3u64u32( 10 ), V3_BUFFER_MODE_QUEUE,
    v3u64u32( 20 ), NULL, v3u64u32( 0 ) ) != V3_PASS ) )
  {
    return ( -2 );
  }

  if ( ( v3BufferCreate( &srcS3, v3u64u32( 80 ), V3_BUFFER_MODE_STACK,
    v3u64u32( 0 ), NULL, v3u64u32( 0 ) ) != V3_PASS ) )
    /* This one cannot grow. */
  {
    return ( -3 );
  }

  if ( ( v3BufferCreate( &dstQ1, v3u64u32( 1000 ), V3_BUFFER_MODE_QUEUE,
    v3u64u32( 0 ), NULL, v3u64u32( 0 ) ) != V3_PASS ) )
  {
    return ( -4 );
  }

  if ( ( v3BufferCreate( &dstS2, v3u64u32( 20 ), V3_BUFFER_MODE_STACK,
    v3u64u32( 2 ), NULL, v3u64u32( 0 ) ) != V3_PASS ) )
  {
    return ( -5 );
  }

  /*
   * Test section B - Fill those buffers.
   */

  /* First stuff some small data into the first and third pipe.
   * and some slightly larger data down the second. */
  for( i = 1; i < 21; i++ )
  {
    item1 = i;
    item2 = i + 100; /* Let data differ so we can separate it.*/
    large_item1 = v3u64Mul( v3u64u32(i), v3u64u32( 23000 ) );

    if ( v3BufferPut( &srcQ1, &item1, v3u64u32( sizeof( item1 ) ) )
      != V3_PASS )
    {
      return ( -6 );
    }

    if ( v3BufferPut( &srcQ2, &large_item1,
      v3u64u32( sizeof( large_item1 ) ) )  != V3_PASS )
    {
      return ( -7 );
    }

    if ( v3BufferPut( &srcS3, &item2,
      v3u64u32( sizeof(item2) ) )  != V3_PASS )
    {
      return ( -8 );
    }


  }
  /*
   * Test section C - Examine them.
   */

  /* First sizes.*/
  if ( !v3u64Eq( v3BufferLength( &srcQ1 ), v3u64u32( 80 ) ) )
  {
    return ( -9 );
  }
  if ( !v3u64Eq( v3BufferLength( &srcQ2 ), v3u64u32( 160 ) ) )
  {
    return ( -10 );
  }
  if ( !v3u64Eq( v3BufferLength( &srcS3 ), v3u64u32( 80 ) ) )
  {
    return ( -11 );
  }


  /* srcS3 should be full, we test that.*/
  if ( v3BufferPut( &srcS3, &item2,
    v3u64u32( sizeof( item2 ) ) ) != V3_BUFFER_ERROR_FULL  )
  {
    return ( -12 );
  }
  if ( !v3u64Eq( v3BufferLength( &srcS3 ), v3u64u32( 80 ) ) )
  {
    return ( -13 );
  }

  /* Then we sample some data from each. */
  if ( !v3u64Eq( v3BufferGet( &item1, v3u64u32( sizeof( item1 ) ), &srcQ1 ),
    v3u64u32( sizeof( item1 ) ) ) )
  {
    return ( -14 );
  }
  if ( item1 != 1 ) /* First item entered. */
  {
    return ( -15 );
  }
  if ( v3BufferUnget( &srcQ1, &item1, v3u64u32( sizeof( item1 ) ) )
    != V3_PASS )
  {
    return ( -16 );
  }

  if ( !v3u64Eq( v3BufferGet( &large_item1,
    v3u64u32( sizeof( large_item1 ) ), &srcQ2 ),
    v3u64u32( sizeof( large_item1 ) ) ) )
  {
    return ( -17 );
  }
  if ( !v3u64Eq( large_item1, v3u64u32( 23000 ) ) ) /* First item entered. */
  {
    return ( -18 );
  }
  if ( v3BufferUnget( &srcQ2, &large_item1,
    v3u64u32( sizeof( large_item1 ) ) ) != V3_PASS )
  {
    return ( -19 );
  }

  if ( !v3u64Eq( v3BufferGet( &item1, v3u64u32( sizeof( item1 ) ), &srcS3 ),
    v3u64u32( sizeof( item1 ) ) ) )
  {
    return ( -20 );
  }
  if ( item1 != 120 ) /* Last item entered. */
  {
    return ( -21 );
  }
  if ( v3BufferUnget( &srcS3, &item1, v3u64u32( sizeof(item1) ) )
    != V3_PASS )
  {
    return ( -22 );
  }

  /*
   * Test section D - Shuffle from source to destination
   */
  /* There are 20 items in each buffer.*/
  for ( i=0; i<20; i++ )
  {
    if ( !v3u64Eq( v3BufferGet( &item1, v3u64u32( sizeof( item1 ) ),
      &srcQ1 ), v3u64u32( sizeof( item1 ) ) ) )
    {
      return ( -23 );
    }
    if ( !v3u64Eq( v3BufferGet( &large_item1,
      v3u64u32( sizeof( large_item1 ) ), &srcQ2 ),
      v3u64u32( sizeof( large_item1 ) ) ) )
    {
      return ( -24 );
    }
   if ( !v3u64Eq( v3BufferGet( &item2, v3u64u32( sizeof( item2 ) ), &srcS3 ),
      v3u64u32( sizeof( item2 ) ) ) )
    {
      return ( -25 );
    }

    if ( item1 != ( i + 1 ) )
    {
      return ( -26 );
    }
    if ( item2 != ( (20-i) + 100 ) )
    {
      return ( -27 );
    }
    if ( !v3u64Eq( large_item1,
      v3u64Mul( v3u64u32(i+1), v3u64u32( 23000 ) ) ) )
    {
      return ( -28 );
    }

    /* Everything is fine. Push into dsts.*/
    if ( v3BufferPut( &dstQ1, &item1, v3u64u32( sizeof( item1 ) ) ) !=
      V3_PASS )
    {
      return ( -29 );
    }
    if ( v3BufferPut( &dstS2, &item1, v3u64u32( sizeof( item1 ) ) ) !=
      V3_PASS )
    {
      return ( -30 );
    }

    if ( v3BufferPut( &dstQ1, &item2, v3u64u32( sizeof( item2 ) ) ) !=
      V3_PASS )
    {
      return ( -31 );
    }
    if ( v3BufferPut( &dstS2, &item2, v3u64u32( sizeof( item2 ) ) ) !=
      V3_PASS )
    {
      return ( -32 );
    }

    if ( v3BufferPut( &dstQ1, &large_item1,
       v3u64u32( sizeof( large_item1 ) ) ) !=  V3_PASS )
    {
      return ( -33 );
    }
    if ( v3BufferPut( &dstS2, &large_item1,
       v3u64u32( sizeof( large_item1 ) ) ) !=  V3_PASS )
    {
      return ( -34 );
    }
  }

  /*
   * Test section E - Examine them.
   */

  /* src buffers should now be empty! And the destinations
   * quite loaded. */
  if ( !v3u64Eq( v3BufferLength( &srcQ1 ), v3u64u32( 0 ) ) )
  {
    return ( -35 );
  }
  if ( !v3u64Eq( v3BufferLength( &srcQ2 ), v3u64u32( 0 ) ) )
  {
    return ( -36 );
  }
  if ( !v3u64Eq( v3BufferLength( &srcS3 ), v3u64u32( 0 ) ) )
  {
    return ( -37 );
  }
  if ( !v3u64Eq( v3BufferLength( &dstQ1 ), v3u64u32( 80 + 80 + 160 ) ) )
  {
    return ( -38 );
  }
  if ( !v3u64Eq( v3BufferLength( &dstS2 ), v3u64u32( 80 + 80 + 160 ) ) )
  {
    return ( -39 );
  }

  /*
   * Test section F -  Shuffle some data crosswise
   *   We move from dst buffers back to src, using unget's.
   */
  for ( i = 0; i < 20; i++ )
  {

    if ( !v3u64Eq( v3BufferGet( &item1, v3u64u32( sizeof( item1 ) ),
      &dstQ1 ), v3u64u32( sizeof( item1 ) ) ) )
    {
      return ( -40 );
    }
    if ( !v3u64Eq( v3BufferGet( &item2, v3u64u32( sizeof( item2 ) ),
      &dstQ1 ), v3u64u32( sizeof( item2 ) ) ) )
    {
      return ( -41 );
    }
    if ( !v3u64Eq( v3BufferGet( &large_item1,
      v3u64u32( sizeof( large_item1 ) ), &dstQ1 ),
      v3u64u32( sizeof( large_item1 ) ) ) )
    {
      return ( -42 );
    }

    /* This set of gets is backwards since it is a LIFO. */
    if ( !v3u64Eq( v3BufferGet( &large_item2,
      v3u64u32( sizeof( large_item2 ) ), &dstS2 ),
      v3u64u32( sizeof( large_item2 ) ) ) )
    {
      return ( -43 );
    }
    if ( !v3u64Eq( v3BufferGet( &item4, v3u64u32( sizeof( item3 ) ),
      &dstS2 ), v3u64u32( sizeof( item3 ) ) ) )
    {
      return ( -44 );
    }
    if ( !v3u64Eq( v3BufferGet( &item3, v3u64u32( sizeof( item4 ) ),
      &dstS2 ), v3u64u32( sizeof( item4 ) ) ) )
    {
      return ( -45 );
    }

    /*
     * Check that data is as expected.
     */

    /* item1,item2 and large_item1 belong to dstQ1  */
    if ( item1 != ( i + 1 ) )
    {
      return ( -46 );
    }

    /* Backwards since it has been stored in srcS3. */
    if ( item2 != ( ( 20 - i ) + 100 ) )
    {
      return ( -47 );
    }

    if ( !v3u64Eq( large_item1, v3u64Mul( v3u64u32( i + 1 ),
      v3u64u32( 23000 ) ) ) )
    {
      return ( -48 );
    }

    /* item3, item4 and large_item2 belong to dstS2.
     * These are backwards. */
    if ( item3 != ( 20 - i ) )
    {
      return ( -49 );
    }

    /* Backwards backwards. Been in both dstS2 and srcS3*/
    if ( item4 != ( i + 101 ) )
    {
      return ( -50 );
    }
    if ( !v3u64Eq( large_item2, v3u64Mul( v3u64u32( 20 - i ),
      v3u64u32( 23000 ) ) ) )
    {
      return ( -51 );
    }
  }

  /*
   * Test section G - Check for sizes.
   */

  /* We checked src buffers above, no changes have been made to them. */
  if ( !v3u64Eq( v3BufferLength( &dstQ1 ), v3u64u32( 0 ) ) )
  {
    return ( -52 );
  }
  if ( !v3u64Eq( v3BufferLength( &dstS2 ), v3u64u32( 0 ) ) )
  {
    return ( -53 );
  }

  /*
   * Test section H - Destruct it all
   */
  v3BufferFree( &srcQ1 );
  v3BufferFree( &srcQ2 );
  v3BufferFree( &srcS3 );
  v3BufferFree( &dstQ1 );
  v3BufferFree( &dstS2 );

  /*
   * Test section I - Test some special cases
   */

  /* I.1 Testing BufferCreate with item as argument.*/
  large_item1 = v3u64u32( 321 );
  if ( v3BufferCreate( &srcS3, v3u64u32( 10 ), V3_BUFFER_MODE_STACK,
    v3u64u32( 29 ), &large_item1, v3u64u32( sizeof( large_item1 ) ) )
    != V3_PASS )
  {
    return ( -54 );
  }

  for ( i = 0 ; i < 10 ; i++ )
  {
    item1 = i * 3;
    if ( v3BufferPut( &srcS3, &item1, v3u64u32( sizeof( item1 ) ) )
      != V3_PASS )
    {
      return( -55 );
    }
  }

  for ( i = 0; i < 10; i++ )
  {
    if ( !v3u64Eq( v3BufferGet( &item1, v3u64u32( sizeof( item1 ) ), &srcS3 ),
      v3u64u32( sizeof( item1 ) ) ) )
    {
      return( -56 );
    }
  }

  return( V3_PASS );
}
