/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

#ifndef V3_BUFFER_H
#define V3_BUFFER_H

#include "cputypes.h"
#include "cosmtask.h"

typedef struct
{
  u64 data_length;
  u64 mem_length;
  u64 grow_size;
  u64 head;
  u64 tail;
  void * memory;
  u32 mode;
  v3_MUTEX lock;
} v3_BUFFER;

/*
  Notes:
  Buffers never have to worry about secure memory situations
  Buffers are implemented as circular buffers
*/

#define V3_BUFFER_MODE_NONE      0
#define V3_BUFFER_MODE_QUEUE     1 /* FIFO */
#define V3_BUFFER_MODE_STACK     2 /* FILO */

#define V3_BUFFER_ERROR_NONE     0
#define V3_BUFFER_ERROR_MEMORY  -1 /* Not enough memory */
#define V3_BUFFER_ERROR_FULL    -2 /* Buffer is full */
#define V3_BUFFER_ERROR_MODE    -3 /* Unknown mode */

s32 v3BufferCreate( v3_BUFFER * buffer, u64 size, u32 mode, u64 grow,
  const void * const data, u64 length );
  /*
    Setup the v3_BUFFER struct and do an initial memory allocation of size
    bytes save alot of reallocs later. (i.e. good chance to boost performance)
    grow is the amount to increase the buffer by when more space is needed.
    If grow is zero, it will not increase and may return a FULL error.
    If data is non-NULL then copy length bytes of initial data into buffer.
    Returns: V3_PASS on success, or an error code on failure.
  */

void v3BufferFree( v3_BUFFER * buffer );
  /*
    Free all data in the buffer and return it to an uninitialized state.
    Returns: nothing.
  */

u64 v3BufferLength( v3_BUFFER * buffer );
  /*
    Returns: the length of the data currently in the buffer.
  */

s32 v3BufferPut( v3_BUFFER * buffer, const void * data, u64 length );
  /*
    Insert length bytes of data into the buffer.
    Returns: V3_PASS on success, or an error code on failure.
  */

u64 v3BufferGet( void * data, u64 length, v3_BUFFER * buffer );
  /*
    Remove up to length bytes of data from the buffer.
    Returns: length on success, otherwise bytes read out of the buffer.
  */

s32 v3BufferUnget( v3_BUFFER * buffer, const void * data, u64 length );
  /*
    Insert length bytes of data back into the buffer after it was taken out
    with a v3BufferGet().
    We shouldn't need to warn you about using this function without thinking
    in a multithreaded environment.
    Returns: V3_PASS on success, or an error code on failure.
  */

/* Low level function. */

s32 v3_BufferPut( v3_BUFFER * buffer, const void * data, u64 length );
  /*
    This function does the work for v3BufferPut, except for no locking, also
    called from v3BufferCreate.
    Returns: V3_PASS on success, or an error code on failure
  */

/* testing */

s32 v3_TestBuffer( void );
  /*
    Test functions in this header.
    Returns: V3_PASS on success, or a negative number corresponding to the
      test that failed.
  */

#endif
