/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

#include "compress.h"
#include "bzip2.h"
#include "cosmmath.h"
#include "cosmmem.h"

s32 v3CompressInit( v3_BUFFER * output, v3_COMPRESS_TMP * tmp,
  u32 type, u32 level )
{
  s32 result;
  u64 size, grow;

  if ( ( output == NULL ) || ( tmp == NULL ) ||
    ( level < 1 ) || ( level > 9 ) )
  {
    return( V3_COMPRESS_ERROR_PARAM );
  }

  v3MutexLock( &tmp->lock, V3_MUTEX_WAIT );

  if ( tmp->tmp_data != NULL )
  {
    /* dont want to lose a working stream */
    v3MutexUnlock( &tmp->lock );
    return( V3_COMPRESS_ERROR_STATE );
  }

  _V3_SET64( size, 00000000, 00000400 );
  _V3_SET64( grow, 00000000, 00000400 );
  if ( v3BufferCreate( output, size, V3_BUFFER_MODE_QUEUE, grow, NULL,
    v3u64u32( 0 ) ) == V3_FAIL )
  {
    v3MutexUnlock( &tmp->lock );
    return( V3_COMPRESS_ERROR_NOMEM );
  }

  switch( type )
  {
    case V3_COMPRESS_BZIP2:
      tmp->tmp_data = v3MemAlloc( v3u64u32( sizeof( bz_stream ) ),
        V3_MEM_NORMAL );
      if ( tmp->tmp_data == NULL )
      {
        tmp->state = V3_COMPRESS_STATE_ERROR;
        v3BufferFree( output );
        v3MutexUnlock( &tmp->lock );
        return( V3_COMPRESS_ERROR_STATE );
      }
      result = BZ2_bzCompressInit( tmp->tmp_data, level, 0, 0 );
      if ( result != BZ_OK )
      {
        tmp->state = V3_COMPRESS_STATE_ERROR;
        v3BufferFree( output );
        v3MemFree( tmp->tmp_data );
        v3MutexUnlock( &tmp->lock );
        return( V3_COMPRESS_ERROR_STATE );
      }
      break;
    default:
      v3MutexUnlock( &tmp->lock );
      return( V3_COMPRESS_ERROR_TYPE );
  }

  tmp->type = type;
  tmp->state = V3_COMPRESS_STATE_INIT;
  tmp->mode = V3_COMPRESS_MODE_COMP;

  v3MutexUnlock( &tmp->lock );

  return( V3_PASS );
}

s32 v3Compress( v3_BUFFER * output, v3_COMPRESS_TMP * tmp,
  const void * const data, u64 length )
{
/* This is where the problem lies. The first thing that should be checked
   is pointers. I'm could be doing something very very wrong with some of
   them. Also check casts. Some of the pointer assignment, dereferencing,
   etc. is probably wrong as you wouldn't believe what I can do to some code
   after it's 3am and it won't work, and I don't have time to fix it back now
   so don't laugh :). If you still can't get it to work, look into
   bzlib.c (bzip2 distribution source), how the bzWrite* (or something like
   that) do it. It would be a "good thing"(tm) to check our bzip2.c with
   that code. May the source with you.
*/
#ifdef WORKING_RIGHT
  s32 result;
  bz_stream * strm;

  if ( ( output == NULL ) || ( tmp == NULL ) || ( data == NULL ) )
    return( V3_COMPRESS_ERROR_PARAM );

  if ( ( tmp->state != V3_COMPRESS_STATE_DATA &&
    tmp->state != V3_COMPRESS_STATE_INIT ) || tmp->tmp_data == NULL)
    return( V3_COMPRESS_ERROR_STATE );

  v3MutexLock( &tmp->lock, V3_MUTEX_WAIT );

  tmp->state = V3_COMPRESS_STATE_DATA;

  strm=tmp->tmp_data;

  switch( tmp->type )
  {
    case V3_COMPRESS_BZIP2:
      strm->next_in = data;
      strm->next_out = output->memory;
      strm->avail_in = v3u32u64( length );
      strm->avail_out = v3u32u64( output->mem_length );
      result = bzCompress( strm, BZ_FINISH );
      break;

    default:
      tmp->state = V3_COMPRESS_STATE_ERROR;
      v3MutexUnlock( &tmp->lock );
      return( V3_COMPRESS_ERROR_STATE );
  }

  v3MutexUnlock( &tmp->lock );
#endif /* working_right */

  return( V3_PASS );
}

s32 v3CompressEnd( v3_BUFFER * output, v3_COMPRESS_TMP * tmp )
{
  if ( ( output == NULL ) || ( tmp == NULL ) )
  {
    return( V3_COMPRESS_ERROR_PARAM );
  }

  v3MutexLock( &tmp->lock, V3_MUTEX_WAIT );

  if ( ( ( tmp->state != V3_COMPRESS_STATE_INIT ) &&
    ( tmp->state != V3_COMPRESS_STATE_DATA ) ) || ( tmp->tmp_data == NULL ) ||
    ( tmp->mode != V3_COMPRESS_MODE_COMP ) )
  {
    v3MutexUnlock( &tmp->lock );
    return( V3_COMPRESS_ERROR_STATE );
  }


  switch( tmp->type )
  {
    case V3_COMPRESS_BZIP2:
      /* finish compressing stream*/
      /* !!! */
      if ( BZ2_bzCompressEnd( tmp->tmp_data ) != BZ_OK )
      {

      }
      break;
    default:
      tmp->state = V3_COMPRESS_STATE_ERROR;
      v3MutexUnlock( &tmp->lock );
      return( V3_COMPRESS_ERROR_TYPE );
  }

  tmp->state = V3_COMPRESS_STATE_DONE;
  v3MemFree( tmp->tmp_data );

  v3MutexUnlock( &tmp->lock );
  v3MutexFree( &tmp->lock );

  return( V3_PASS );
}

s32 v3DecompressInit( v3_BUFFER * output, v3_COMPRESS_TMP * tmp, u32 type )
{
  return( V3_PASS );
}

s32 v3Decompress( v3_BUFFER * output, v3_COMPRESS_TMP * tmp,
  const void * const data, u64 length )
{
  return( V3_PASS );
}

s32 v3DecompressEnd( v3_BUFFER * output, v3_COMPRESS_TMP * tmp )
{
  return( V3_PASS );
}

s32 v3_TestCompress( void )
{
  v3_BUFFER buf;
  v3_COMPRESS_TMP tmp;

  v3MemSet( &buf, v3u64u32( sizeof( v3_BUFFER ) ), 0 );
  v3MemSet( &tmp, v3u64u32( sizeof( v3_COMPRESS_TMP ) ), 0 );

  if ( v3CompressInit( &buf, &tmp, V3_COMPRESS_BZIP2, 9 ) != V3_PASS )
  {
    return( -1 );
  }
  if ( v3CompressEnd( &buf, &tmp ) != V3_PASS )
  {
    return( -2 );
  }

  return( V3_PASS );
}
