/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

/* CPU/OS Layer - CPU and OS specific code is allowed */

#ifndef V3_COSMFILE_H
#define V3_COSMFILE_H

#include "cputypes.h"
#include "cosmtask.h"

#if ( OS_TYPE == OS_MACOS )
#include <Fsp_fopen.h>
#elif ( ( OS_TYPE == OS_OS2 ) || ( OS_TYPE == OS_WIN32 ) )
#else
#if ( OS_TYPE == OS_FREEBSD )
#include <sys/types.h>
#endif
#include <dirent.h>
#endif

#define V3_FILE_STATUS_CLOSED   0
#define V3_FILE_STATUS_OPEN     1

#define V3_FILE_MODE_NONE     0x00
#define V3_FILE_MODE_EXIST    0x01 /* Test for file existence only */
#define V3_FILE_MODE_READ     0x02 /* Open for reading */
#define V3_FILE_MODE_WRITE    0x04 /* Open for writing */
#define V3_FILE_MODE_APPEND   0x08 /* Open at end of file, write only */
#define V3_FILE_MODE_CREATE   0x10 /* Create the file if it doesn't exist */
#define V3_FILE_MODE_TRUNCATE 0x20 /* Truncate an existing file on open */
#define V3_FILE_MODE_SYNC     0x40 /* Write functions sync() before return */

/* these will probably be made automatic and non-optional */
#define V3_FILE_LOCK_NONE       0x00 /* No Locking */
#define V3_FILE_LOCK_READ       0x01 /* Shared read lock */
#define V3_FILE_LOCK_WRITE      0x02 /* Exclusive write lock */

#define V3_FILE_ERROR_NONE      0
#define V3_FILE_ERROR_DENIED    -1   /* Access Denied */
#define V3_FILE_ERROR_NOTFOUND  -2   /* File/Handle Not Found */
#define V3_FILE_ERROR_READMODE  -3   /* Read Access denied */
#define V3_FILE_ERROR_WRITEMODE -4   /* Write access denied */
#define V3_FILE_ERROR_READLOCK  -5   /* Unable to read lock */
#define V3_FILE_ERROR_WRITELOCK -6   /* Unable to write lock */
#define V3_FILE_ERROR_EOF       -7   /* End Of File */
#define V3_FILE_ERROR_APPEND    -8   /* Unable to append */
#define V3_FILE_ERROR_CREATE    -9   /* Unable to create new file */
#define V3_FILE_ERROR_CLOSED    -10  /* File is closed */
#define V3_FILE_ERROR_SEEK      -11  /* Seek Error */
#define V3_FILE_ERROR_NAME      -12  /* Invalid Filename */
#define V3_FILE_ERROR_MODE      -13  /* Open flags conflict */
#define V3_FILE_ERROR_UNLOCK    -14  /* Unable to unlock */
#define V3_FILE_ERROR_TELL      -15  /* Tell Error */
#define V3_FILE_ERROR_LENGTH    -16  /* Unable to discover file size */
#define V3_FILE_ERROR_NOSPACE   -17  /* Device is full */

#define V3_FILE_MAX_FILENAME    255

typedef ascii v3_FILENAME[V3_FILE_MAX_FILENAME + 1];

/*
  Paths are always in the cosm "native|/path/path2/filename" format
  This is important for the system because any platform that can will do a
  chroot() before running, non-chroot systems will tack addition path on the
  front at runtime. The 'other' slash is translated on the fly if needed, but
  internally, it's all stored in the same format. The native portion of the
  path (everything before and including the '|') is platform specific for
  devices/nodes on weird OSes. e.g. "e:|/temp/a.txt"(win32), or
  "//4|/tmp/a.txt" (qnx). If paths are ever sent over the net, this native
  portion will be stripped.
*/

typedef struct
{
  v3_FILENAME filename;
  u64 offset;
  u64 length;
  u64 handle;
  u32 status;
  s32 last_error;
  u32 mode;
  u32 lockmode;
  v3_MUTEX lock;
} v3_FILE;

#define V3_FILE_TYPE_UNKNOWN   0  /* Unknown, filessytem error, etc */
#define V3_FILE_TYPE_FILE      1  /* Normal file */
#define V3_FILE_TYPE_DIR       2  /* Directory */
#define V3_FILE_TYPE_DEVICE    3  /* OS device file/handle */
#define V3_FILE_TYPE_SPECIAL   4  /* Something valid, but an above type */

#define V3_FILE_RIGHTS_NONE    0x00 /* No access */
#define V3_FILE_RIGHTS_READ    0x01 /* Readable */
#define V3_FILE_RIGHTS_WRITE   0x02 /* Writeable (implies append) */
#define V3_FILE_RIGHTS_APPEND  0x04 /* Appendable */
#define V3_FILE_RIGHTS_EXEC    0x08 /* Executeable program */

typedef struct
{
  u64 length;
  u32 type;
  u32 rights;
  v3_time create;
  v3_time modify;
  v3_time access;
} v3_FILE_INFO;

#define V3_DIR_STATUS_CLOSED   0
#define V3_DIR_STATUS_OPEN     1

#define V3_DIR_MODE_NONE       0x00
#define V3_DIR_MODE_EXIST      0x01 /* Test for dir existance only */
#define V3_DIR_MODE_READ       0x02 /* Open for reading only */
#define V3_DIR_MODE_CREATE     0x04 /* Create the dir if it doesnt exist */

#define V3_DIR_ERROR_NONE      0
#define V3_DIR_ERROR_DENIED    -1  /* Access Denied */
#define V3_DIR_ERROR_NOTFOUND  -2  /* Directory not found */
#define V3_DIR_ERROR_READMODE  -3  /* Read access denied */
#define V3_DIR_ERROR_EOF       -4  /* End of Directory */
#define V3_DIR_ERROR_CREATE    -5  /* Unable to create new directory */
#define V3_DIR_ERROR_NOTEMPTY  -6  /* Directory is not empty (no delete) */
#define V3_DIR_ERROR_MODE      -7  /* Open flags conflict */
#define V3_DIR_ERROR_CLOSED    -8  /* Directory is closed */
#define V3_DIR_ERROR_NAME      -9  /* Invalid Dirname */

typedef struct
{
  v3_FILENAME dirname;
  u64 offset;
  u32 status;
  s32 last_error;
  v3_MUTEX lock;
/* dirctories are handled far differently by every OS */
#if ( ( OS_TYPE == OS_MACOS ) || ( OS_TYPE == OS_OS2 ) )
  s32 handle;
#elif ( OS_TYPE == OS_WIN32 )
  HANDLE handle;
  WIN32_FIND_DATA find_data;
#else
  DIR * handle;
#endif
} v3_DIR;

/*
  File Functions
*/

s32 v3FileOpen( v3_FILE * file, const ascii * filename, u32 mode, u32 lock );
  /*
    Attempt to open the filename in the mode with the lock.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

u64 v3FileRead( void * buffer, u64 length, v3_FILE * file );
  /*
    Read length bytes from the file into the buffer.
    Returns: 0 or < length if an error - set last_error, length on success.
  */

u64 v3FileWrite( v3_FILE * file, const void * const buffer, u64 length );
  /*
    Write length bytes to the file from the buffer.
    Returns: 0 or < length if an error - set last_error, length on success.
  */

s32 v3FileSeek( v3_FILE * file, u64 offset );
  /*
    Move the current file offset (in bytes).
    If the file is open in mode V3_FILE_MODE_APPEND, you can't seek.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

s32 v3FileTell( u64 * offset, v3_FILE * file );
  /*
    Get the current file offset.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

s32 v3FileEOF( v3_FILE * file );
  /*
    Test EOF state.
    Returns: V3_PASS if there is more to read, V3_FILE_ERROR_EOF if EOF, or
      an error code on failure.
  */

s32 v3FileLength( u64 * length, v3_FILE * file );
  /*
    Set length to the length of the open file.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

s32 v3FileTruncate( v3_FILE * file, u64 length );
  /*
    Truncate the file at length bytes.
    Current offset will be moved to end of file.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

s32 v3FileDelete( const ascii * filename );
  /*
    Attempt to delete the file.
    Deletion should involve a full wipe of the file data.
    Returns: V3_PASS on success, or an error code on failure.
  */

s32 v3FileInfo( v3_FILE_INFO * info, const ascii * filename );
  /*
    Get Information about a file.
    Returns: V3_PASS on success, or an error code on failure.
  */

s32 v3FileClose( v3_FILE * file );
  /*
    Close the file.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

/*
  Directory Functions
*/

s32 v3DirOpen( v3_DIR * dir, const ascii * dirname, u32 mode );
  /*
    Attempt to open the dirname in the mode.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

u64 v3DirRead( v3_FILENAME * buffer, u64 length, v3_DIR * dir );
  /*
    Read length filenames from the dir into the buffer.
    Returns: 0 or < length if an error - set last_error, length on success.
  */

s32 v3DirDelete( const ascii * dirname );
  /*
    Attempt to delete the dir. Will only work if the directory is empty.
    Returns: V3_PASS on success, or an error code on failure.
  */

s32 v3DirClose( v3_DIR * dir );
  /*
    Close the directory.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

/*
  Utility Functions to deal with endian issues and keep
  everything portable. ALL data must be run through these when moving
  from the machine to the network or a file. We need to deal with the
  (normal) case where the pointers are the same memory location.
  We don't need different functions for signed data types since we're just
  swapping bytes.
*/

void v3u16Load( u16 * num, const u8 * bytes );
  /*
    Load the big endian bytes into native u16/s16
  */

void v3u16Save( u8 * bytes, const u16 * const num );
  /*
    Save the native u16/s16 num into big endian bytes
  */

void v3u32Load( u32 * num, const u8 * bytes );
  /*
    Load the big endian bytes into native u32/s32
  */

void v3u32Save( u8 * bytes, const u32 * const num );
  /*
    Save the native u32/s32 num into big endian bytes
  */

void v3u64Load( u64 * num, const u8 * bytes );
  /*
    Load the big endian bytes into native u64/s64
  */

void v3u64Save( u8 * bytes, const u64 * const num );
  /*
    Save the native u64/s64 num into big endian bytes
  */

void v3u128Load( u128 * num, const u8 * bytes );
  /*
    Load the big endian bytes into native u128/s128
  */

void v3u128Save( u8 * bytes, const u128 * const num );
  /*
    Save the native u128/s128 num into big endian bytes
  */

/* Low level Functions */

s32 v3_FileNativePath( ascii * native, const ascii * path );
  /*
    Translate path into a native system formatted path.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

u64 v3_FileWrite( v3_FILE * file, const void * const buffer, u64 length );
  /*
    Write length bytes to the file from the buffer. Mutex already set.
    Returns: 0 or < length if an error - set last_error, length on success.
  */

s32 v3_FileSeek( v3_FILE * file, u64 offset );
  /*
    Move the current file offset (in bytes). Mutex already set.
    If the file is open in mode V3_FILE_MODE_APPEND, you can't seek.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

s32 v3_FileTell( u64 * offset, v3_FILE * file );
  /*
    Get the current file offset. Mutex already set.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

s32 v3_FileLength( u64 * length, v3_FILE * file );
  /*
    Set length to the length of the open file. Mutex already set.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

s32 v3_FileTruncate( v3_FILE * file, u64 length );
  /*
    Truncate the file at length bytes. Mutex already set.
    Current offset will be moved to end of file.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

s32 v3_FileLock( v3_FILE * file, u32 lock );
  /*
    Lock an open file. Mutex already set.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

s32 v3_FileUnLock( v3_FILE * file );
  /*
    Unlock an open file. Mutex already set.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

s32 v3_FileWipe( v3_FILE * file, u64 length );
  /*
    Securely wipe all file data past length bytes into the file which
    must be open and writeable.
    Current offset will be moved to new end of file.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

/* testing */

s32 v3_TestCosmFile( void );
  /*
    Test functions in this header.
    Returns: V3_PASS on success, or a negative number corresponding to the
      test that failed.
  */

#endif
