/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

/* CPU/OS Layer - CPU and OS specific code is allowed */

#ifndef V3_COSMMATH_H
#define V3_COSMMATH_H

#include "cputypes.h"

/* include the ANSI math function headers */
#if ( !defined( NO_IEEE_FLOAT ) )
#include <math.h>
#endif

/*
  Define arithmatic (+ - * / % ++ --) logical (& | ^ ~ << >>),
  and comparison ( == > < ) for u64, s64, u128, and s128
  Portability and clarity are important here, not speed.
  We do not need logical operations for signed values, use a typecase
  to unsigned value.
  Note: != <= >= aren't needed, that would just reverse the if statement.
*/

/*
  Convertions
*/

/* bigger <- smaller */
#if ( defined( CPU_32BIT ) )
u64 v3u64u32( u32 a ); /* u64 = u32 */
s64 v3s64s32( s32 a ); /* s64 = s32 */
#else /* CPU_64BIT */
#define v3u64u32( a ) ( (u64) a )
#define v3s64s32( a ) ( (s64) a )
#endif /* bits */
u128 v3u128u64( u64 a ); /* u128 = u64 */
s128 v3s128s64( s64 a ); /* s128 = s64 */
u128 v3u128u32( u32 a ); /* u128 = u32 */
s128 v3s128s32( s32 a ); /* s128 = s32 */

/* smaller <- bigger - truncation will occur */
#if ( defined( CPU_32BIT ) )
#define v3u32u64( a ) ( (u32) a.lo )
#define v3s32s64( a ) ( (s32) a.lo )
u64 v3u64u128( u128 a ); /* u64 = u128 */
s64 v3s64s128( s128 a ); /* s64 = s128 */
#define v3u32u128( a ) ( (u32) a.lo.lo )
s32 v3s32s128( s128 a ); /* s32 = s128 */
#else /* CPU_64BIT */
#define v3u32u64( a ) ( (u32) ( a & 0xFFFFFFFF ) )
#define v3s32s64( a ) ( (s32) ( a & 0xFFFFFFFF ) )
#define v3u64u128( a ) ( (u64) a.lo )
#define v3s64s128( a ) ( (s64) a.lo )
#define v3u32u128( a ) ( (u32) ( a.lo & 0xFFFFFFFF ) )
#define v3s32s128( a ) ( (s32) ( a.lo & 0xFFFFFFFF ) )
#endif /* bits */

/* signed <-> unsigned, sign may change */
#if ( defined( CPU_32BIT ) )
u64 v3u64s64( s64 a ); /* u64 <- s64 */
s64 v3s64u64( u64 a ); /* s64 <- u64 */
#else /* CPU_64BIT */
#define v3u64s64( a ) ( (u64) a )
#define v3s64u64( a ) ( (s64) a )
#endif /* bits */
u128 v3u128s128( s128 a ); /* u128 <- s128 */
s128 v3s128u128( u128 a ); /* s128 <- u128 */

/*
  u64 Functions
*/

#if ( defined( CPU_32BIT ) )
u64 v3u64Add( u64 a, u64 b );  /* c = a + b */
u64 v3u64Sub( u64 a, u64 b );  /* c = a - b */
u64 v3u64Mul( u64 a, u64 b );  /* c = a * b */
u64 v3u64Div( u64 a, u64 b );  /* c = a / b */
u64 v3u64Mod( u64 a, u64 b );  /* c = a % b */
u64 v3u64Inc( u64 * a );       /* (*a)++ */
u64 v3u64Dec( u64 * a );       /* (*a)-- */

u64 v3u64And( u64 a, u64 b );  /* c = a & b */
u64 v3u64Or( u64 a, u64 b );   /* c = a | b */
u64 v3u64Xor( u64 a, u64 b );  /* c = a ^ b */
u64 v3u64Not( u64 a );         /* c = ^a */
u64 v3u64Lsh( u64 a, u32 x );  /* c = a << x */
u64 v3u64Rsh( u64 a, u32 x );  /* c = a >> x */

u32 v3u64Eq( u64 a, u64 b );   /* a == b ? 1 : 0; */
u32 v3u64Gt( u64 a, u64 b );   /* a > b ? 1 : 0; */
u32 v3u64Lt( u64 a, u64 b );   /* a < b ? 1 : 0; */
#else /* CPU_64BIT */
#define v3u64Add( a, b ) ( a + b )
#define v3u64Sub( a, b ) ( a - b )
#define v3u64Mul( a, b ) ( a * b )
#define v3u64Div( a, b ) ( ( b == 0 ) ? 0 : a / b )
#define v3u64Mod( a, b ) ( ( b == 0 ) ? 0 : a % b )
#define v3u64Inc( a ) ( (*a)++ )
#define v3u64Dec( a ) ( (*a)-- )

#define v3u64And( a, b ) ( a & b )
#define v3u64Or( a, b ) ( a | b )
#define v3u64Xor( a, b ) ( a ^ b )
#define v3u64Not( a ) ( ~a )
#define v3u64Lsh( a, x ) ( a << x )
#define v3u64Rsh( a, x ) ( a >> x )

#define v3u64Eq( a, b ) ( a == b )
#define v3u64Gt( a, b ) ( a > b )
#define v3u64Lt( a, b ) ( a < b )
#endif /* bits */

/*
  s64 Functions
*/

#if ( defined( CPU_32BIT ) )
s64 v3s64Add( s64 a, s64 b );  /* c = a + b */
s64 v3s64Sub( s64 a, s64 b );  /* c = a - b */
s64 v3s64Mul( s64 a, s64 b );  /* c = a * b */
s64 v3s64Div( s64 a, s64 b );  /* c = a / b */
s64 v3s64Mod( s64 a, s64 b );  /* c = a % b */
s64 v3s64Inc( s64 * a );       /* (*a)++ */
s64 v3s64Dec( s64 * a );       /* (*a)-- */

u32 v3s64Eq( s64 a, s64 b );   /* a == b ? 1 : 0; */
u32 v3s64Gt( s64 a, s64 b );   /* a > b ? 1 : 0; */
u32 v3s64Lt( s64 a, s64 b );   /* a < b ? 1 : 0; */
#else /* CPU_64BIT */
#define v3s64Add( a, b ) ( a + b )
#define v3s64Sub( a, b ) ( a - b )
#define v3s64Mul( a, b ) ( a * b )
#define v3s64Div( a, b ) ( ( b == 0 ) ? 0 : a / b )
s64 v3s64Mod( s64 a, s64 b );  /* c = a % b */
#define v3s64Inc( a ) ( (*a)++ )
#define v3s64Dec( a ) ( (*a)-- )

#define v3s64Eq( a, b ) ( a == b )
#define v3s64Gt( a, b ) ( a > b )
#define v3s64Lt( a, b ) ( a < b )
#endif /* bits */

/*
  u128 Functions
*/

u128 v3u128Add( u128 a, u128 b );  /* c = a + b */
u128 v3u128Sub( u128 a, u128 b );  /* c = a - b */
u128 v3u128Mul( u128 a, u128 b );  /* c = a * b */
u128 v3u128Div( u128 a, u128 b );  /* c = a / b */
u128 v3u128Mod( u128 a, u128 b );  /* c = a % b */
u128 v3u128Inc( u128 * a );        /* (*a)++ */
u128 v3u128Dec( u128 * a );        /* (*a)-- */

u128 v3u128And( u128 a, u128 b );  /* c = a & b */
u128 v3u128Or ( u128 a, u128 b );  /* c = a | b */
u128 v3u128Xor( u128 a, u128 b );  /* c = a ^ b */
u128 v3u128Not( u128 a );          /* c = ^a */
u128 v3u128Lsh( u128 a, u32 x );   /* c = a << x */
u128 v3u128Rsh( u128 a, u32 x );   /* c = a >> x */

u32 v3u128Eq( u128 a, u128 b );    /* a == b ? 1 : 0; */
u32 v3u128Gt( u128 a, u128 b );    /* a > b ? 1 : 0; */
u32 v3u128Lt( u128 a, u128 b );    /* a < b ? 1 : 0; */

/*
  s128 Functions
*/

s128 v3s128Add( s128 a, s128 b );  /* c = a + b */
s128 v3s128Sub( s128 a, s128 b );  /* c = a - b */
s128 v3s128Mul( s128 a, s128 b );  /* c = a * b */
s128 v3s128Div( s128 a, s128 b );  /* c = a / b */
s128 v3s128Mod( s128 a, s128 b );  /* c = a % b */
s128 v3s128Inc( s128 * a );        /* (*a)++ */
s128 v3s128Dec( s128 * a );        /* (*a)-- */

u32 v3s128Eq( s128 a, s128 b );    /* a == b ? 1 : 0; */
u32 v3s128Gt( s128 a, s128 b );    /* a > b ? 1 : 0; */
u32 v3s128Lt( s128 a, s128 b );    /* a < b ? 1 : 0; */

/* Low level functions */

void v3_u64DivMod( u64 a, u64 b, u64 * div, u64 * mod );
  /* div = a / b; mod = a % b; */

void v3_u128DivMod( u128 a, u128 b, u128 * div, u128 * mod );
  /* div = a / b; mod = a % b; */

/* testing */

s32 v3_TestCosmMath( void );
  /*
    Test functions in this header.
    Returns: V3_PASS on success, or a negative number corresponding to the
      test that failed.
  */

#endif
