/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

/* CPU/OS Layer - CPU and OS specific code is allowed */

#include "cputypes.h"
#include "cosmmem.h"
#include "cosmmath.h"
#include "cosmio.h"

#include <stdlib.h>
#include <string.h> /* for memmove */

#define V3_MEM_64BITS  64
#define V3_MEM_32BITS  32

#undef V3_MEM_SIZE
#if ( ( OS_TYPE == OS_TRU64 ) || ( OS_TYPE == OS_IRIX64 ) \
  || ( ( OS_TYPE == OS_SOLARIS ) && ( CPU_WORD_SIZES == 842 ) ) \
  || ( ( OS_TYPE == OS_LINUX ) && ( CPU_TYPE == CPU_ALPHA ) ) )
#define V3_MEM_SIZE V3_MEM_64BITS
#else
#define V3_MEM_SIZE V3_MEM_32BITS
#endif

/* one of the above must be set at this point */
#if ( !defined( V3_MEM_SIZE ) )
#error "Define V3_MEM_SIZE - see cosmmem.c"
#endif


#define V3_MEM_IS_SECURE_NO  0
#define V3_MEM_IS_SECURE_YES 1

/* Someday I hope this line will need an #if */
#define V3_MEM_IS_SECURE V3_MEM_IS_SECURE_NO

void * v3MemAlloc( u64 bytes, u32 secure )
{
  void * mem;
  u64 align;
  u64 request;

  /* Return NULL if bytes is 0 */
  if ( v3u64Eq( bytes, v3u64u32( 0 ) )
    || ( ( secure != V3_MEM_NORMAL ) && ( secure != V3_MEM_NORMAL ) ) )
  {
    return( NULL );
  }

#if ( V3_MEM_IS_SECURE == V3_MEM_IS_SECURE_NO ) /* not secure */
  if ( secure == V3_MEM_SECURE )
  {
    v3MemWarning();
    return( NULL );
  }
#endif

/*
  we still need to find a way to make sure we return a
  16 byte aligned address, but it should be done already
*/
  _V3_SET64( align, 00000000, 00000010 );
  request = v3u64Mul( v3u64Div( bytes, align ), align );
  if ( !v3u64Eq( request, bytes ) )
  {
    request = v3u64Add( request, align );
  }

  /* check for alignment overflow */
  if ( v3u64Lt( request, bytes ) )
  {
    return( NULL );
  }

#if ( V3_MEM_SIZE == V3_MEM_64BITS )
  mem = calloc( 1, request );
#else /* 32bits */
  if ( v3u64Gt( bytes, v3u64u32( (u32) 0xFFFFFFFF ) ) )
  {
    return( NULL );
  }
  mem = calloc( 1, v3u32u64( request ) );
#endif

  if ( mem == NULL )
  {
    return( NULL );
  }

  if ( secure == V3_MEM_SECURE )
  {
    /*
      Do as much as possible to make sure the RAM is secure. Some OSes
      will allow us to at least tighted security a little bit.
      Lock Memory into RAM, secure it from other processes, etc.
      So far mainstream OS security is a joke and we can't do much.
    */
  }

  return( mem );
}

void * v3MemRealloc( void * memory, u64 bytes, u32 secure )
{
  u64 align;
  u64 request;

  /* Free memory and return NULL if bytes is 0 and 'memory' is not NULL */
  if ( ( v3u64Eq( bytes, v3u64u32( 0 ) ) ) && ( memory != NULL ) )
  {
    v3MemFree( memory );
    return( NULL );
  }

#if ( V3_MEM_IS_SECURE == V3_MEM_IS_SECURE_NO )
  if ( secure == V3_MEM_SECURE )
  {
    v3MemWarning();
    return( NULL );
  }
#endif

  /* alignment issues */
  _V3_SET64( align, 00000000, 00000010 );
  request = v3u64Mul( v3u64Div( bytes, align ), align );
  if ( !v3u64Eq( request, bytes ) )
  {
    request = v3u64Add( request, align );
  }

#if ( V3_MEM_SIZE == V3_MEM_64BITS )
  return( realloc( memory, request ) );
#else /* 32bits */
  if ( v3u64Gt( bytes, v3u64u32( (u32) 0xFFFFFFFF ) ) )
  {
    return( NULL );
  }
  return( realloc( memory, v3u32u64( request ) ) );
#endif
}

s32 v3MemCopy( void * dest, const void * src, u64 length )
{
  /* Fail if either dest or src is NULL */
  if ( ( dest == NULL ) || ( src == NULL ) )
  {
    return( V3_FAIL );
  }

  /* Do nothing if dest == src, or length is zero */
  if ( ( dest == src ) || ( v3u64Eq( length, v3u64u32( 0 ) ) ) )
  {
    return( V3_PASS );
  }

#if ( V3_MEM_SIZE == V3_MEM_64BITS )
  memmove( dest, src, length );
#else /* 32bits */
  if ( v3u64Gt( length, v3u64u32( (u32) 0xFFFFFFFF ) ) )
  {
    return( V3_FAIL );
  }
#if ( OS_TYPE == OS_SUNOS )
  bcopy( src, dest, v3u32u64( length ) );
#else
  memmove( dest, src, v3u32u64( length ) );
#endif
#endif

  return( V3_PASS );
}

s32 v3MemSet( void * memory, u64 length, u8 value )
{
#if ( V3_MEM_SIZE == V3_MEM_64BITS )
  memset( memory, value, length );
#else /* 32bits */
  if ( v3u64Gt( length, v3u64u32( (u32) 0xFFFFFFFF ) ) )
  {
    return( V3_FAIL );
  }
  memset( memory, value, v3u32u64( length ) );
#endif

  return( V3_PASS );
}

s32 v3MemCmp( const void * blockA, const void * blockB, u64 max_bytes )
{
  u64 i;
  s32 a, b;
  u8 * pA, * pB;

  if ( ( blockA == NULL ) || ( blockB == NULL )
    || ( v3u64Eq( max_bytes, v3u64u32( 0 ) ) ) )
  {
    if ( blockA != blockB )
    {
      return( -1 );
    }

    return( 0 );
  }

  pA = (u8 *) blockA;
  pB = (u8 *) blockB;
  _V3_SET64( i, 00000000, 00000000 );
  do
  {
    a = (s32) *(pA)++;
    b = (s32) *(pB)++;
    v3u64Inc( &i );
  } while ( ( a == b ) && v3u64Lt( i, max_bytes ) );

  return( a - b );
}

void * v3MemOffset( const void * memory, u64 offset )
{
  u8 * tmp;

  /*
    We let the compiler do the pointer math correctly, but then check for
    wrapping the only way we can.
  */
  if ( memory == NULL )
  {
    return( NULL ) ;
  }

  tmp = (u8 *) memory;

#if ( V3_MEM_SIZE == V3_MEM_64BITS )
  tmp = &tmp[offset];
  if ( (u64) tmp < (u64) memory )
  {
    return( NULL ); /* wrap */
  }
#else /* 32bit */
  if ( v3u64Gt( offset, v3u64u32( (u32) 0xFFFFFFFF ) ) )
  {
    return( NULL ) ;
  }
  tmp = &tmp[v3u32u64( offset )];
  if ( (void *) tmp < (void *) memory )
  {
    return( NULL ); /* wrap */
  }
#endif

  return( (void *) tmp );
}

void v3MemFree( void * memory )
{
  if ( memory != NULL )
  {
    free( memory );
  }
}

s32 v3MemSystem( u64 * amount )
{
  /*
    We must find out how much as a normal user and without any 'tricks'
    like reading /proc or /dev.
  */
#if ( OS_TYPE == OS_WIN32 )
  /* this has some major limitations (2GB limit) */
  MEMORYSTATUS info;

  info.dwLength = sizeof( MEMORYSTATUS );
  GlobalMemoryStatus( &info );
  *amount = v3u64u32( (u32) info.dwTotalPhys );

  return( V3_PASS );
#endif

  return( V3_FAIL );
}


void v3MemWarning( void )
{
  v3PrintA( (ascii *) "Secure memory requested, but not available.\n" );
}

/* testing */

s32 v3_TestCosmMem( void )
{
  u8 * ptr1, * ptr2;
  u8 * mem;
  u8 * offset;
  u32 i;
  u64 size;

  /* First be sure our V3_MEM_SIZE is correct - set at the top of cosmmem.c */
#if ( V3_MEM_SIZE == V3_MEM_64BITS )
  if ( sizeof( void * ) != 8 )
  {
    return( -1 );
  }
#else /* the 32bit case */
  if ( sizeof( void * ) != 4 )
  {
    return( -1 );
  }
#endif

  /* Next be sure our memory size matches the calloc size_t */
#if ( V3_MEM_SIZE == V3_MEM_64BITS )
  if ( sizeof( size_t ) != sizeof( u64 ) )
  {
    return( -2 );
  }
#else /* the 32bit case */
  if ( sizeof( size_t ) != sizeof( u32 ) )
  {
    return( -2 );
  }
#endif

  /*
    We simply try to allocate more memory then the machine
    can possibly be able to give up.
  */
#if ( V3_MEM_SIZE == V3_MEM_64BITS )
  _V3_SET64( size, FFFFFFFF, FFFFFFFF );
#else /* the 32bit case */
  _V3_SET64( size, 00000001, 00000000 );
#endif
  if ( ( mem = (u8 *) v3MemAlloc( size, V3_MEM_NORMAL ) ) != NULL )
  {
    v3MemFree( mem );
    return( -3 );
  }

  /* allocate 0 bytes */
  _V3_SET64( size, 00000000, 00000000 );
  if ( ( mem = (u8 *) v3MemAlloc( size, V3_MEM_NORMAL ) ) != NULL )
  {
    v3MemFree( mem );
    return( -4 );
  }
  /* bad secure flag */
  _V3_SET64( size, 00000000, 00000020 );
  if ( ( mem = (u8 *) v3MemAlloc( size, 0xFFFF ) ) != NULL )
  {
    v3MemFree( mem );
    return( -5 );
  }

  /* test v3MemAlloc */
  _V3_SET64( size, 00000000, 00000020 );
  ptr1 = (u8 *) v3MemAlloc( size, V3_MEM_NORMAL );
  ptr2 = (u8 *) v3MemAlloc( size, V3_MEM_NORMAL );
  if ( ( ptr1 == NULL ) || ( ptr2 == NULL ) )
  {
    v3MemFree( ptr1 );
    v3MemFree( ptr2 );
    return( -6 );
  }

  /* is it zeroed? */
  for ( i = 0 ; i < 0x20 ; i++ )
  {
    if ( ( ptr1[i] != 0 ) || ( ptr2[i] != 0 ) )
    {
      v3MemFree( ptr1 );
      v3MemFree( ptr2 );
      return( -7 );
    }
  }

  /* Put the number 0 .. 31 in  ptr1. */
  for ( i = 0 ; i < 0x20 ; i++ )
  {
    ptr1[i] = (u8) i;
  }

  /* Test v3MemCopy, copy a bit of ptr1 to ptr2 */
  if ( v3MemCopy( ptr2, ptr1, size ) == V3_FAIL )
  {
    v3MemFree( ptr1 );
    v3MemFree( ptr2 );
    return( -8 );
  }
  for ( i = 0 ; i < 0x20 ; i++ )
  {
    if ( ptr1[i] != ptr2[i] )
    {
      v3MemFree( ptr1 );
      v3MemFree( ptr2 );
      return( -9 );
    }
  }

  /* Test v3MemRealloc */
  ptr1 = (u8 *) v3MemRealloc( ptr1, v3u64u32( (u32) 0x40 ), V3_MEM_NORMAL );
  if ( ptr1 == NULL )
  {
    v3MemFree( ptr1 );
    v3MemFree( ptr2 );
    return( -10 );
  }
  ptr1 = (u8 *) v3MemRealloc( ptr1, v3u64u32( (u32) 0x10 ), V3_MEM_NORMAL );
  if ( ptr1 == NULL )
  {
    v3MemFree( ptr1 );
    v3MemFree( ptr2 );
    return( -11 );
  }
  ptr1 = (u8 *) v3MemRealloc( ptr1, v3u64u32( (u32) 0x20 ), V3_MEM_NORMAL );
  if ( ptr1 == NULL )
  {
    v3MemFree( ptr1 );
    v3MemFree( ptr2 );
    return( -12 );
  }
  /* Only first 16 bytes must remain, last 16 can be lost in realloc */
  for ( i = 0 ; i < 0x10 ; i++ )
  {
    if ( ptr1[i] != ptr2[i] )
    {
      v3MemFree( ptr1 );
      v3MemFree( ptr2 );
      return( -13 );
    }
  }

  /* Test v3MemSet */

  if ( v3MemSet( ptr1, v3u64u32( (u32) 0x08 ), (u8) 0xE7 ) != V3_PASS )
  {
    v3MemFree( ptr1 );
    v3MemFree( ptr2 );
    return( -14 );
  }
  /* Check first 8 bytes == 0xE7 */
  for ( i = 0 ; i < 0x08 ; i++ )
  {
    if ( ptr1[i] != (u8) 0xE7 )
    {
      v3MemFree( ptr1 );
      v3MemFree( ptr2 );
      return( -15 );
    }
  }
  /* Check last 8 bytes are the same */
  for ( i = 0x08 ; i < 0x10 ; i++ )
  {
    if ( ptr1[i] != ptr2[i] )
    {
      v3MemFree( ptr1 );
      v3MemFree( ptr2 );
      return( -16 );
    }
  }

  /* test v3MemOffset */
  _V3_SET64( size, 00000000, 0000000F );
  offset = (u8 *) v3MemOffset( ptr2, size );
  if ( *offset != 0x0F )
  {
    v3MemFree( ptr1 );
    v3MemFree( ptr2 );
    return( -17 );
  }
  _V3_SET64( size, 00000000, 00000014 );
  offset = (u8 *) v3MemOffset( ptr2, size );
  if ( *offset != 0x14 )
  {
    v3MemFree( ptr1 );
    v3MemFree( ptr2 );
    return( -18 );
  }

  /* test v3MemCmp */
  _V3_SET64( size, 00000000, 00000010 );
  for ( i = 0 ; i < 0x10 ; i++ )
  {
    ptr1[i] = (u8) i;
    ptr2[i] = (u8) i;
  }
  if ( v3MemCmp( ptr1, ptr2, size ) != 0 )
  {
    v3MemFree( ptr1 );
    v3MemFree( ptr2 );
    return( -19 );
  }
  ptr1[6] = (u8) 42;
  if ( v3MemCmp( ptr1, ptr2, size ) == 0 )
  {
    v3MemFree( ptr1 );
    v3MemFree( ptr2 );
    return( -19 );
  }

  /* No way to test v3MemFree */
  v3MemFree( ptr1 );
  v3MemFree( ptr2 );

  return( V3_PASS );
}
