/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

/* CPU/OS Layer - CPU and OS specific code is allowed */

#ifndef V3_NET_H
#define V3_NET_H

#include "cputypes.h"
#include "cosmtime.h"
#include "cosmtask.h"

/*
These port defines need to be moved out of the cpu/os layer eventually
cosm-client     4700/tcp  Cosm Client
cosm-client     4700/udp  Cosm Client
cosm-proxy      4701/tcp  Cosm Proxy
cosm-proxy      4701/udp  Cosm Proxy
cosm-server     4702/tcp  Cosm Server
cosm-server     4702/udp  Cosm Server
cosm-console    4703/tcp  Cosm Console
cosm-console    4703/udp  Cosm Console
#                         Adam L. Beberg <beberg@mithral.com>
*/
#define V3_NET_PORT_CLIENT   4700
#define V3_NET_PORT_PROXY    4701
#define V3_NET_PORT_SERVER   4702

#define V3_NET_STATUS_CLOSED   0
#define V3_NET_STATUS_OPEN     1
#define V3_NET_STATUS_LISTEN   2

#define V3_NET_MODE_NONE  0
#define V3_NET_MODE_TCP   23
#define V3_NET_MODE_UDP   32

#define V3_NET_ERROR_NONE      0
#define V3_NET_ERROR_HOST      -1  /* Unable to connect to host */
#define V3_NET_ERROR_PORT      -2  /* Unable to connect/listen on port */
#define V3_NET_ERROR_FIREHOST  -3  /* Invalid firewall host */
#define V3_NET_ERROR_FIREPORT  -4  /* Invalid firewall port */
#define V3_NET_ERROR_FIREPASS  -5  /* Bad firewall Password */
#define V3_NET_ERROR_CLOSED    -6  /* Connection Closed */
#define V3_NET_ERROR_WAIT      -7  /* Wait time exceded in Recv */
#define V3_NET_ERROR_MODE      -8  /* TCP/UDP mode conflict */
#define V3_NET_ERROR_NO_NET    -9  /* No Networking Support */

#define V3_NET_ACCEPT_NOWAIT  0
#define V3_NET_ACCEPT_WAIT    1

typedef struct
{
  u64 handle;
  u32 status;
  u32 host;
  u32 my_host;
  u16 port;
  u16 my_port;
  u32 mode;
  s32 last_error;
  ascii fire_pass[64];
  v3_MUTEX lock;
  u32 fire_host;
  u16 fire_port;
} v3_NET;

#define V3_NET_MAX_HOSTNAME   255

typedef ascii v3_NET_HOSTNAME[V3_NET_MAX_HOSTNAME + 1];

#define V3_NET_ALLOW  1
#define V3_NET_DENY   0

typedef struct
{
  v3_time expires;
  u32 ip;
  u32 mask;
  u32 permission;
  u32 null;  /* for memory alignment */
} v3_NET_ACL_ENTRY;

typedef struct
{
  u32 length;
  u32 count;
  v3_NET_ACL_ENTRY * list;
} v3_NET_ACL;

/* Network send and receive functions */

s32 v3NetOpen( v3_NET * net, u32 host, u16 port, u32 my_host, u16 my_port,
  u32 mode, u32 firehost, u16 fireport, ascii * firepass );
  /*
    Open a network connection to the remote host and port. my_host and
    my_port are the host/port you attempt to connect from, 0 for either
    means the OS picks. If mode is V3_NET_MODE_TCP, TCP protocol is used.
    If mode is V3_NET_MODE_UDP, then the UDP protocol is used.
    firehost, fireport, firepass Are for proxies that need to go through
    a firewall. Use semi-colon to separate username from password for
    socks5 firewall. ONLY proxies will ever cross firewalls.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

#define _V3_NETOPEN( net, host, port ) \
  v3NetOpen( net, host, port, 0, 0, V3_NET_MODE_TCP, 0, 0, NULL )
  /*
    Since v3NetOpen is rather complex, a macro for simpler cases is also
    defined. _V3_NETOPEN(...) allows just the first 3 parameters to be
    specified for when you do not need to worry about firewalls and just
    need a TCP connection.
  */

s32 v3NetSend( v3_NET * net, const void * data, u32 length );
  /*
    Send the data over the network connection. Returns once data is sent.
    Connection must be opened/accepted in V3_NET_MODE_TCP mode.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

u32 v3NetRecv( void * buffer, u32 length, v3_NET * net, u32 wait_ms );
  /*
    Read whatever data is available, up to length bytes, into the buffer.
    If wait_ms is non-zero, then delay up to wait_ms milliseconds until
    length bytes have been received.
    Connection must be opened/accepted in V3_NET_MODE_TCP mode.
    Returns: 0 or < length if an error - set last_error, length on success.
  */

s32 v3NetSendUDP( v3_NET * net, const void * data, u32 length );
  /*
    Send the data over the network connection.
    Connection must be opened/listening in V3_NET_MODE_UDP mode.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

u32 v3NetRecvUDP( void * buffer, u32 length, v3_NET * net, v3_NET_ACL * acl,
  u32 wait_ms );
  /*
    Read whatever data is available, up to length bytes, into the buffer.
    If wait_ms is non-zero, then delay up to wait_ms milliseconds until length
    bytes have been received. Connection must be opened/listening in
    V3_NET_MODE_UDP mode. Any data from a host that fails the acl masks
    will be discarded.
    Returns: 0 or < length if an error - set last_error, length on success.
  */

s32 v3NetListen( v3_NET * net, u32 ip, u16 port, u32 mode, u32 queue );
  /*
    Set up a listening point on port that will accept connections.
    Listen on ip if possible, if ip is zero listen on all interfaces.
    If mode is V3_NET_MODE_TCP, TCP protocol is used. If mode is
    V3_NET_MODE_UDP, then the UDP protocol is used.
    If port is zero, let OS pick one. Allow queue connections to be
    queued before rejecting connections.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

s32 v3NetAccept( v3_NET * new_connection, v3_NET * net, v3_NET_ACL * acl,
  u32 wait );
  /*
    Accept a network connection if one is waiting, and fill in all the fields
    of new_connection. If wait is V3_NET_ACCEPT_WAIT then do not return
    until a connection is accepted, otherwise if wait is V3_NET_ACCEPT_NOWAIT
    return immediately. Any host that fails the acl masks will be ignored.
    Returns: V3_PASS on success, or set last_error of new_connection and
      return an error code.
  */

s32 v3NetClose( v3_NET * net );
  /*
    Close the Network connection and clear out any remaining data.
    Returns: V3_PASS on success, or set last_error and return an error code.
  */

u32 v3NetDNS( u32 * ip, u32 length, ascii * name );
  /*
    Perform DNS lookup. Sets up to length ip's to the IP of the name'd host.
    Returns: Number of IP's set, 0 indicates failure.
  */

s32 v3NetRevDNS( v3_NET_HOSTNAME * name, u32 ip );
  /*
    Perform reverse DNS. Sets name to the hostname of the ip.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s32 v3NetMyIP( u32 * ip );
  /*
    Sets ip to the hosts IP.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s32 v3NetACLAdd( v3_NET_ACL * acl, u32 ip, u32 mask, u32 permission,
  v3_time expires );
  /*
    Add the ip/mask pair to the acl. perm should be either V3_NET_ALLOW or
    V3_NET_DENY. expires is the time after which the entry will be deleted
    automatically, and should be based off of v3SystemClock time.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s32 v3NetACLDelete( v3_NET_ACL * acl, u32 ip, u32 mask );
  /*
    Delete the ip/mask entry from the acl.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s32 v3NetACLTest( v3_NET_ACL * acl, u32 ip );
  /*
    Test if an ip is accepted by the acl. An ip is accepted if it passed
    through each of the acl entries in order and has a V3_NET_ALLOW setting
    at the end. The default is to allow i.e. a NULL acl.
    Returns: V3_NET_ALLOW if accepted, or V3_NET_DENY if rejected.
  */

void v3NetACLFree( v3_NET_ACL * acl );
  /*
    Free the internal acl data.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s32 v3_NetGlobalInit( void );
  /*
    Initialize any global networking states that an OS needs.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

/* testing */

s32 v3_TestCosmNet( void );
  /*
    Test functions in this header.
    Returns: V3_PASS on success, or a negative number corresponding to the
      test that failed.
  */

#endif
