/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

/* CPU/OS Layer - CPU and OS specific code is allowed */

#include "cosmtask.h"
#include "cosmfile.h"
#include "cosmio.h"
#include "cosmmem.h"

#include <errno.h>
#if ( OS_TYPE == OS_MACOS )
#include <Processes.h>
#include <Threads.h>
#include <TextUtils.h>
#include <AppleEvents.h>
#include <MultiProcessing.h>
#include <Gestalt.h>
#elif ( OS_TYPE == OS_WIN32 )
#include <stddef.h>
#include <signal.h>
#include <process.h>
#include <time.h>
#include <dos.h>
static u32 __V3_WIN32_SIGNALS_INIT = 0;
static HANDLE __V3_WIN32_SIGINT = NULL;
static HANDLE __V3_WIN32_SIGTERM = NULL;
void v3_SignalWaitThread( void * arg );
#elif ( OS_TYPE == OS_OS2 )
#include <stddef.h>
#include <signal.h>
#include <process.h>
#include <time.h>
#define INCL_DOSERRORS
#define INCL_DOSPROCESS
#define INCL_DOSDATETIME
#include <os2.h>
#elif ( OS_TYPE == OS_BEOS )
#include <OS.h>
#include <signal.h>
#include <unistd.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <stdlib.h>
#elif ( OS_TYPE == OS_SUNOS )
#include <signal.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <stdlib.h>
#include <unistd.h>
#if ( !defined( NO_THREADS ) )
#include <lwp/lwp.h>
#include <lwp/lwpmachdep.h>
#include <lwp/stackdep.h>
#endif
#else /* OS */
#include <sched.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <stdlib.h>
#endif /* OS */

#if ( OS_TYPE != OS_MACOS )
#include <sys/types.h>
#endif

#if ( OS_TYPE == OS_FREEBSD )
#include <sys/sysctl.h>
#endif

#if ( ( OS_TYPE == OS_IRIX ) || ( OS_TYPE == OS_IRIX64 ) )
#include <sys/sysmp.h>
#endif

#if ( OS_TYPE == OS_SOLARIS )
#include <sys/types.h>
#include <sys/processor.h>
#include <sys/procset.h>
#endif

/* Setup the size of Process ID */
#if ( 0 )
#define V3_TASKID_64
#else
#undef V3_TASKID_64
#endif

/* Solaris lacks PRIO_MAX and PRIO_MIN. */
#if ( OS_TYPE == OS_SOLARIS )
#define PRIO_MAX  20
#define PRIO_MIN  -20
#endif

u64 v3ProcessID( void )
{
#if ( OS_TYPE == OS_MACOS )
  OSErr my_err;
  ProcessSerialNumber my_psn;

  my_err = GetCurrentProcess(&my_psn);
  if (my_err == noErr)
  {
    u64 my_pid;
#if ( defined( CPU_64BIT ) )
    my_pid = ( my_psn.highLongOfPSN << 32 ) + my_psn.lowLongOfPSN;
#else
    my_pid.hi = my_psn.highLongOfPSN;
    my_pid.lo = my_psn.lowLongOfPSN;
#endif
    return( my_pid );
  }
#elif ( OS_TYPE == OS_WIN32 )
  return( v3u64u32( (u32) GetCurrentProcessId() ) );
#else /* unix... */

#if ( defined( V3_TASKID_64 ) )
  return( (u64) getpid() );
#else
  return( v3u64u32( (u32) getpid() ) );
#endif /* 64 bit */

#endif /* OS */
}

u8 v3ProcessPriority( u8 priority )
{
#if ( OS_TYPE == OS_OS2 )
  /* current process not distinguished from current thread */
  return( v3ThreadPriority( priority ) );
#elif ( ( OS_TYPE == OS_MACOS ) || ( OS_TYPE == OS_BEOS ) )
  return( 0 );
#elif ( OS_TYPE == OS_WIN32 )
  /* same code as for threads */
  return( v3ThreadPriority( priority ) );
#else /* UNIX */
  s32 get_pri;

  if ( priority != 0 )
  {
    if ( setpriority( PRIO_PROCESS, 0, v3_PriorityToOS( priority ) ) != 0 )
    {
      return( 0 );
    }
  }

  errno = 0;
  get_pri = getpriority( PRIO_PROCESS, 0 );

  if ( errno == 0 )
  {
    return( v3_PriorityToCosm( get_pri ) );
  }
  else
  {
    return( 0 );
  }
#endif
}

u64 v3ProcessSpawn( const ascii * command, const ascii * arguments, ... )
{
  const ascii * arg_array[1024];
  v3_FILENAME filename;
  va_list args;
  u32 i;
  int new_pid;

#if ( OS_TYPE == OS_WIN32 )
  ascii command_line[1024];
  STARTUPINFO startup_info;
  PROCESS_INFORMATION process_info;
  u32 j;
  u32 length;
  u32 total;
  u32 done;
#elif ( OS_TYPE == OS_MACOS )
  OSErr my_err;
  FSSpec spec;
  u64 my_pid;
#endif

  if ( v3_FileNativePath( filename, command ) == V3_FAIL )
  {
    return( v3u64u32( (u32) 0 ) );
  }

  arg_array[0] = (const ascii *) filename;
  arg_array[1] = arguments;

  va_start( args, arguments );
  i = 1;
  while ( ( arg_array[i] != NULL ) && ( i < 1023 ) )
  {
    arg_array[++i] = va_arg( args, const ascii * );
  }
  arg_array[1023] = NULL;
  va_end( args );

#if ( OS_TYPE == OS_WIN32 )
  /* put together a command line, win32 doesn't have a working spawnv() */
  total = 0;
  j = 0;
  done = 0;
  while ( ( j < i ) && ( done == 0 ) )
  {
    length = v3u32u64( v3StrLengthA( arg_array[j] ) );
    if ( length > ( 1022 - total ) )
    {
      done = 1;
    }
    else
    {
      if ( j > 0 )
      {
        command_line[total++] = ' ';
      }
      v3StrCopyA( &command_line[total], arg_array[j],
        v3u64u32( 1024 - total ) );
      total = total + length;
      j++;
    }
  }
  v3MemSet( &startup_info, v3u64u32( sizeof( STARTUPINFO ) ), (u8) 0 );
  startup_info.cb = sizeof( STARTUPINFO );
  if ( CreateProcess( filename, command_line, NULL, NULL, 0, 0, NULL, NULL,
    &startup_info, &process_info ) == 0 )
  {
    new_pid = 0;
  }
  else
  {
    /* process will not go away until all the handles are closed */
    CloseHandle( process_info.hProcess );
    CloseHandle( process_info.hThread );
    new_pid = process_info.dwProcessId;
  }
#elif ( OS_TYPE == OS_OS2 )
  new_pid = spawnv( P_NOWAIT, (char *) arg_array[0], (char * *) arg_array );
#elif ( OS_TYPE == OS_MACOS )
  /* First convert the path to a pascal string */
  Str255 path = "\p";
  c2pstrcpy( path, (const char *)command );
  /* Create a filespec. We could also use __path2fss here. See MSL C p.137 */
  my_err = MakeFSSpec( 0, 0, path, spec );
  if ( my_err == noErr )
  {
    /* Launch */
    LaunchParamBlockRec launch_params;
    /* extendedBlock and extendedBlockLen are defined in Processes.h */
    launch_params.launchBlockID = extendedBlock;
    launch_params.launchEPBLength = extendedBlockLen;
    launch_params.launchFileFlags = 0;
    launch_params.launchControlFlags = launchContinue + launchNoFileFlags;
    launch_params.launchAppSpec = &spec;
    launch_params.launchAppParameters = nil;

    my_err = LaunchApplication( &launch_params );
    if ( my_err == noErr )
    {
      /* convert the psn to u64 */
#if ( defined( CPU_64BIT ) )
      my_pid = ( launch_params.launchProcessSN.highLongOfPSN << 32 ) +
               ( launch_params.launchProcessSN.lowLongOfPSN );
#else
      my_pid.hi = launch_params.launchProcessSN.highLongOfPSN;
      my_pid.lo = launch_params.launchProcessSN.lowLongOfPSN;
#endif
      return( my_pid );
    }
    else
    {
      return( v3u64u32( (u32) 0 ) );
    }
  }
  else
  {
    return( v3u64u32( (u32) 0 ) );
  }
#else /* unix... */
#if ( defined( SIGCHLD ) )
  /* SIGCHLD has to be ignored or zombies remain and confuse kill() */
  signal( SIGCHLD, SIG_IGN );
#endif
  new_pid = fork();
  if ( new_pid == -1 ) /* check for failure */
  {
    return( v3u64u32( (u32) 0 ) );
  }
  if ( new_pid == 0 ) /* are we in the child or parent */
  {
    /* in child */
    if ( execv( (char *) arg_array[0], (char * *) arg_array ) == -1 )
    {
      /* since we are in the child, this is an unrecoverable error. */
      exit( 0 );
    }
  }
#endif

#if ( defined( V3_TASKID_64 ) )
  return( (u64) new_pid );
#else
  return( v3u64u32( (u32) new_pid ) );
#endif
}

s32 v3ProcessSignal( u64 process_id, u32 signal_type )
{
#if ( OS_TYPE == OS_WIN32 )
  HANDLE process;
  ascii name[128];
#elif ( OS_TYPE == OS_MACOS )
  OSErr my_err;
  ProcessSerialNumber my_psn;
  AEAddressDesc target_address;
#endif

  if ( signal_type == V3_PROCESS_SIGPING )
  {
#if ( OS_TYPE == OS_MACOS )
    /* !!! */
#elif ( OS_TYPE == OS_WIN32 )
    if ( ( process = OpenProcess( PROCESS_ALL_ACCESS, 0,
      v3u32u64( process_id ) ) ) == NULL )
    {
      return( V3_FAIL );
    }
    else
    {
      CloseHandle( process );
      return( V3_PASS );
    }
#elif ( OS_TYPE == OS_OS2 )
    return( V3_FAIL ); /* no interprocess signals in OS/2 */
#else /* OS */

#if ( defined( V3_TASKID_64 ) )
    if ( kill( process_id, 0 ) == -1 )
#else
    if ( kill( v3u32u64( process_id ), 0 ) == -1 )
#endif
    {
      if ( errno == ESRCH )
      {
        return( V3_FAIL );
      }
    }
#endif /* OS */
  }
  else
  {
#if ( OS_TYPE == OS_MACOS )
    /* create a psn from the process_id */
#if ( defined( CPU_64BIT ) )
    my_psn.highLongOfPSN = ( process_id >> 32 );
    my_psn.highLongOfPSN = (u32) process_id;
#else
    my_psn.highLongOfPSN = process_id.hi;
    my_psn.lowLongOfPSN = process_id.lo;
#endif
    my_err = AECreateDesc( typeProcessSerialNumber, &my_psn,
      sizeof( my_psn ), &target_address );
    if ( my_err == noErr )
    {
      AppleEvent the_apple_event;
      switch ( signal_type )
      {
        case V3_PROCESS_SIGINT:  /* Save/CheckPoint */
          my_err = AECreateAppleEvent( 'COSM', 'SAVE', &target_address,
            kAutoGenerateReturnID, kAnyTransactionID, &the_apple_event );
          break;
        case V3_PROCESS_SIGTERM:  /* Save & Terminate */
          my_err = AECreateAppleEvent( 'COSM', 'SADI', &target_address,
            kAutoGenerateReturnID, kAnyTransactionID, &the_apple_event );
          break;
        case V3_PROCESS_SIGKILL:  /* Terminate with extreme prejudice */
          my_err = AECreateAppleEvent( 'COSM', 'DIE ', &target_address,
            kAutoGenerateReturnID, kAnyTransactionID, &the_apple_event );
          break;
        default:
          return( V3_FAIL );
      }

      if ( my_err == noErr )
      {
        my_err = AESend( &the_apple_event, nil, kAENoReply, kAENormalPriority,
          0, nil, nil );
        if ( my_err == noErr )
        {
          return( V3_PASS );
        }
      }
    }
    return( V3_FAIL );
#elif ( OS_TYPE == OS_WIN32 )
    if ( v3u64Eq( process_id, v3ProcessID() ) )
    {
      /* process is this one */
      switch ( signal_type )
      {
        case V3_PROCESS_SIGINT:
          raise( SIGTERM );
          break;
        case V3_PROCESS_SIGTERM:
          raise( SIGINT );
          break;
        case V3_PROCESS_SIGKILL:
          v3ProcessEnd( -1 );
          break;
        default:
          return( V3_FAIL );
          break;
      }
      return( V3_PASS );
    }
    /* open the correct Semaphore, and trigger it */
    switch ( signal_type )
    {
      case V3_PROCESS_SIGINT:
        v3PrintAStr( name, v3u64u32( 128 ), (ascii *) "%016Y-SIGINT",
          process_id );
        if ( ( process = OpenSemaphore( SEMAPHORE_ALL_ACCESS, 0, name ) )
          == NULL )
        {
          return( V3_FAIL );
        }
        if ( ReleaseSemaphore( process, 1, NULL ) == 0 )
        {
          CloseHandle( process );
          return( V3_FAIL );
        }
        else
        {
          CloseHandle( process );
          return( V3_PASS );
        }
        break;
      case V3_PROCESS_SIGTERM:
        v3PrintAStr( name, v3u64u32( 128 ), (ascii *) "%016Y-SIGTERM",
          process_id );
        if ( ( process = OpenSemaphore( SEMAPHORE_ALL_ACCESS, 0, name ) )
          == NULL )
        {
          return( V3_FAIL );
        }
        if ( ReleaseSemaphore( process, 1, NULL ) == 0 )
        {
          CloseHandle( process );
          return( V3_FAIL );
        }
        else
        {
          CloseHandle( process );
          return( V3_PASS );
        }
        break;
      case V3_PROCESS_SIGKILL:
        /* slay the process, this is not good for the system */
        if ( ( process = OpenProcess( PROCESS_ALL_ACCESS, 0,
          v3u32u64( process_id ) ) ) == NULL )
        {
          return( V3_FAIL );
        }
        if ( TerminateProcess( process, 1 ) == 0 )
        {
          CloseHandle( process );
          return( V3_FAIL );
        }
        else
        {
          CloseHandle( process );
          return( V3_PASS );
        }
        break;
      default:
        return( V3_FAIL );
    }
    return( V3_FAIL );
#elif ( OS_TYPE == OS_OS2 )
/* !!! */
    raise( signal_type );
#else /* OS */
    switch ( signal_type )
    {
      case V3_PROCESS_SIGINT:
        signal_type = SIGINT;
        break;
      case V3_PROCESS_SIGTERM:
        signal_type = SIGTERM;
        break;
      case V3_PROCESS_SIGKILL:
        signal_type = SIGKILL;
        break;
      default:
        return( V3_FAIL );
    }
#if ( defined( V3_TASKID_64 ) )
    if ( kill( (pid_t) process_id, (int) signal_type ) == -1 )
#else
    if ( kill( (pid_t) v3u32u64( process_id ), (int) signal_type ) == -1 )
#endif
    {
      return( V3_FAIL );
    }
#endif /* OS */
  }
  return( V3_PASS );
}

void v3ProcessEnd( int status )
{
#if ( OS_TYPE == OS_MACOS )
  v3ProcessSignal( v3ProcessID(), V3_PROCESS_SIGKILL );
#else
  exit( status );
#endif
}

u32 v3CPUCount( void )
{
  u32 ncpus;

  /*
    There must be one CPU, but we must find out how many as a normal user
    and without any 'tricks' like reading /proc or /dev.
  */

#if ( OS_TYPE == OS_MACOS )
  if ( MPLibraryIsLoaded() )
  {
    ncpus = MPProcessorsScheduled();
  }
#elif ( OS_TYPE == OS_BEOS )
  system_info Sysinfo;

  get_system_info( &Sysinfo );
  ncpus = Sysinfo.cpu_count;
#elif ( OS_TYPE == OS_WIN32 )
  SYSTEM_INFO CPUInfo;

  GetSystemInfo( & CPUInfo );
  if ( CPUInfo.dwNumberOfProcessors < 1 )
  {
    ncpus = 1;
  }
  else
  {
    ncpus = CPUInfo.dwNumberOfProcessors;
  }
#elif ( OS_TYPE == OS_LINUX )
  /* can't do it without wierd-ass tricks */
  ncpus = 1;
#elif ( OS_TYPE == OS_FREEBSD )
  int mib[2], ncpu;
  size_t len;

  mib[0] = CTL_HW;
  mib[1] = HW_NCPU;
  len = sizeof( ncpu );
  sysctl( mib, 2, &ncpu, &len, NULL, 0 );
  ncpus = ncpu;
#elif ( ( OS_TYPE == OS_IRIX ) || ( OS_TYPE == OS_IRIX64 ) )
  ncpus = sysmp( MP_NPROCS );
#elif ( ( OS_TYPE == OS_TRU64 ) || ( OS_TYPE == OS_SOLARIS ) )
  /* _SC_NPROCESSORS_CONF == number of physical processors */
  /* _SC_NPROCESSORS_ONLN == number of ACTIVE processors   */
  ncpus = sysconf( _SC_NPROCESSORS_ONLN );
#elif ( OS_TYPE == OS_SUNOS )
  /* SunOS does have multiple CPU capabilities, but no sysconf()
     interface to access it. */
  ncpus = 1;
#else /* OS = other */
  ncpus = 1;
#endif /* OS */

  return( ncpus );
}

s32 v3CPUGet( u32 * cpu )
{
  /* Note: On multi-CPU systems, each thread could concurrently be on a CPU */

#if ( OS_TYPE == OS_MACOS )
  return( V3_FAIL );
#elif ( OS_TYPE == OS_WIN32 )
  /* can get the processor affinity mask but not the processor we're on */
  return( V3_FAIL );
#elif ( OS_TYPE == OS_SOLARIS )
  /* (cramer) TODO: insert code here. */
  return( V3_FAIL );
#else /* OS */
  return( V3_FAIL );
#endif
}

s32 v3CPULock( u64 process_id, u32 cpu )
{
#if ( OS_TYPE == OS_WIN32 )
  /* Can't determine which process we are running on */
  return( V3_FAIL );
#elif ( OS_TYPE == OS_SOLARIS )
  int result;
  result = processor_bind( P_PID, (pid_t) v3u32u64( process_id ),
    (processorid_t) cpu, NULL );
  if ( result == 0 )
  {
    return( V3_PASS );
  }
  return( V3_FAIL );
#else /* OS */
  return( V3_FAIL );
#endif
}

s32 v3CPUUnlock( u64 process_id )
{
#if ( OS_TYPE == OS_WIN32 )
  /* v3CPULock always fails so this function will as well... */
  return( V3_FAIL );
#elif ( OS_TYPE == OS_SOLARIS )
  int result;
  result = processor_bind( P_PID, (pid_t) v3u32u64( process_id ),
    PBIND_NONE, NULL );
  if ( result == 0 )
  {
    return( V3_PASS );
  }
  return( V3_FAIL );
#else /* OS */
  return( V3_FAIL );
#endif
}

u64 v3ThreadBegin( void (*start)(void *), void * arg, u32 stack_size )
{
  u64 result;

#if ( defined( NO_THREADS ) )
  /* set result to fail by default */
  _V3_SET64( result, FFFFFFFF, FFFFFFFF );
  return( result );
#elif ( OS_TYPE == OS_MACOS )
  OSErr my_err;
  long response;
  short comparebit = gestaltThreadMgrPresent;
  my_err = Gestalt( gestaltThreadMgrAttr, &response );
  if ( my_err == noErr )
  {
    if ( BitTst( &response, 31 - comparebit ) == true )
    {
      /* got thread manager */
      ThreadID thread_made;
      my_err = NewThread( kCooperativeThread,
        (pascal void * (*)(void *))start, arg, stack_size, kCreateIfNeeded,
        nil, &thread_made );
      if ( my_err == noErr )
      {
        return( v3u64u32( thread_made ) );
      }
    }
  }
  return( result );
#elif ( OS_TYPE == OS_BEOS )
  thread_id new_thread;
  new_thread = spawn_thread( (thread_func) start, "Cosm Thread",
    B_NORMAL_PRIORITY, (void *) arg );
  switch( new_thread )
  {
    case B_NO_MORE_THREADS:
    case B_NO_MEMORY:
      _V3_SET64( result, FFFFFFFF, FFFFFFFF );
      break;
    default:
      /* Technicall, resume_thread could return !B_OK */
      resume_thread( new_thread );
      result = v3u64u32( new_thread );
      break;
  }
  return( result );
#elif ( OS_TYPE == OS_WIN32 )
  u32 thread_id;
  HANDLE thandle;

  thandle = (HANDLE) _beginthreadex( NULL, stack_size,
    (unsigned int (__stdcall *)(void *)) start, arg, 0, &thread_id );
  if ( thandle == 0 )
  {
    _V3_SET64( result, FFFFFFFF, FFFFFFFF );
  }
  else
  {
    /* thread will not go away until all the handles are closed */
    CloseHandle( thandle );
    result = v3u64u32( (u32) thread_id );
  }
  return( result );
#elif ( OS_TYPE == OS_OS2 )
  TID thread;
  stack_size += 8192;  /* ensure have, at least, the minimum */
  stack_size &= ~(4096-1); /* and align on 4k boundry */
  thread = _beginthread( start, 0, stack_size, arg );
  if ( thread != -1 )
  {
    result = v3u64u32( (u32) thread );
  }
  return( result );
#elif ( OS_TYPE == OS_SUNOS )
  thread_t tid;
  stkalign_t *stack;
  int ret;

  /* !!! use a thread stack cache? [init] */
  /* minimum stack size */
  if ( stack_size < ( MINSIGSTKSZ ) )
  {
    stack_size = MINSIGSTKSZ;
  }

  /* Allocate the stack */
  stack = (stkalign_t *) malloc( sizeof(stkalign_t) * stack_size );
  if ( stack == NULL )
  {
    _V3_SET64( result, FFFFFFFF, FFFFFFFF );
    return( result );
  }

  /* Create the critter */
  ret = lwp_create(&tid, (void * (*)( void * )) start, MINPRIO, 0,
    stack, 1, (int) arg, (int) 0, (int) 0, (int) 0);
  if ( ret != 0 )
  {
    _V3_SET64( result, FFFFFFFF, FFFFFFFF );
  }
  else
  {
    /* SunOS is 32bit only. */
    result.hi = (u32) tid.thread_id;
    result.lo = (u32) tid.thread_key;
  }
  return( result );
#else /* POSIX Threads? */
  pthread_t thread;
  pthread_attr_t attr;

  pthread_attr_init( &attr );

#if ( ( OS_TYPE == OS_IRIX ) || ( OS_TYPE == OS_IRIX64 ) )
  pthread_attr_setstacksize( &attr, (size_t) stack_size );
#elif ( OS_TYPE == OS_SOLARIS )
  /* system-wide contention */
  pthread_attr_setscope( &attr, PTHREAD_SCOPE_SYSTEM );
#endif

  result = v3u64u32( pthread_create( &thread, &attr,
    (void * (*)( void * )) start, arg ) );
  if ( !v3u64Eq( result, v3u64u32( 0 ) ) )
  {
    _V3_SET64( result, FFFFFFFF, FFFFFFFF );
  }
  else
  {
    result = v3u64u32( (u32) thread );
  }

  pthread_attr_destroy( &attr );
  return( result );
#endif
}

s32 v3ThreadID( u64 * thread_id )
{
#if ( defined( NO_THREADS ) )
  return( V3_FAIL );
#elif ( OS_TYPE == OS_MACOS )
  OSErr my_err;
  ThreadID thread_found;
  my_err = GetCurrentThread( &thread_found );
  if ( my_err == noErr )
  {
    *thread_id = v3u64u32( thread_found );
    return( V3_PASS );
  }
  return( V3_FAIL );
#elif ( OS_TYPE == OS_BEOS )
  int32 my_thread;
  my_thread = find_thread( NULL );
  if ( my_thread != B_NAME_NOT_FOUND )
  {
    *thread_id = v3u64u32( my_thread );
    return( V3_PASS );
  }
  return( V3_FAIL );
#elif ( OS_TYPE == OS_WIN32 )
  *thread_id = v3u64u32( (u32) GetCurrentThreadId() );
  return( V3_PASS );
#elif ( OS_TYPE == OS_OS2 )
  PTIB ptib;  /*  A pointer to the address of the TIB
                in which the current thread is returned. */
  PPIB ppib;  /* The address of a pointer to the PIB
                in which the current process is returned. */

  DosGetInfoBlocks( &ptib, &ppib );
  *thread_id = v3u64u32( ptib -> tib_ptib2 -> tib2_ultid );
  return( V3_PASS );
#elif ( OS_TYPE == OS_SUNOS )
  int result;
  thread_t tid;

  result = lwp_self( &tid );
  if ( result != 0 )
  {
    *thread_id = v3u64u32( 0 );
    return( V3_FAIL );
  }

  /* SunOS is 32bit only. */
  thread_id->hi = (u32) tid.thread_id;
  thread_id->lo = (u32) tid.thread_key;

  return( V3_PASS );
#else /* OS */
  *thread_id = v3u64u32( (u32) pthread_self() );
  return( V3_PASS );
#endif
}

u8 v3ThreadPriority( u8 priority )
{
#if ( defined( NO_THREADS ) )
  return( 0 );
#elif ( OS_TYPE == OS_MACOS )
  return( 0 );
#elif ( OS_TYPE == OS_BEOS )
  /* This needs to be worked out... */
  return( 0 );
#elif ( OS_TYPE == OS_WIN32 )
  s32 goal, process, thread;
  if ( priority != 0 )
  {
    goal = v3_PriorityToOS( priority );
    /*
      this table uses the priority scale for foreground processes, using the
      values for background proccesses caused major problems
    */
    switch ( goal )
    {
      case 1:
        process = NORMAL_PRIORITY_CLASS;
        thread = THREAD_PRIORITY_IDLE;
        break;
      case 2:
        process = IDLE_PRIORITY_CLASS;
        thread = THREAD_PRIORITY_LOWEST;
        break;
      case 3:
        process = IDLE_PRIORITY_CLASS;
        thread = THREAD_PRIORITY_BELOW_NORMAL;
        break;
      case 4:
        process = IDLE_PRIORITY_CLASS;
        thread = THREAD_PRIORITY_NORMAL;
        break;
      case 5:
        process = IDLE_PRIORITY_CLASS;
        thread = THREAD_PRIORITY_ABOVE_NORMAL;
        break;
      case 6:
        process = IDLE_PRIORITY_CLASS;
        thread = THREAD_PRIORITY_HIGHEST;
        break;
      case 7:
        process = NORMAL_PRIORITY_CLASS;
        thread = THREAD_PRIORITY_LOWEST;
        break;
      case 8:
        process = NORMAL_PRIORITY_CLASS;
        thread = THREAD_PRIORITY_BELOW_NORMAL;
        break;
      case 9:
        process = NORMAL_PRIORITY_CLASS;
        thread = THREAD_PRIORITY_NORMAL;
        break;
      default:
        process = NORMAL_PRIORITY_CLASS;
        thread = THREAD_PRIORITY_NORMAL;
        break;        
    }
    if ( ( SetPriorityClass( GetCurrentProcess(), process ) == 0 ) ||
      ( SetThreadPriority( GetCurrentThread(), thread ) == 0 ) )
    {
      return( 0 );
    }
  }

  if ( ( ( process = GetPriorityClass( GetCurrentProcess() ) ) == 0 ) ||
    ( ( thread = GetThreadPriority( GetCurrentThread() ) ) ==
    THREAD_PRIORITY_ERROR_RETURN ) )
  {
    return( 0 );
  }

  if ( process == IDLE_PRIORITY_CLASS )
  {
    switch ( thread )
    {
      case THREAD_PRIORITY_IDLE:
        goal = 1;
        break;
      case THREAD_PRIORITY_LOWEST:
        goal = 2;
        break;
      case THREAD_PRIORITY_BELOW_NORMAL:
        goal = 3;
        break;
      case THREAD_PRIORITY_NORMAL:
        goal = 4;
        break;
      case THREAD_PRIORITY_ABOVE_NORMAL:
        goal = 5;
        break;
      case THREAD_PRIORITY_HIGHEST:
        goal = 6;
        break;
      default:
        goal = 9;
        break;
    }
  }
  else if ( process == NORMAL_PRIORITY_CLASS )
  {
    switch ( thread )
    {
      case THREAD_PRIORITY_IDLE:
        goal = 1;
        break;
      case THREAD_PRIORITY_LOWEST:
        goal = 7;
        break;
      case THREAD_PRIORITY_BELOW_NORMAL:
        goal = 8;
        break;
      case THREAD_PRIORITY_NORMAL:
        goal = 9;
        break;
      default:
        goal = 9;
        break;
    }
  }
  else
  {
    goal = 9;
  }

  return( v3_PriorityToCosm( (u8) goal ) );
#elif ( OS_TYPE == OS_OS2 )
  PTIB  ptib;  /*  A pointer to the address of the TIB
  in which the current thread is returned. */
  PPIB  ppib;  /*  The address of a pointer to the PIB
  in which the current process is returned. */
  s32 pri_class;
  s32 pri_level;
  u32 ulClass;  /*  Priority class of a process. */
  s32 delta;    /*  Change to apply to the current base priority
  level of the process. */

  DosGetInfoBlocks( &ptib, &ppib );
  /* got current settings - small magic to get values */
  pri_class = ( ( ( ptib->tib_ptib2->tib2_ulpri ) >> 8 ) & 0x00FF );
  pri_level = ( ( ptib->tib_ptib2->tib2_ulpri ) & 0x001F );

  if ( priority == 0 )
  {  /* only asking current status */
    if ( pri_class == PRTYC_IDLETIME )
      return( v3_PriorityToCosm( pri_level ) );
    else
      return( V3_PRI_MAX );
  }
  else
  {
    if ( priority == 255 )
    { /* return to normal class / level 0 */
      ulClass = PRTYC_REGULAR;
      delta = -1 * pri_level;
    }
    else
    {  /* set idle class / level(mapped) */
      ulClass = PRTYC_IDLETIME;
      delta = v3_PriorityToOS( priority ) - pri_level;
    }
    if ( NO_ERROR == DosSetPriority( PRTYS_THREAD, ulClass, delta, 0 ) )
      return( priority );
    else
      return( 0 );
  }
#elif ( OS_TYPE == OS_SUNOS )
  int result;

  result = lwp_setpri( SELF, priority );
  if ( result != 0 )
  {
    /* !!! call pod_setmaxpri(255) maybe? [init] */
    return( 0 );
  }
  return( priority );
#else /* OS */
  pthread_attr_t attr;
  struct sched_param param;
  u8 pri_set;
  u8 pri_result;

  pri_set = v3_PriorityToCosm( priority );
  pri_result = 0;

  if ( priority != 0 )
  {
    pthread_attr_init( &attr );

    pthread_attr_getschedparam( &attr, &param );
    if ( param.sched_priority != pri_set )
    {
      param.sched_priority = pri_set;
      pri_result = pthread_attr_setschedparam( &attr, &param );
    }
    pthread_attr_destroy( &attr );
  }
  return( v3_PriorityToCosm( pri_result ) );
#endif
}

void v3ThreadEnd( void )
{
#if ( defined( NO_THREADS ) )
  return;
#elif ( OS_TYPE == OS_MACOS )
  u64 current;

  if ( v3ThreadID( &current ) == V3_PASS )
  {
    DisposeThread( v3u32u64( current ), nil, 0 );
  }
#elif ( OS_TYPE == OS_BEOS )
  exit_thread( 0 );
#elif ( ( OS_TYPE == OS_WIN32 ) || ( OS_TYPE == OS_OS2 ) )
  _endthreadex( 0 );
#elif ( OS_TYPE == OS_SUNOS )
  lwp_destroy( SELF );
#else /* OS */
  pthread_exit( NULL );
#endif
}

s32 v3MutexLock( v3_MUTEX * mutex, u32 wait )
{
#if ( defined( NO_THREADS ) )
  /* If we dont have threads, Mutex's dont need to be thread safe */
  if ( ( wait == V3_MUTEX_NOWAIT ) || ( wait == V3_MUTEX_WAIT ) )
  {
    /* we cant actually Yield() or anything without threads, so just fail */
    if ( mutex->lock == 0 )
    {
      mutex->lock = 1;
      return( V3_PASS );
    }
    return( V3_FAIL );
  }
  else /* wait is invalid */
  {
    return( V3_FAIL );
  }
#else
   /*
     !!!  (this is where we need a mini-protect, 1 static OS-mutex maybe)
     protect, check not already INIT then do INIT, unprotect
     since a Free is a user explicit thing, we'll be ok inside the outer if

     mini-protect
     if ( mutex->state != V3_MUTEX_STATE_INIT )
     {

     }
     un-mini-protect
   */
#if ( OS_TYPE == OS_WIN32 )
  if ( mutex->state != V3_MUTEX_STATE_INIT )
  {
    mutex->os_mutex = CreateSemaphore( NULL, 1, 1, NULL );
    if ( mutex->os_mutex == NULL )
    {
      return( V3_FAIL );
    }
    mutex->state = V3_MUTEX_STATE_INIT;
  }
#elif ( OS_TYPE == OS_MACOS )
    /* !!! mac stuff */
#elif ( OS_TYPE == OS_SUNOS )
  /* !!! SunOS LWP stuff */
#elif ( defined( _POSIX_THREADS ) )
  if ( mutex->state != V3_MUTEX_STATE_INIT )
  {
    if ( pthread_mutex_init( &mutex->os_mutex, NULL ) != 0 )
    {
      return( V3_FAIL );
    }
    mutex->state = V3_MUTEX_STATE_INIT;
  }
#else /* no OS mutex support, but threads */
  /* should never be allowed to happen */
#error "threads but no mutexes? check v3MutexLock()"
#endif /* OS */

  if ( wait == V3_MUTEX_NOWAIT )
  {
#if ( OS_TYPE == OS_WIN32 )
    if ( WaitForSingleObject( mutex->os_mutex, 0 ) != WAIT_OBJECT_0 );
    {
      return( V3_FAIL );
    }
#elif ( OS_TYPE == OS_MACOS )
    /* !!! mac stuff */
#elif ( OS_TYPE == OS_SUNOS )
    /* !!! SunOS LWP stuff */
#elif ( defined( _POSIX_THREADS ) )
    if ( pthread_mutex_trylock( &mutex->os_mutex ) != 0 )
    {
      return( V3_FAIL );
    }
#else /* no OS mutex support, but threads */
  /* should never be allowed to happen */
#error "threads but no mutexes? check v3MutexLock()"
#endif /* OS */
  }
  else if ( wait == V3_MUTEX_WAIT )
  {
#if ( OS_TYPE == OS_WIN32 )
    if ( WaitForSingleObject( mutex->os_mutex, INFINITE )
      != WAIT_OBJECT_0 );
    {
      return( V3_FAIL );
    }
#elif ( OS_TYPE == OS_MACOS )
    /* !!! mac stuff */
#elif ( OS_TYPE == OS_SUNOS )
    /* !!! SunOS LWP stuff */
#elif ( defined( _POSIX_THREADS ) )
    if ( pthread_mutex_lock( &mutex->os_mutex ) != 0 )
    {
      return( V3_FAIL );
    }
#else /* no OS mutex support, but threads */
  /* should never be allowed to happen */
#error "threads but no mutexes? check v3MutexLock()"
#endif /* OS */
  }
  else /* wait is invalid */
  {
    return( V3_FAIL );
  }
#endif /* NO_THREADS */

  return( V3_PASS );
}

s32 v3MutexUnlock( v3_MUTEX * mutex )
{
#if ( defined( NO_THREADS ) )
  /* If we dont have threads, Mutex's dont need to be thread safe */
  if ( mutex->lock == 1 )
  {
    mutex->lock = 0;
  }
  else
  {
    return( V3_FAIL );
  }
#else /* have threads */
  if ( mutex->state != V3_MUTEX_STATE_INIT )
  {
    return( V3_FAIL );
  }

#if ( OS_TYPE == OS_WIN32 )
  if ( ReleaseSemaphore( mutex->os_mutex, 1, NULL ) == 0 )
  {
    return( V3_FAIL );
  }
#elif ( OS_TYPE == OS_MACOS )
  /* !!! mac stuff */
#elif ( OS_TYPE == OS_SUNOS )
    /* !!! SunOS LWP stuff */
#elif ( defined( _POSIX_THREADS ) )
  if ( pthread_mutex_unlock( &mutex->os_mutex ) != 0 )
  {
    return( V3_FAIL );
  }
#else /* no OS mutex support, but threads */
  /* should never be allowed to happen */
#error "threads but no mutexes? check v3MutexUnlock()"
#endif /* OS */
#endif /* NO_THREADS */

  return( V3_PASS );
}

void v3MutexFree( v3_MUTEX * mutex )
{
#if ( defined( NO_THREADS ) )
  /* do nothing */
#else /* have threads */
  if ( mutex->state != V3_MUTEX_STATE_INIT )
  {
    return;
  }
#if ( OS_TYPE == OS_WIN32 )
  CloseHandle( mutex->os_mutex );
#elif ( OS_TYPE == OS_MACOS )
    /* !!! mac stuff */
#elif ( OS_TYPE == OS_SUNOS )
    /* !!! SunOS LWP stuff */
#elif ( defined( _POSIX_THREADS ) )
  pthread_mutex_destroy( &mutex->os_mutex );
#else /* no OS mutex support, but threads */
  /* should never be allowed to happen */
#error "threads but no mutexes? check v3MutexFree()"
#endif /* OS */
  mutex->state = V3_MUTEX_STATE_NONE;
#endif /* NO_THREADS */
}

u32 v3Sleep( u32 millisec )
{
#if ( OS_TYPE == OS_MACOS )
  unsigned long final;
  /*
    we'll avoid the use of all kinds of time managers and sacrifice some
    resolution. Delay in ticks (1 tick = 1/60 sec -> 16 msec
  */
  Delay( millisec >> 4, &final );
#elif ( OS_TYPE == OS_BEOS )
  int i;
  /* Check for bounds error here, MAX_u32/1000 */
  if ( millisec <= ( 0xFFFFFFFF / 1000 ) )
  {
    snooze( millisec * 1000 );
  }
  else
  {
    for ( i = 0 ; i < 1000 ; i++ )
    {
      snooze( millisec );
    }
  }
#elif ( OS_TYPE == OS_WIN32 )
  Sleep( millisec );
#elif ( OS_TYPE == OS_NUTC )
  int i;
  /* Check for boundry (1,000,000 is the limit) */
  if ( millisec < 1000 )
  {
    usleep( millisec * 1000 );
  }
  else
  {
    sleep( millisec / 1000 );
    usleep( ( millisec % 1000 ) * 1000 );
  }
#elif ( OS_TYPE == OS_OS2 )
  DosSleep( millisec );
#else
  int i;
  /* Check for bounds error here, MAX_u32/1000 */
  if ( millisec <= ( 0xFFFFFFFF / 1000 ) )
  {
    usleep( millisec * 1000 );
  }
  else
  {
    for ( i = 0 ; i < 1000 ; i++ )
    {
      usleep( millisec );
    }
  }
#endif
  return( 0 );
}

void v3Yield( void )
{
#if ( !defined( NO_THREADS ) )
#if ( OS_TYPE == OS_MACOS )
  YieldToAnyThread();
#elif ( OS_TYPE == OS_BEOS )
  /* there isn't any such creature, so sleep for 0.1 millisec */
  /* NOTE: snooze( 0 ) works, but it's not documented */
  snooze( 100 );
#elif ( OS_TYPE == OS_WIN32 )
  Sleep( 0 );
#elif ( OS_TYPE == OS_OS2 )
  DosSleep( 0 );
#elif ( OS_TYPE == OS_SUNOS )
  /* This doesn't do what you think it does... */
  /* (this will only yeild to the same lwp priority group) */
  lwp_yield( SELF );
#else /* OS */
#ifdef _POSIX_PRIORITY_SCHEDULING
  sched_yield();
#endif
#endif /* OS */
#endif /* !NO_THREADS */
}

s32 v3SignalRegister( u32 signal_type, void (*handler)(int) )
{
#if ( OS_TYPE == OS_WIN32 )
  ascii name[128];
  u64 pid;
  u64 thread_id;
  if ( __V3_WIN32_SIGNALS_INIT == 0 )
  {
    /* setup the Semaphores in unsignaled state */
    pid = v3ProcessID();
    if ( _V3_EQ64( pid, 00000000, 00000000 ) )
    {
      return( V3_FAIL );
    }
    v3PrintAStr( name, v3u64u32( 128 ), (ascii *) "%016Y-SIGINT",
      v3ProcessID() );
    __V3_WIN32_SIGINT = CreateSemaphore( NULL, 0, 1, name );
    v3PrintAStr( name, v3u64u32( 128 ), (ascii *) "%016Y-SIGTERM",
      v3ProcessID() );
    __V3_WIN32_SIGTERM = CreateSemaphore( NULL, 0, 1, name );
    if ( ( __V3_WIN32_SIGINT == NULL ) || ( __V3_WIN32_SIGINT == NULL ) )
    {
      return( V3_FAIL );
    }
    /* spawn a thread to go wait for a signals */
    thread_id = v3ThreadBegin( v3_SignalWaitThread, NULL, 4096 );
    if ( _V3_EQ64( thread_id, FFFFFFFF, FFFFFFFF ) )
    {
      return( V3_FAIL );
    }
    /* mark init as done */
    __V3_WIN32_SIGNALS_INIT = 1;
  }
  /* register the signal catchers */
  /* Win32 signal default is ignore SIGTERM and exit on SIGINT */
  switch ( signal_type )
  {
    case V3_PROCESS_SIGINT:
      if ( signal( SIGTERM, handler ) != SIG_ERR )
      {
        return( V3_PASS );
      }
      break;
    case V3_PROCESS_SIGTERM:
      if ( signal( SIGINT, handler ) != SIG_ERR )
      {
        return( V3_PASS );
      }
      break;
    default:
      return( V3_FAIL );
  }
#else
  switch ( signal_type )
  {
    case V3_PROCESS_SIGINT:
      if ( signal( SIGINT, handler ) != SIG_ERR )
      {
        return( V3_PASS );
      }
      break;
    case V3_PROCESS_SIGTERM:
      if ( signal( SIGTERM, handler ) != SIG_ERR )
      {
        return( V3_PASS );
      }
      break;
    default:
      return( V3_FAIL );
  }
#endif /* OS */

  return( V3_FAIL );
}

s32 v3SystemClock( v3_time * local_time )
{
  s128 os_local_time, unix_epoc;

#if ( OS_TYPE == OS_WIN32 )
  time_t utctime;
  SYSTEMTIME systime;
#elif ( OS_TYPE == OS_OS2 )
  DATETIME time_value; /* to get .hh time resolution */
  time_t utctime; /* secs since 1 Jan. 1970 UTC */
#else
  struct timeval time_value;
#endif

  if ( local_time == NULL )
  {
    return( V3_FAIL );
  }

#if ( OS_TYPE == OS_WIN32 )
  do
  {
    time( &utctime );
    GetSystemTime( &systime );
  } while ( utctime != time( NULL ) );

  os_local_time.hi.hi = 0;
  os_local_time.hi.lo = (u32) utctime;
  _V3_SET64( os_local_time.lo, 00418937, 4BC6A7EF );
  os_local_time.lo = v3u64Mul( v3u64u32( (u32) systime.wMilliseconds ),
    os_local_time.lo );
#elif ( OS_TYPE == OS_OS2 )
  do
  {
    time( &utctime );
    if ( DosGetDateTime( &time_value ) != NO_ERROR )
    {
      return( V3_FAIL );
    }
  } while ( utctime != time( NULL ) );

  os_local_time.hi.hi = 0;
  os_local_time.hi.lo = utctime;
  _V3_SET64( os_local_time.lo, 028F5C28, F5C28F5C );
  os_local_time.lo = v3u64Mul( v3u64u32( (u32) time_value.hundredths ),
    os_local_time.lo );
#else
  /* Query the OS for current time */
  if ( gettimeofday( &time_value, NULL ) != 0)
  {
    return( V3_FAIL );
  }

  /* Transform OS time to v3_time format */
  /* 2^64 / 1000000 = 0x000010C6F7A0B5ED */

#if ( defined( CPU_64BIT ) )
  os_local_time.hi = time_value.tv_sec;
  os_local_time.lo = time_value.tv_usec * 0x000010C6F7A0B5ED;
#else
  os_local_time.hi.hi = 0;
  os_local_time.hi.lo = time_value.tv_sec;
  _V3_SET64( os_local_time.lo, 000010C6, F7A0B5ED );
  os_local_time.lo = v3u64Mul( v3u64u32( time_value.tv_usec ),
    os_local_time.lo );
#endif /* 64 bit */

#endif
  /*
    gettimeofday gives the number of second since 1, January 1970
    In this count, there is no leap seconds, and all years divisible by 4
    are leap. So between 01/01/1970 and 01/01/2000 there is
    ( ( ( 366 * 7 ) + (365 * 23 ) ) * ( 24 * 60 * 60 ) ) =
    946.684.800 ( = 0x386D4380 ) seconds
  */
  _V3_SET128( unix_epoc, 00000000, 386D4380, 00000000, 00000000 );

  *local_time = v3s128Sub( os_local_time, unix_epoc );

  return( V3_PASS );
}

/* low level functions */

u8 v3_PriorityToCosm( int pri )
{
#if ( OS_TYPE == OS_MACOS )
  return( 0 );
#elif ( OS_TYPE == OS_BEOS )
  /* 1-120 -> 1-255 */
  /* This will need some magic to account for "realtime" processing */
  return( 0 );
#elif ( OS_TYPE == OS_WIN32 )
  /* 1-9 -> 1-255 */
  return( (u8) ( ( ( pri - 1 ) * 31 ) + 1 ) );
#elif ( OS_TYPE == OS_OS2 )
  /* (0 .. 31) => (1 .. 254); 32=255 */
  if ( pri >= 32 )
  {
    return( 255 );
  }
  else
  {
    return( 1 + ( 253 * (int) pri ) / 31 );
  }
#else /* OS */
  int ratio;

  /* Someone made UNIX's PRIO_MAX the low priority, Cosm has it straight */
  if ( ( pri > PRIO_MAX ) || ( pri < 0 ) )
  {
    /* out of our known range */
    return( 0 );
  }
  /* 20-0 -> 1-255 */
  ratio = ( PRIO_MAX * 1000 ) / ( V3_PRI_MAX - V3_PRI_MIN );
  return( V3_PRI_MAX - ( ( pri * 1000 ) / ratio ) );
#endif
}

int v3_PriorityToOS( u8 pri )
{
#if ( OS_TYPE == OS_MACOS )
  return( 0 );
#elif ( OS_TYPE == OS_BEOS )
  /* Reverse the logic in v3_PriorityToCosm */
  return( 0 );
#elif ( OS_TYPE == OS_WIN32 )
  /* 1-255 -> 1-9 */
  return( (u8) ( ( pri / 31 ) + 1 ) );
#elif ( OS_TYPE == OS_OS2 )
  /* (1 .. 254) => (0 .. 31); 255=32 */
  if ( pri >= V3_PRI_MAX )
  {
    return( 32 );
  }
  else
  {
    return( 31 * ( pri - 1 ) / 253 );
  }
#else
  int ratio;

  /* Someone made UNIX's PRIO_MAX the low priority, Cosm has it straight */
  /* 1-255 -> 20-0 */
  ratio = ( ( V3_PRI_MAX - V3_PRI_MIN ) * 1000 ) / PRIO_MAX;
  return( PRIO_MAX - ( ( pri * 1000 ) / ratio ) );
#endif
}

#if ( OS_TYPE == OS_WIN32 )
void v3_SignalWaitThread( void * arg )
{
  HANDLE handles[2];
  u32 go;

  handles[0] = __V3_WIN32_SIGINT;
  handles[1] = __V3_WIN32_SIGTERM;

  for ( ; ; )
  {
    /* wait for the signals */
    go = WaitForMultipleObjects( 2, handles, 0, INFINITE );
    switch ( go )
    {
      case WAIT_OBJECT_0:
        raise( SIGTERM );
        break;
      case ( WAIT_OBJECT_0 + 1 ):
        raise( SIGINT );
        break;
      default:
        /* if this happens, the process is probably already dead */
        break;
    }

  }

}
#endif /* win32 */

/* test code */

#if ( !defined( NO_THREADS ) )

#define V3_THREAD_BUFFERSIZE  64
#define V3_THREAD_MAXCOUNTER  255

typedef struct
{
  u64 tid_src;
  u64 tid_dest;
  u8 buffer[V3_THREAD_BUFFERSIZE];
  u32 head;
  u32 tail;
  s32 exit_src;
  s32 exit_dest;
  v3_MUTEX lock;
} thread_data;

void v3_ThreadTestSrc( void * arg )
{
  thread_data *data;
  u32 done;
  u32 count;

  data = (thread_data *) arg;

  if ( v3ThreadID( &data->tid_src ) != V3_PASS )
  {
    data->exit_src = -1;
    return;
  }

  done = 0;
  count = 0;

  while ( done == 0 )
  {
    v3MutexLock( &data->lock, V3_MUTEX_WAIT );
    if ( data->exit_dest != 0 )
    {
      /* noone to receive data */
      data->exit_src = -2;
      done = 1;
    }
    /* fill buffer */
    while ( ( ( ( data->tail + 1 ) % V3_THREAD_BUFFERSIZE ) != data->head ) &&
      ( count < V3_THREAD_MAXCOUNTER ) )
    {
      data->buffer[data->tail] = (u8) count++;
      data->tail = ( data->tail + 1 ) % V3_THREAD_BUFFERSIZE;
      if ( count == V3_THREAD_MAXCOUNTER )
      {
        data->exit_src = 1;
        done = 1;
      }
    }
    v3MutexUnlock( &data->lock );
    v3Yield();
  }

  v3ThreadEnd();
}

void v3_ThreadTestDest( void * arg )
{
  thread_data *data;
  u32 done;
  u32 count;
  u8 value;

  data = (thread_data *) arg;

  if ( v3ThreadID( &data->tid_dest ) != V3_PASS )
  {
    data->exit_dest = -1;
    return;
  }

  done = 0;
  count = 0;

  while ( done == 0 )
  {
    v3MutexLock( &data->lock, V3_MUTEX_WAIT );
    /* empty buffer */
    while ( ( data->head != data->tail ) && ( count < V3_THREAD_MAXCOUNTER ) )
    {
      value = data->buffer[data->head];
      if ( value != count++ )
      {
        data->exit_dest = -2;
        done = 1;
      }
      data->head = ( data->head + 1 ) % V3_THREAD_BUFFERSIZE;
    }
    if ( count == V3_THREAD_MAXCOUNTER )
    {
      data->exit_dest = 1;
      done = 1;
    }
    if ( data->exit_src != 0 )
    {
      /* noone to receive send more data? */
      if ( done != 1 )
      {
        /* if we're not already done that is bad */
        data->exit_src = -2;
        done = 1;
      }
    }
    v3MutexUnlock( &data->lock );
    v3Yield();
  }

  v3ThreadEnd();
}

#endif /* have threads */

s32 v3_TestCosmTask( void )
{
  u32 ncpu;

#if ( defined( NO_THREADS ) )
  u64 thread_id;
  u64 res;
#else /* have threads */
  thread_data data;
  u64 thread_id[2];
  u32 done;
#endif /* NO_THREADS */

#if ( OS_TYPE == OS_SOLARIS )
  /* have CPULock/Unlock */
  u32 cpu;
#endif 

#if ( defined( NO_THREADS ) )
  /* make sure thread functions return correct errors */
  res = v3ThreadBegin( NULL, NULL, 1024 );

  if ( !_V3_EQ64( res, FFFFFFFF, FFFFFFFF ) )
  {
    return( -1 );
  }

  if ( v3ThreadID( &thread_id ) != V3_FAIL )
  {
    return( -2 );
  }

  if ( v3ThreadPriority( 255 ) != 0 )
  {
    return( -3 );
  }

  v3ThreadEnd();
#else /* have threads */
  /* Tests the Thread functions */
  v3MemSet( &data, v3u64u32( sizeof( thread_data ) ), 0 );

  _V3_SET64( thread_id[0], 00000000, 00000000 );
  _V3_SET64( thread_id[1], 00000000, 00000000 );

  thread_id[0] = v3ThreadBegin( v3_ThreadTestSrc, (void *) &data, 4096 );
  thread_id[1] = v3ThreadBegin( v3_ThreadTestDest, (void *) &data, 4096 );

  /* Test for bad thread creation */
  if ( _V3_EQ64( thread_id[0], FFFFFFFF, FFFFFFFF ) ||
    _V3_EQ64( thread_id[1], FFFFFFFF, FFFFFFFF ) )
  {
    return( -4 );
  }

  /* threads are running, wait for them to finish */
  done = 0;
  while ( done == 0 )
  {
    v3MutexLock( &data.lock, V3_MUTEX_WAIT );
    if ( ( data.exit_src != 0 ) && ( data.exit_dest != 0 ) )
    {
      done = 1;
    }
    v3MutexUnlock( &data.lock );
    v3Yield();
  }

  if ( data.exit_src != 1 )
  {
    return( -5 );
  }

  if ( data.exit_dest != 1 )
  {
    return( -6 );
  }

  /* Test for v3ThreadID being correct */
  if ( !v3u64Eq( data.tid_src, thread_id[0] ) ||
    !v3u64Eq( data.tid_dest, thread_id[1] ) )
  {
    return( -7 );
  }
#endif /* NO_THREADS */

  ncpu = v3CPUCount();
  if ( ncpu == 0 )
  {
    return( -8 );
  }

  /* Test CPULock and Unlock */

  /* some things that should always fail */
  if ( v3CPULock( v3u64u32( 0 ), (u32) -1 ) != V3_FAIL )
  {
    return( -9 );
  }
  if ( v3CPULock( v3ProcessID(), (u32) ncpu ) != V3_FAIL )
  {
    return( -10 );
  }
  if ( v3CPUUnlock( v3u64u32( 0 ) ) != V3_FAIL )
  {
    return( -11 );
  }

#if ( OS_TYPE == OS_SOLARIS )
  /* we should be able to lock and unlock */
  for ( cpu = 0 ; cpu < ncpu ; cpu++ )
  {
    if ( v3CPULock( v3ProcessID(), (u32) cpu ) != V3_PASS )
    {
      return( -12 );
    }
  }
  if ( v3CPUUnlock( v3ProcessID() ) != V3_PASS )
  {
    return( -13 );
  }
#else /* OS */
  /* OS without CPULock/Unlock support, will fail */
  if ( v3CPULock( v3ProcessID(), (u32) 0 ) != V3_FAIL )
  {
    return( -14 );
  }
  if ( v3CPUUnlock( v3ProcessID() ) != V3_FAIL )
  {
    return( -15 );
  }
#endif /* CPULock/Unlock */

  return( V3_PASS );
}
