/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

/* CPU/OS Layer - CPU and OS specific code is allowed */

#ifndef V3_COSMTASK_H
#define V3_COSMTASK_H

#include "cputypes.h"
#include "cosmmath.h"

#if ( OS_TYPE == OS_WIN32 )
#include <windows.h>
#include <winbase.h>
#elif ( OS_TYPE == OS_MACOS )
  /* !!! macos stuff */
#elif ( OS_TYPE == OS_BEOS )
  /* !!! BeOS stuff */
#elif ( OS_TYPE == OS_OS2 )
  /* !!! OS/2 stuff */
#else /* OS */
#include <unistd.h>
#if ( !defined( NO_THREADS ) )
#if ( OS_TYPE == OS_LINUX )
#define _POSIX_THREADS 1 /* linux forgots this on most systems */
#endif /* linux */
#if ( defined( _POSIX_THREADS ) )
#include <pthread.h>
#endif /* POSIX threads */
#endif /* have threads */
#endif /* OS */

#define V3_PROCESS_SIGPING  0  /* Test for Process Existance */
#define V3_PROCESS_SIGINT   1  /* Save/Checkpoint */
#define V3_PROCESS_SIGTERM  2  /* Save and Terminate */
#define V3_PROCESS_SIGKILL  3  /* Terminate with extreme prejudice */

#define V3_PRI_MAX 255  /* maximum thread priority */
#define V3_PRI_MIN 1    /* minimum thread priority */

#define V3_MUTEX_NOWAIT 13
#define V3_MUTEX_WAIT   14

#define V3_MUTEX_STATE_NONE  0
#define V3_MUTEX_STATE_INIT  42

typedef struct
{
  u32 state;
#if ( defined( NO_THREADS ) )
  u32 lock;
#elif ( OS_TYPE == OS_WIN32 )
  HANDLE os_mutex; /* a semaphore */
#elif ( OS_TYPE == OS_MACOS )
  /* !!! mac stuff */
#elif ( defined( _POSIX_THREADS ) )
  pthread_mutex_t os_mutex;
#endif
} v3_MUTEX;

/* Process Functions */

u64 v3ProcessID( void );
  /*
    Returns: the current Process ID.
  */

u8 v3ProcessPriority( u8 priority );
  /*
    Attempt to set the task priority from 1 to 255. If 0 is passed, then
    no change will be made and the current priority will be returned.
    1 is low priority, 255 is high priority. The procedure will remap the
    priority into values appropriate for the system. This procedure will never
    increase the priority beyond the default normal (255) priority for the OS.
    Returns: Process priority (mapped to 1-255 scale) after call is completed,
      0 indicates failure.
   */

u64 v3ProcessSpawn( const ascii * command, const ascii * arguments, ... );
  /*
    Spawn a new process running command. The command must be a full path in
    Cosm format to the binary.
    All additional arguments will not be modified. The arguments MUST
    end with a NULL. e.g. v3ProcessSpawn( "cat", "file.txt", NULL );
    Returns: The child's process ID, or 0 on error.
  */

s32 v3ProcessSignal( u64 process_id, u32 signal_type );
  /*
    Sends signal to process.
    Valid signals are:
      V3_PROCESS_SIGPING  Test for Process Existance
      V3_PROCESS_SIGINT   Save/Checkpoint
      V3_PROCESS_SIGTERM  Save and Terminate
      V3_PROCESS_SIGKILL  Terminate with extreme prejudice
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

void v3ProcessEnd( int status );
  /*
    End the Process and all threads, return status as the exit code.
    Returns: Nothing.
  */

/* CPU Functions */

u32 v3CPUCount( void );
  /*
    Returns: number of CPUs in the system, 0 indicates failure.
  */

s32 v3CPUGet( u32 * cpu );
  /*
    Get the CPU# the process is running on and set cpu. Not real meaningful
    unless it has been locked, but may avoid a process move if called
    before a v3CPULock().
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s32 v3CPULock( u64 process_id, u32 cpu );
  /*
    Lock the process, and any threads, to the CPU cpu. This function
    is primarily for implimenting V3_MEM_SECURE memory allocation.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s32 v3CPUUnlock( u64 process_id );
  /*
    Unlock the process from a set CPU. If you are using any V3_MEM_SECURE
    memory, calling this function would be a very very bad idea.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

/* Thread Functions */

u64 v3ThreadBegin( void (*start)(void *), void * arg, u32 stack_size );
  /*
    Begin a thread, make a stack, call the function start.
    Returns: Assigned thread ID, 0xFFFFFFFFFFFFFFFF (-1) indicates failure.
  */

s32 v3ThreadID( u64 * thread_id );
  /*
    Puts the current Thread ID into thread_id.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

u8 v3ThreadPriority( u8 priority );
  /*
    Attempt to set the thread priority from 1 to 255. If 0 is passed, then
    no change will be made and the current priority will be returned.
    1 is low priority, 255 is high priority. The procedure will remap the
    priority into values appropriate for the system. This procedure will
    never increase the priority beyond the default normal (255) priority for
    the OS.
    Returns: Thread priority (mapped to 1-255 scale) after call is completed,
      0 indicates failure.
  */

void v3ThreadEnd( void );
  /*
    End the calling thread.
    Returns: Nothing, thread's dead.
  */

/* Data locks for threadsafe operations */

s32 v3MutexLock( v3_MUTEX * mutex, u32 wait );
  /*
    Set an exclusive lock.
    If wait is V3_MUTEX_NOWAIT, then return immediately.
    If wait is V3_MUTEX_WAIT then do not return until lock is available.
    You should _never_ have more then one lock alive at a time per thread,
    as this may result in deadlock. Note that if you create a mutex you MUST
    free it eventually with v3MutexFree.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s32 v3MutexUnlock( v3_MUTEX * mutex );
  /*
    Unset an exclusive lock.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

void v3MutexFree( v3_MUTEX * mutex );
  /*
    Free the mutex and related system resources. Do not call this unless
    you are completely done with the mutex, very bad things will happen.
    Returns: nothing.
  */

/* Sleep */

u32 v3Sleep( u32 millisec );
  /*
    Sleep for millisec milliseconds.
    Returns: 0 if slept for complete time, or time remaining if returned
      early.
  */

void v3Yield( void );
  /*
    Yield to another thread/process that wants to run.
    Returns: nothing
  */

/* Signals */

s32 v3SignalRegister( u32 signal_type, void (*handler)(int) );
  /*
    Register handler as the handler for the signal.
    signal_type must be one of V3_PROCESS_SIG*. See v3ProcessSignal above.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

/* Time */

s32 v3SystemClock( v3_time * local_time );
  /*
    Get the time from the system clock and set local_time.
    v3_time is a signed number (s128) of seconds in 64b.64b fixed point
    format based on the time 0 = 00:00:00 UTC, Jan 1, 2000 AD.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

/* Low level helper functions */

u8 v3_PriorityToCosm( int pri );
  /*
    Translate native priority range to Cosm range.
    Returns: Priority in Cosm range.
  */

int v3_PriorityToOS( u8 pri );
  /*
    Translate Cosm priority range to system range.
    Returns: Priority in native system terms.
  */

/* testing */

void v3_ThreadTestSrc( void * arg );
  /*
    This thread will produce data and put it into the buffer starting with
    0 and counting to 255.
  */

void v3_ThreadTestDest( void * arg );
  /*
    This thread will consume data and take it out of the buffer,
    it will expect numbers starting with 0 and counting to 255.
  */

s32 v3_TestCosmTask( void );
  /*
    Test functions in this header.
    Returns: V3_PASS on success, or a negative number corresponding to the
      test that failed.
  */

#endif
