/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

#include "cosmmath.h"
#include "cosmtask.h"
#include "cosmtime.h"
#include "cosmfile.h"
#include "cosmnet.h"

s32 v3Time( v3_time * dest, const v3_TIME_CORRECTION * const corrections )
{
  v3_time lt, diff, drift, result;

  /* Start with a few sanity checkes. */
  if ( ( dest == NULL ) || ( corrections == NULL ) )
  {
    return( V3_FAIL );
  }

  /* Get the lock on the correction. */
  if ( v3MutexLock( (v3_MUTEX *) &corrections->lock,
    V3_MUTEX_WAIT ) == V3_FAIL )
  {
    return( V3_FAIL );
  }

  /* Check if initialized properly.*/
  if ( v3u64Eq( corrections->initialized, v3u64u32( 0 ) ) )
  {
    v3MutexUnlock( (v3_MUTEX *) &corrections->lock );
    return( V3_FAIL );
  }

  if ( v3SystemClock( &lt ) == V3_FAIL )
  {
    v3MutexUnlock( (v3_MUTEX *) &corrections->lock );
    return( V3_FAIL );
  }

  if ( v3s128Gt( corrections->time_last_set, lt ) )
  {
    v3MutexUnlock( (v3_MUTEX *) &corrections->lock );
    return( V3_FAIL );
  }

  /* time = clock + offset + drift * seconds( mytime - time_last_set ) */

  /* sneaky way to shift diff right 64 */
  diff = v3s128s64( v3s128Sub( lt, corrections->time_last_set ).hi );
  drift = v3s128Mul( diff, corrections->time_drift );
  result = v3s128Add( v3s128Add( lt, drift ), corrections->time_offset );

  /* release the lock. */
  v3MutexUnlock( (v3_MUTEX *) &corrections->lock );

  *dest = result;

  return( V3_PASS );
}

s32 v3TimeSet( v3_TIME_CORRECTION * corrections,
  u32 * ip_list, u32 ip_count )
{
  /* INSERT CODE HERE */
  return(-1);
}

s32 v3TimeUnitsGregorian( v3_TIME_UNITS * units, v3_time time )
{
  s64 seconds, leaps;
  u32 max_days[] = { 30, 28, 30, 29, 30, 29, 30, 30, 29, 30, 29, 30 };
  u32 total_days[] = { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273,
                       304, 334, 365 };

  if ( units == NULL )
  {
    return( V3_FAIL );
  }

  units->subsec = time.lo;
  seconds = time.hi;

  /* Day of week */
  units->wday = (u32) v3s32s64( v3s64Mod( v3s64Div( seconds,
    v3s64s32( 24 * 60 * 60 ) ), v3s64s32( 7 ) ) );

  /* January 1, 2000 is a Saturday */
  if ( v3s64Lt( seconds, v3s64s32( 0 ) ) )
  {
    units->wday = ( units->wday + 5 ) % 7;
  }
  else
  {
    units->wday = ( units->wday + 6 ) % 7;
  }

  /* Year */
  /* 365.2425 * 24 * 60 * 60 = 0x1E18558 */
  /* First approximation of the year */
  units->year = v3s64Div( seconds, v3s64s32( 0x1E18558 ) );
  seconds = v3s64Sub( seconds, v3s64Mul( units->year,
    v3s64s32( 365 * 24 * 60 * 60 ) ) );

  /* Need a positive remaining number of seconds to simplify calculation */
  if ( v3s64Lt( seconds, v3s64s32( 0 ) ) )
  {
    v3s64Dec( &units->year );
    seconds = v3s64Add( seconds, v3s64s32( 365 * 24 * 60 * 60 ) );
  }

  /* Using the approximate number of year, also count for leap days */
  if ( v3s64Lt( units->year, v3s64s32( 0 ) ) )
  {
    leaps = v3s64Add( v3s64Sub ( v3s64Div( units->year,
      v3s64s32( 4 ) ), v3s64Div( units->year, v3s64s32( 100 ) ) ),
      v3s64Div( units->year, v3s64s32( 400 ) ) );
  }
  else
  {
    leaps = v3s64Add( v3s64Add( v3s64Sub( v3s64Div( units->year,
      v3s64s32( 4 ) ), v3s64Div( units->year, v3s64s32( 100 ) ) ),
      v3s64Div( units->year, v3s64s32( 400 ) ) ), v3s64s32( 1 ) );
  }
  if ( !v3s64Lt( units->year, v3s64s32( 0 ) ) && 
    v3_TimeIsYearLeap( v3s64Add( units->year, v3s64s32( 2000 ) ) ) )
  {
    v3s64Dec( &leaps );
  }

  seconds = v3s64Sub( seconds, v3s64Mul( leaps, v3s64s32( 24 * 60 * 60 ) ) );

  /* Set known year value and check if remainder is less than one year */
  if ( v3s64Lt( seconds, v3s64s32( 0 ) ) )
  {
    units->year = v3s64Add( units->year, v3s64s32( 1999 ) );
    if ( v3_TimeIsYearLeap( units->year ) )
    {
      seconds = v3s64Add( seconds, v3s64s32( 366 * 24 * 60 * 60 ) );
    }
    else
    {
      seconds = v3s64Add( seconds, v3s64s32( 365 * 24 * 60 * 60 ) );
    }
  }
  else
  {
    units->year = v3s64Add( units->year, v3s64s32( 2000 ) );
  }

  /* If remainder is still more than one year, fix it */
  if ( !v3s64Lt( seconds, v3s64s32( 366 * 24 * 60 * 60 ) ) ||
    ( !v3_TimeIsYearLeap( units->year ) &&
    !v3s64Lt( seconds, v3s64s32( 365 * 24 * 60 * 60 ) ) ) )
  {
    if ( v3_TimeIsYearLeap( units->year ) )
    {
      seconds = v3s64Sub( seconds, v3s64s32( 366 * 24 * 60 * 60 ) );
    }
    else
    {
      seconds = v3s64Sub( seconds, v3s64s32( 365 * 24 * 60 * 60 ) );
    }
    v3s64Inc( &units->year );
  }

  /* Seconds */
  units->sec = (u32) v3s32s64( v3s64Mod( seconds, v3s64s32( 60 ) ) );
  seconds = v3s64Div( seconds, v3s64s32( 60 ) );

  /* minutes */
  units->min = (u32) v3s32s64( v3s64Mod( seconds, v3s64s32( 60 ) ) );
  seconds = v3s64Div( seconds, v3s64s32( 60 ) );

  /* Hours */
  units->hour = (u32) v3s32s64( v3s64Mod( seconds, v3s64s32( 24 ) ) );
  seconds = v3s64Div( seconds, v3s64s32( 24 ) );

  /* Day of year */
  units->yday = (u32) v3s32s64( seconds );

  /* Month */
  units->month = units->yday / 30;
  if ( units->month > 11 )
  {
    units->month = 11;
  }
  if ( v3_TimeIsYearLeap( units->year ) && units->yday >= 59 )
  {
    if ( total_days[units->month] + 1 > units->yday )
    {
      units->month--;
    }
    if ( total_days[units->month + 1] + 1 <= units->yday )
    {
      /* Can this happen on leap years? */
      units->month++;
    }
    if ( units->month > 1 )
    {
      /* Remove Feb 29 from seconds now */
      v3s64Dec( &seconds );
    }
  }
  else
  {
    if ( total_days[units->month] > units->yday )
    {
      units->month--;
    }
    if ( total_days[units->month + 1] <= units->yday )
    {
      /* Probably only useful on march 1st */
      units->month++;
    }
  }
  seconds = v3s64Sub( seconds, v3s64s32( total_days[units->month] ) );
  if ( units->month > 11 )
  {
    /* Something is broken in the code */
    return( V3_FAIL );
  }

  /* Day of month */
  units->day = (u32) v3s32s64( seconds );
  if ( units->day > max_days[units->month] )
  {
    /* Something is broken in the code */
    return( V3_FAIL );
  }

  return( V3_PASS );
}

s32 v3TimeDigestGregorian( v3_time * time, const v3_TIME_UNITS * const units )
{
  s64 tmp, seconds;
  u32 max_days[] = { 30, 28, 30, 29, 30, 29, 30, 30, 29, 30, 29, 30 };
  s32 total_days[] = { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273,
                       304, 334 };

  if ( ( time == NULL ) || ( units == NULL ) )
  {
    return( V3_FAIL );
  }

  /* years */
  tmp = v3s64Sub( units->year, v3s64s32( 2000 ) );
  seconds = v3s64Add( v3s64Sub( v3s64Add(
    v3s64Mul( tmp, v3s64s32( 365 ) ), v3s64Div( tmp, v3s64s32( 4 ) ) ),
    v3s64Div( tmp, v3s64s32( 100 ) ) ), v3s64Div( tmp, v3s64s32( 400 ) ) );
  if ( v3s64Lt( units->year, v3s64s32( 2000 ) ) )
  {
    if ( ( units->month > 1 ) && v3_TimeIsYearLeap( units->year ) )
    {
      /* After Feb 29 on a leap year */
      v3s64Inc( &seconds );
    }
  }
  else
  {
    v3s64Inc( &seconds ); /* Year 2000 leap day */
    if ( ( units->month < 2 ) && v3_TimeIsYearLeap( units->year ) )
    {
      /* Before Feb 29 on a leap year */
      v3s64Dec( &seconds );
    }
  }

  /* months */
  if ( units->month > 11 )
  {
    /* Invalid month number */
    return( V3_FAIL );
  }

  seconds = v3s64Add( seconds, v3s64s32( total_days[units->month] ) );

  /* days */
  if ( units->day > max_days[units->month] )
  {
    /* Invalid number of days for that month */
    return( V3_FAIL );
  }

  if ( ( units->month == 1) && ( units->day == 28 ) )
  {
    /* Feb 29 -- Make sure the year IS leap */
    if ( !v3_TimeIsYearLeap( units->year ) )
    {
      return( V3_FAIL );
    }
  }

  seconds = v3s64Add( seconds, v3s64s32( (s32) units->day ) );

  /* Bound verifications */
  if ( !v3s64Lt( units->year, v3s64s32( 2000 ) ) )
  {
    if ( v3s64Lt( seconds, v3s64Sub( units->year, v3s64s32( 2000 ) ) ) )
    {
      /* Overflow */
      return( V3_FAIL );
    }
  }
  else
  {
    if ( v3s64Gt( seconds, v3s64Sub( units->year, v3s64s32( 2000 ) ) ) )
    {
      /* Underflow */
      return( V3_FAIL );
    }
  }

  /* hours */
  if ( units->hour > 23 )
  {
    /* Invalid number of hours */
    return( V3_FAIL );
  }
  seconds = v3s64Mul( seconds, v3s64s32( 24 ) );
  seconds = v3s64Add( seconds, v3s64s32( (s32) units->hour ) );

  /* minutes */
  if ( units->min > 59 )
  {
    /* Invalid number of minutes */
    return( V3_FAIL );
  }
  seconds = v3s64Mul( seconds, v3s64s32( 60 ) );
  seconds = v3s64Add( seconds, v3s64s32( (s32) units->min ) );

  /* seconds */
  if ( units->sec > 59 )
  {
    /* !!! Accept more if it's on a leap second !!! */
    /* Invalid number of seconds */
    return( V3_FAIL );
  }
  seconds = v3s64Mul( seconds, v3s64s32( 60 ) );
  seconds = v3s64Add( seconds, v3s64s32( (s32) units->sec ) );
  /* !!! add leap seconds here if necessary !!! */

  /* Now fill up the time struct */
  time->hi = seconds;
  if ( v3s64Lt( seconds, v3s64s32( 0 ) ) )
  {
    time->lo = v3u64Sub( v3u64u32( 0 ), units->subsec );
  }
  else
  {
    time->lo = units->subsec;
  }

  return( V3_PASS );
}

/* low level */

s32 v3_TimeIsYearLeap( s64 year )
{
  /* Based on Great Britain Dominions (in sept. 1752) calendar change
     This includes USA(except Alaska), part of Canada, Australia,
     Ireland, Scotland and Wales */
  /* With no year 0, so before year 1, it was year -1 */

  if ( v3s64Gt( year, v3s64s32( 1752 ) ) )
  {
    /* Gregorian Calendar */
    /* There is no official year 4000 rule yet */
    if ( v3s64Eq( v3s64Mod( year, v3s64s32( 4 ) ), v3s64s32( 0 ) ) &&
      ( !v3s64Eq( v3s64Mod( year, v3s64s32( 100 ) ), v3s64s32( 0 ) ) ||
      v3s64Eq( v3s64Mod( year, v3s64s32( 400 ) ), v3s64s32( 0 ) ) ) )
    {
      /* Year is leap */
      return( 1 );
    }
    return( 0 );
  }
  if ( v3s64Gt( year, v3s64s32( 7 ) ) )
  {
    /* Corrected Julian Calendar */
    return( v3s64Eq( v3s64Mod( year, v3s64s32( 4 ) ), v3s64s32( 0 ) ) );
  }
  if ( !v3s64Lt( year, v3s64s32( -45 ) ) )
  {
    /* Wrong start with the Julian Calendar */
    if ( v3s64Gt( year, v3s64s32( -9 ) ) )
    {
      /* No leap years in that period to compensate for the extra days
         introduced in the previous years */
      return( 0 );
    }
    return( v3s64Eq( v3s64Mod( year, v3s64s32( 3 ) ), v3s64s32( 0 ) ) );
  }
  /* Before the Julian Calendar, it was a real mess!!! */
  return( 0 );
}

/* testing */

s32 v3_TestCosmTime( void )
{
  v3_time time_test;

  if ( v3SystemClock( &time_test ) != V3_PASS )
  {
    return( -1 );
  }

  return( V3_PASS );
}
