/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

#ifndef V3_COSMTIME_H
#define V3_COSMTIME_H

#include "cputypes.h"
#include "cosmtask.h"

/*
  Times are signed s128's, 64b.64b format based 0 = 00:00:00 UTC, Jan 1,
  2000 AD. This gives a Range of +/- 2.923E11 years and a resolution of
  5.421E-20 seconds. This allows times from bang to crunch - assuming a closed
  universe (if it's not closed, we have bigger problems). It's also useful for
  timing the half lives of particles like Neutral Sigma Baryons.

  Useful constants:
  Beginning of time: 80000000, 00000001
  End of time: 7FFFFFFF, FFFFFFFF
  microsecond: 000010C6, F7A0B5ED
  millisecond: 00418937, 4BC6A7EF
  centisecond: 028F5C28, F5C28F5C
  UNIX epoc = Cosm + 0x386D4380
*/

/*
  Use:
  const ascii months[12] = V3_TIME_MONTHS;
  const ascii days[7] = V3_TIME_DAYS;
  etc...
*/
#define V3_TIME_MONTHS { (ascii *) "January", (ascii *) "Feburary", \
  (ascii *) "March", (ascii *) "April", (ascii *) "May", \
  (ascii *) "June", (ascii *) "July", (ascii *) "August", \
  (ascii *) "September", (ascii *) "October", (ascii *) "November", \
  (ascii *) "December" }

#define V3_TIME_MONTHS3 { (ascii *) "Jan", (ascii *) "Feb", \
  (ascii *) "Mar", (ascii *) "Apr", (ascii *) "May", (ascii *) "Jun", \
  (ascii *) "Jul", (ascii *) "Aug", (ascii *) "Sep", (ascii *) "Oct", \
  (ascii *) "Nov", (ascii *) "Dec" }

#define V3_TIME_DAYS { (ascii *) "Sunday", (ascii *) "Monday", \
  (ascii *) "Tuesday", (ascii *) "Wednesday", (ascii *) "Thursday", \
  (ascii *) "Friday", (ascii *) "Saturday" }

#define V3_TIME_DAYS3 { (ascii *) "Sun", (ascii *) "Mon", \
    (ascii *) "Tue", (ascii *) "Wed", (ascii *) "Thr", (ascii *) "Fri", \
    (ascii *) "Sat" }

typedef struct
{
  /* all values are zero based */
  s64 year;   /* year */
  u32 month;  /* month */
  u32 day;    /* day of month */
  u32 wday;   /* day of week */
  u32 yday;   /* day of year */
  u32 hour;   /* hours */
  u32 min;    /* minutes */
  u32 sec;    /* seconds */
  u64 subsec; /* sub-seconds, note: unsigned unlike v3_time */
} v3_TIME_UNITS;

typedef struct
{
  v3_time time_last_set;  /* when originally set, for setting drift */
  v3_time time_offset;    /* offset from system time */
  v3_time time_drift;     /* drift/sec */
  u64     initialized;    /* setup or not */
  v3_MUTEX lock;
} v3_TIME_CORRECTION;

s32 v3Time( v3_time * dest, const v3_TIME_CORRECTION * const corrections );
  /*
    Get the current time. Already Synchronized with network time.
    Compensate for system clock drift and correct accordingly.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s32 v3TimeSet( v3_TIME_CORRECTION * corrections,
  u32 * ip_list, u32 ip_count );
  /*
    Synchronize with network time. Dont attempt to set the system clock even
    though that would be nice. Set the value of time_offset and related values
    relative to the system clock instead. Use the ip_list as the list of
    time servers to sync with.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s32 v3TimeUnitsGregorian( v3_TIME_UNITS * units, v3_time time );
  /*
    Fill in the v3_TIME_UNITS structure with the time in Gregorian format.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s32 v3TimeDigestGregorian( v3_time * time,
  const v3_TIME_UNITS * const units );
  /*
    Convert back to a v3_time value from Gregorian, sanity check the values.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

/*
  With the above we can print whatever format we want.
  However, the year/month/day order should always be: year, month, day
  Since this is no fun to reverse parse, make sure we never need to.
  Keep all data in machine format, and only display it when needed.
*/

/* low level */

s32 v3_TimeIsYearLeap( s64 year );
  /*
    Check if year is a leap year.
    Returns: 1 if the year is a leap year, or 0 if it is not a leap year.
  */

/* testing */

s32 v3_TestCosmTime( void );
  /*
    Test functions in this header.
    Returns: V3_PASS on success, or a negative number corresponding to the
      test that failed.
  */

#endif
