/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

/* CPU/OS Layer - CPU and OS specific code is allowed */

#ifndef V3_CPUTYPES_H
#define V3_CPUTYPES_H

/* One of the following must be defined */
#if ( ( !defined( CPU_64BIT ) && !defined( CPU_32BIT ) ) \
  || ( defined( CPU_64BIT ) && defined( CPU_32BIT ) ) )
#error "Define either CPU_64BIT or CPU_32BIT"
#endif

/* CPU_WORD_SIZES must be defined to sizeof(long, int, short) */
#if ( !defined( CPU_WORD_SIZES ) )
#error "Define CPU_WORD_SIZES - see cputypes.h"
#endif

/* Define 8bit types */
typedef unsigned char u8;
typedef signed char s8;

/* Define 16bit types */
typedef unsigned short u16;
typedef signed short s16;

/* Define the 32bit types */
#if ( CPU_WORD_SIZES == 422 )
typedef unsigned long u32;
typedef signed long s32;
#elif ( CPU_WORD_SIZES == 442 )
typedef unsigned int u32;
typedef signed int s32;
#elif ( CPU_WORD_SIZES == 842 )
typedef unsigned int u32;
typedef signed int s32;
#else
#error "Invalid CPU_WORD_SIZES - see cputypes.h"
#endif

/* Define 64bit types */
#if ( defined( CPU_64BIT ) )
#if ( CPU_WORD_SIZES == 842 )
typedef unsigned long u64;
typedef signed long s64;
#else
typedef unsigned long long u64;
typedef signed long long s64;
#endif
#else
typedef struct { u32 hi; u32 lo; } u64;
typedef struct { s32 hi; u32 lo; } s64;
#endif

/*
  Define the 64 bit setting and equality macros
  usage: _V3_SET64( a, fedcba98, 76543210 );
  usage: if ( _V3_EQ64( a, fedcba98, 76543210 ) ) ...
*/
#if ( defined( CPU_64BIT ) )
#define _V3_SET64( dest, a, b ) \
  dest = 0x##a##b;
#define _V3_EQ64( value, a, b ) \
  ( value == 0x##a##b )
#else
#define _V3_SET64( dest, a, b ) \
  dest##.hi = 0x##a; dest##.lo = 0x##b;
#define _V3_EQ64( value, a, b ) \
  ( ( value.hi == 0x##a ) && ( value.lo == 0x##b ) )
#endif

/* We don't support native 128bit CPUs. All systems will fake 128bit */
typedef struct { u64 hi; u64 lo; } u128;
typedef struct { s64 hi; u64 lo; } s128;

/*
  Define the 128 bit setting and equality macros
  usage: _V3_SET128( a, fedcba98, 76543210, 01234567, 89abcdef );
  usage: if ( _V3_EQ128( a, fedcba98, 76543210, 01234567, 89abcdef ) ) ...
*/
#if ( defined( CPU_64BIT ) )
#define _V3_SET128( dest, a, b, c, d ) \
  dest##.hi = 0x##a##b; dest##.lo = 0x##c##d;
#define _V3_EQ128( value, a, b, c, d ) \
  ( ( value.hi == 0x##a##b ) && ( value.lo == 0x##c##d ) )
#else
#define _V3_SET128( dest, a, b, c, d ) \
  dest##.hi.hi = 0x##a; dest##.hi.lo = 0x##b; \
  dest##.lo.hi = 0x##c; dest##.lo.lo = 0x##d;
#define _V3_EQ128( value, a, b, c, d ) \
  ( ( value.hi.hi == 0x##a ) && ( value.hi.lo == 0x##b ) \
  && ( value.lo.hi == 0x##c ) && ( value.lo.lo == 0x##d ) )
#endif

/* Floating point, not everything has a IEEE floating point unit */
#if ( !defined( NO_IEEE_FLOAT ) )
typedef float f32;
typedef double f64;
#endif

/* Character types - ascii and UTF-16BE */
typedef u8 ascii;
typedef u16 unicode;

/* Time type */
typedef s128 v3_time;

/* Make sure that NULL is defined */
#if ( !defined( NULL ) )
#define NULL ((void *) 0)
#endif

#define V3_PASS  0
#define V3_FAIL -1

/* We don't want or need fine resolution on either OS or CPU for compiles */
#define CPU_UNKNOWN     0
#define CPU_X86         1
#define CPU_POWERPC     2
#define CPU_MIPS        3
#define CPU_MIPS64      4
#define CPU_ALPHA       5
#define CPU_PA_RISC     6
#define CPU_68K         7
#define CPU_SPARC       8
#define CPU_SPARC64     9
#define CPU_POWER       10
#define CPU_VAX         11
#define CPU_ARM         12
#define CPU_88K         13
#define CPU_S390        14
#define CPU_SH4         15
#define V3_CPU_TYPE_MAX 15

#define V3_CPU_TYPES { (ascii *) "UNKNOWN", (ascii *) "x86", \
  (ascii *) "PowerPC", (ascii *) "MIPS", (ascii *) "MIPS64", \
  (ascii *) "Alpha", (ascii *) "PA_RISC", (ascii *) "68K", \
  (ascii *) "Sparc", (ascii *) "Sparc64", (ascii *) "POWER", \
  (ascii *) "VAX", (ascii *) "ARM", (ascii *) "88K", \
  (ascii *) "S390", (ascii *) "SH4", (ascii *) NULL }

#define OS_UNKNOWN      0
#define OS_WIN32        1   /* Win95 and up */
#define OS_MACOS        2
#define OS_MACOSX       3
#define OS_LINUX        4
#define OS_BSDI         5
#define OS_NETBSD       6
#define OS_FREEBSD      7
#define OS_OPENBSD      8
#define OS_NEXTSTEP     9
#define OS_BEOS         10
#define OS_IRIX         11
#define OS_IRIX64       12
#define OS_SUNOS        13
#define OS_SOLARIS      14
#define OS_SCO          15
#define OS_QNX          16
#define OS_TRU64        17 /* AKA Dec UNIX, AKA OSF1 */
#define OS_VMS          18
#define OS_OS2          19
#define OS_UNIXWARE     20
#define OS_HPUX         21
#define OS_MACH         22
#define OS_AIX          23
#define OS_AUX          24
#define OS_AMIGAOS      25
#define OS_NETWARE      26
#define OS_MVS          27
#define OS_ULTRIX       28
#define OS_DGUX         29
#define OS_SINIX        30
#define OS_DYNIX        31
#define OS_OS390        32
#define OS_RISCOS       33
#define OS_OS9          34
#define OS_NUTC         35 /* an alternatice Win32 environment */
#define V3_OS_TYPE_MAX  35

#define V3_OS_TYPES { (ascii *) "UNKNOWN", (ascii *) "Win32", \
  (ascii *) "MacOS", (ascii *) "MacOS X", (ascii *) "Linux", \
  (ascii *) "BSDI", (ascii *) "NetBSD", (ascii *) "FreeBSD", \
  (ascii *) "OpenBSD", (ascii *) "NEXTSTEP", (ascii *) "BeOS", \
  (ascii *) "Irix", (ascii *) "Irix64", (ascii *) "SunOS", \
  (ascii *) "Solaris", (ascii *) "SCO", (ascii *) "QNX", \
  (ascii *) "Tru64", (ascii *) "VMS", (ascii *) "OS2", \
  (ascii *) "Unixware", (ascii *) "HPUX", (ascii *) "Mach", \
  (ascii *) "AIX", (ascii *) "AUX", (ascii *) "AmigaOS", \
  (ascii *) "Netware", (ascii *) "MVS", (ascii *) "ULTRIX", \
  (ascii *) "DGUX", (ascii *) "SINIX", (ascii *) "DYNIX", \
  (ascii *) "OS390", (ascii *) "RISCOS", (ascii *) "OS9", \
  (ascii *) "NuTCRACKER", (ascii *) NULL }

/* CPU_TYPE must be defined by now */
#if ( !defined( CPU_TYPE ) || ( CPU_TYPE == CPU_UNKNOWN ) \
  || ( CPU_TYPE > V3_CPU_TYPE_MAX ) )
#error "Define CPU_TYPE - see cputypes.h"
#endif

/* OS_TYPE must be defined by now */
#if ( !defined( OS_TYPE ) || ( OS_TYPE == OS_UNKNOWN ) \
  || ( OS_TYPE > V3_OS_TYPE_MAX ) )
#error "Define OS_TYPE - see cputypes.h"
#endif

#endif
