/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

#include "email.h"
#include "cosmio.h"
#include "cosmmem.h"

s32 v3EmailSMTP( const u32 smtp_server, const ascii * to, const ascii * from,
  const ascii * subject, const ascii * message, u32 length )
{
  v3_NET net;
  ascii reply[4], buffer[64];
  u32 my_ip;
  u64 buffer_length;
  u32 last_idx, idx;

  if ( to == NULL )
  {
    return( V3_EMAIL_ERROR_TO );
  }
  if ( from == NULL )
  {
    return( V3_EMAIL_ERROR_FROM );
  }
  if ( ( message == NULL ) && ( length != 0 ) )
  {
    return( V3_EMAIL_ERROR_MESSAGE );
  }

  if ( v3MemSet( &net, v3u64u32( sizeof( v3_NET ) ), 0 ) != V3_PASS )
  {
    return( V3_EMAIL_ERROR_ABORTED );
  }

  if ( v3NetOpen( &net, smtp_server, 25, 0, 25, V3_NET_MODE_TCP, 0, 0,
    (ascii *) NULL ) != V3_PASS )
  {
    return( V3_EMAIL_ERROR_HOST );
  }

  if ( v3_EmailGetReply( reply, &net ) != V3_PASS )
  {
    v3NetClose( &net );
    return( V3_EMAIL_ERROR_ABORTED );
  }
  if ( reply[0] != '2' )
  {
    v3NetClose( &net );
    return( V3_EMAIL_ERROR_ABORTED );
  }

  if ( v3NetMyIP( &my_ip ) == V3_PASS )
  {
    buffer_length = v3PrintAStr( buffer, v3u64u32( 64 ),
      (ascii *) "HELO %u.%u.%u.%u\r\n",
      ( my_ip >> 24 ) & 0xFF, ( my_ip >> 16 ) & 0xFF,
      ( my_ip >> 8 ) & 0xFF, my_ip & 0xFF );
  }
  else
  {
    /* Can't get my IP, just write something */
    buffer_length = v3PrintAStr( buffer, v3u64u32( 64 ),
      (ascii *) "HELO Unknown\r\n" );
  }

  if ( v3NetSend( &net, buffer, v3u32u64( buffer_length ) ) != V3_PASS )
  {
    /* Can't send HELO */
    v3NetClose( &net );
    return( V3_EMAIL_ERROR_ABORTED );
  }

  if ( v3_EmailGetReply( reply, &net ) != V3_PASS )
  {
    v3NetClose( &net );
    return( V3_EMAIL_ERROR_ABORTED );
  }
  if ( reply[0] != '2' )
  {
    v3NetClose( &net );
    return( V3_EMAIL_ERROR_ABORTED );
  }

  if ( ( v3NetSend( &net, "MAIL FROM:", 10 ) != V3_PASS )
    || ( v3NetSend( &net, from, v3u32u64( v3StrLengthA( from ) ) )
    != V3_PASS ) || ( v3NetSend( &net, "\r\n", 2 ) != V3_PASS ) )
  {
    /* Can't send FROM: command */
    v3NetClose( &net );
    return( V3_EMAIL_ERROR_ABORTED );
  }

  if ( v3_EmailGetReply( reply, &net ) != V3_PASS )
  {
    v3NetClose( &net );
    return( V3_EMAIL_ERROR_ABORTED );
  }

  if ( reply[0] != '2' )
  {
    if ( ( reply[0] == '5' ) && ( reply[1] == '0' ) && ( reply[2] == '1' ) )
    {
      v3NetClose( &net );
      return( V3_EMAIL_ERROR_FROM );
    }
    v3NetClose( &net );
    return( V3_EMAIL_ERROR_ABORTED );
  }

  if ( ( v3NetSend( &net, "RCPT TO:", 8 ) != V3_PASS )
    || ( v3NetSend( &net, to, v3u32u64( v3StrLengthA( to ) ) ) != V3_PASS )
    || ( v3NetSend( &net, "\r\n", 2 ) != V3_PASS ) )
  {
    /* Can't send TO: command */
    v3NetClose( &net );
    return( V3_EMAIL_ERROR_ABORTED );
  }

  if ( v3_EmailGetReply( reply, &net ) != V3_PASS )
  {
    v3NetClose( &net );
    return( V3_EMAIL_ERROR_ABORTED );
  }

  if ( reply[0] != '2' )
  {
    if ( ( reply[0] == '5' ) && ( reply[1] == '5' ) && ( reply[2] == '1' ) )
    {
      v3NetClose( &net );
      return( V3_EMAIL_ERROR_NORELAY );
    }
    if ( ( reply[0] == '5' ) && ( reply[1] == '5' ) && ( reply[2] == '0' ) )
    {
      v3NetClose( &net );
      return( V3_EMAIL_ERROR_TO );
    }
    v3NetClose( &net );
    return( V3_EMAIL_ERROR_ABORTED );
  }

  if ( v3NetSend( &net, "DATA\r\n", 6 ) != V3_PASS )
  {
    /* Can't send DATA command */
    v3NetClose( &net );
    return( V3_EMAIL_ERROR_ABORTED );
  }

  if ( v3_EmailGetReply( reply, &net ) != V3_PASS )
  {
    v3NetClose( &net );
    return( V3_EMAIL_ERROR_ABORTED );
  }

  if ( reply[0] != '3' )
  {
    v3NetClose( &net );
    return( V3_EMAIL_ERROR_ABORTED );
  }

  /* Sending headers */
  if ( ( v3NetSend( &net, "From: ", 6 ) != V3_PASS ) ||
    ( v3NetSend( &net, from,
    v3u32u64( v3StrLengthA( from ) ) ) != V3_PASS ) ||
    ( v3NetSend( &net, "\r\n", 3 ) != V3_PASS ) )
  {
    /* Can't send From: line */
    v3NetClose( &net );
    return( V3_EMAIL_ERROR_ABORTED );
  }

  if ( ( v3NetSend( &net, "To: ", 4 ) != V3_PASS )
    || ( v3NetSend( &net, to, v3u32u64( v3StrLengthA( to ) ) ) != V3_PASS )
    || ( v3NetSend( &net, "\r\n", 3 ) != V3_PASS ) )
  {
    /* Can't send To: line */
    v3NetClose( &net );
    return( V3_EMAIL_ERROR_ABORTED );
  }

  /* send subject only if there is one */
  if ( subject != NULL )
  {
    if ( ( v3NetSend( &net, "Subject: ", 9 ) != V3_PASS )
      || ( v3NetSend( &net, subject, v3u32u64( v3StrLengthA( subject ) ) )
      != V3_PASS ) || ( v3NetSend( &net, "\r\n", 3 ) != V3_PASS ) )
    {
      /* Can't send Subject: line */
      v3NetClose( &net );
      return( V3_EMAIL_ERROR_ABORTED );
    }
  }

  /* Sending message */
  last_idx = idx = 0;
  while ( idx < length )
  {
    if ( ( idx == last_idx ) && ( message[idx] == '.' ) )
    {
      /* Must double a dot at the beginning of a line */
      if ( v3NetSend( &net, ".", 1 ) != V3_PASS )
      {
        /* Can't send the extra dot */
        v3NetClose( &net );
        return( V3_EMAIL_ERROR_ABORTED );
      }
    }
    if ( message[idx] == '\r' )
    {
      if ( v3NetSend( &net, &message[last_idx], ( idx - last_idx + 1 ) )
        != V3_PASS )
      {
        /* Can't send that line of text */
        v3NetClose( &net );
        return( V3_EMAIL_ERROR_ABORTED );
      }
      if ( v3NetSend( &net, "\n", 1 ) != V3_PASS )
      {
        /* Can't send the carriage return */
        v3NetClose( &net );
        return( V3_EMAIL_ERROR_ABORTED );
      }
      if ( ( idx < length - 1 ) && ( message[idx+1] == '\n' ) )
      {
        idx++;
      }
      last_idx = idx + 1;
    }
    else if ( message[idx] == '\n' )
    {
      if ( v3NetSend( &net, &message[last_idx], ( idx - last_idx ) )
        != V3_PASS )
      {
        /* Can't send that line of text */
        v3NetClose( &net );
        return( V3_EMAIL_ERROR_ABORTED );
      }
      if ( v3NetSend( &net, "\r\n", 2 ) != V3_PASS )
      {
        /* Can't send the end of line */
        v3NetClose( &net );
        return( V3_EMAIL_ERROR_ABORTED );
      }
      if ( ( idx < length - 1 ) && ( message[idx+1] == '\r' ) )
      {
        idx++;
      }
      last_idx = idx + 1;
    }
    idx++;
  }
  if ( last_idx < idx )
  {
    /* Send remaining text */
    if ( v3NetSend( &net, &message[last_idx], ( idx - last_idx ) )
      != V3_PASS )
    {
      /* Can't send that line of text */
      v3NetClose( &net );
      return( V3_EMAIL_ERROR_ABORTED );
    }
  }

  /* Close the DATA channel */
  if ( v3NetSend( &net, "\r\n.\r\n", 5 ) != V3_PASS )
  {
    /* Can't send the end of DATA command */
    v3NetClose( &net );
    return( V3_EMAIL_ERROR_ABORTED );
  }

  if ( v3_EmailGetReply( reply, &net ) != V3_PASS )
  {
    v3NetClose( &net );
    return( V3_EMAIL_ERROR_ABORTED );
  }
  if ( reply[0] != '2' )
  {
    v3NetClose( &net );
    return( V3_EMAIL_ERROR_ABORTED );
  }

  v3NetSend( &net, "QUIT\r\n", 6 );
  v3NetClose( &net );
  return( V3_PASS );
}

/* Low level functions */

s32 v3_EmailGetReply( ascii * reply, v3_NET * net )
{
  ascii buffer[64];
  u32 size, timeout, wait = 120000;
  u32 i, j, part, lastline = 0;

  if ( ( reply == NULL ) || ( net == NULL ) )
  {
    return( V3_FAIL );
  }

  if ( v3NetRecv( reply, 4, net, wait ) != 4 )
  {
    /* Reply code not received */
    return( V3_FAIL );
  }
  if ( reply[3] == ' ' )
  {
    /* Last line to receive */
    lastline = 1;
  }
  reply[3] = 0;

  timeout = 0;
  part = 0;
  for ( ; ; )
  {
    if ( ( size = v3NetRecv( &buffer[part], 64 - part, net, timeout ) ) == 0 )
    {
      /* Nothing else was received!! */
      return( V3_FAIL );
    }
    part = 0;
    if ( lastline == 0 )
    {
      for ( i = 0 ; i < ( size - 1 ) ; i++ )
      {
        if ( buffer[i] == '\r' )
        {
          if ( i > ( size - 6 ) )
          {
            part = size - i;
            for ( j = 0 ; i < size ; i++ )
            {
              buffer[j] = buffer[i];
              j++;
            }
            break;
          }
          if ( buffer[i+1] == '\n' )
          {
            i += 5;
            if ( buffer[i] == ' ' )
            {
              /* Only one line left to receive */
              lastline = 1;
              break;
            }
          }
        }
      }
    }
    if ( lastline == 1 )
    {
      if ( buffer[size-2] == '\r' && buffer[size-1] == '\n' )
      {
        /* Reply text all received */
        return( V3_PASS );
      }
      if ( buffer[size-1] == '\r' )
      {
        buffer[0] = '\r';
        part = 1;
      }
    }

    if ( size < 64 )
    {
      /* Leave time for more text to arrive */
      timeout = 30000;
    }
    else
    {
      timeout = 0;
    }
  }
}

/* testing */

s32 v3_TestEmail( void )
{
  /* No automatic test can be done for now */
  return( V3_PASS );
}
