/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

#include "log.h"
#include "cosmio.h"
#include "cosmmem.h"

s32 v3LogOpen( v3_LOG * log, ascii * filename, u32 max_level, u32 mode )
{
  s32 result;

  if ( log == NULL )
  {
    return( V3_LOG_ERROR_INIT );
  }

  v3MutexLock( &log->lock, V3_MUTEX_WAIT );

  log->status = V3_LOG_STATUS_NULL;

  if ( filename == NULL )
  {
    v3MutexUnlock( &log->lock );
    return( V3_LOG_ERROR_NAME );
  }

  if ( ( mode != V3_LOG_MODE_NUMBER ) && ( mode != V3_LOG_MODE_BITS ) )
  {
    v3MutexUnlock( &log->lock );
    return( V3_LOG_ERROR_MODE );
  }

  result = v3MemSet( (void *) &log->file, v3u64u32( sizeof( v3_FILE ) ), 0 );
  if ( result != V3_PASS )
  {
    /* Something bad has happened. */
    v3MutexUnlock( &log->lock );
    return( V3_LOG_ERROR_ACCESS );
  }

  result = v3FileOpen( &log->file, filename, V3_FILE_MODE_CREATE |
    V3_FILE_MODE_APPEND, V3_FILE_LOCK_NONE );

  if ( result == V3_FILE_ERROR_NAME )
  {
    /* Invalid filename */
    v3MutexUnlock( &log->lock );
    return( V3_LOG_ERROR_NAME );
  }

  if ( result != V3_PASS )
  {
    /* Problem creating or accessing the logfile. */
    v3MutexUnlock( &log->lock );
    return( V3_LOG_ERROR_ACCESS );
  }

  /* We have an open file at this point.  Close it. */
  v3FileClose( &log->file );

  /* Everything checks out.  Fill in the v3_LOG structure and return. */
  v3StrCopyA( log->filename, filename, v3u64u32( V3_FILE_MAX_FILENAME ) );
  log->mode = mode;
  log->level = max_level;
  log->status = V3_LOG_STATUS_INIT;

  v3MutexUnlock( &log->lock );

  return( V3_PASS );
}

s32 v3LogClose( v3_LOG * log )
{
  v3MutexFree( &log->lock );

  return( V3_PASS );
}

s32 v3Log( v3_LOG * log, u32 level, u32 echo, const ascii * format, ... )
{
  va_list ap;
  s32 result;
  u64 tmp;

  if ( ( log == NULL ) || ( format == NULL ) )
  {
    return( V3_LOG_ERROR_INIT );
  }

  v3MutexLock( &log->lock, V3_MUTEX_WAIT );

  /* Sanity checks */
  if ( ( log->filename == NULL ) || ( log->status != V3_LOG_STATUS_INIT ) ||
    ( ( log->mode != V3_LOG_MODE_NUMBER ) &&
    ( log->mode != V3_LOG_MODE_BITS ) ) )
  {
    v3MutexUnlock( &log->lock );
    return( V3_LOG_ERROR_INIT );
  }

  /* First, check if we should write to the log */
  if ( level != 0 )
  {
    if ( log->mode == V3_LOG_MODE_NUMBER )
    {
      /* We're using 'number' */
      if ( level > log->level )
      {
        /* Then we don't log */
        v3MutexUnlock( &log->lock );
        return( V3_PASS );
      }
    }
    else
    {
      /* We're using 'bits' */
      if ( ( level & log->level ) == 0 )
      {
        /* Then we don't log */
        v3MutexUnlock( &log->lock );
        return( V3_PASS );
      }
    }
  }

  /* Looks like we're writing a log entry. */

  result = v3MemSet( (void *) &log->file, v3u64u32( sizeof(v3_FILE) ), 0 );
  if ( result != V3_PASS )
  {
    /* Something bad has happened */
    v3MutexUnlock( &log->lock );
    return( V3_LOG_ERROR_ACCESS );
  }

  result = v3FileOpen( &log->file, log->filename, V3_FILE_MODE_CREATE |
    V3_FILE_MODE_APPEND, V3_FILE_LOCK_WRITE );
  if ( result != V3_PASS )
  {
    /* Problem creating or accessing the logfile. */
    v3MutexUnlock( &log->lock );
    return( V3_LOG_ERROR_ACCESS );
  }

  /* Write to file */
  va_start( ap, format );
  _V3_SET64( tmp, FFFFFFFF, FFFFFFFF );
  v3_Print( &log->file, NULL, 1, tmp, format, ap );
  va_end( ap );
  if ( echo == V3_LOG_ECHO )
  {
    va_start( ap, format );
    v3_Print( NULL, NULL, 1, tmp, format, ap );
    va_end( ap );
  }

  if ( log->file.last_error != V3_FILE_ERROR_NONE )
  {
    /* Some file error occured. */
    v3MutexUnlock( &log->lock );
    return( V3_LOG_ERROR_ACCESS );
  }

  /* Ok. That's it. Close up and go home. */
  v3FileClose( &log->file );
  v3MutexUnlock( &log->lock );

  return( V3_PASS );
}

s32 v3_TestLog( void )
{
  v3_LOG log;
  v3_FILE file;
  ascii *correct, buf[512];
  u64 real_read;
  s32 ret;

  /* Tests 1-4 - Clear stuff */

  if ( v3MemSet( &log, v3u64u32( sizeof( v3_LOG ) ), 0x00 ) != V3_PASS )
  {
    return( -1 );
  }

  if ( v3MemSet( &file, v3u64u32( sizeof( v3_FILE ) ), 0x00 ) != V3_PASS )
  {
    return( -2 );
  }

  /* clean up from previous test */
  ret = v3FileDelete( (ascii *) "test.log" );
  if ( ( ret != V3_PASS ) && ( ret != V3_FILE_ERROR_NOTFOUND ) )
  {
    return( -4 );
  }

  /*
    5. Test v3LogOpen NULL log
    6. Test v3LogOpen NULL filename
    7. Test v3LogOpen invalid mode
    8. Test v3Log and NULL log
    9. Test v3Log and NULL format
  */

  if ( v3LogOpen( (v3_LOG *) NULL, (ascii *) "test.log", 0,
    V3_LOG_MODE_NUMBER ) != V3_LOG_ERROR_INIT )
  {
    return( -5 );
  }

  if ( v3LogOpen( &log, (ascii *) NULL, 0, V3_LOG_MODE_NUMBER )
    != V3_LOG_ERROR_NAME )
  {
    return( -6 );
  }

  if ( v3LogOpen( &log, (ascii *) "test.log", 0,
   ( V3_LOG_MODE_NUMBER | V3_LOG_MODE_BITS ) ) != V3_LOG_ERROR_MODE )
  {
    return( -7 );
  }

  if ( v3Log( (v3_LOG *) NULL, 0, V3_LOG_NOECHO, (ascii *) "foo" )
    != V3_LOG_ERROR_INIT )
  {
    return( -8 );
  }

  if ( v3Log( &log, 0, V3_LOG_NOECHO, (ascii *) NULL ) != V3_LOG_ERROR_INIT )
  {
    return( -9 );
  }

  /*
    Create a V3_LOG_MODE_NUMBER log and
    use it for the first set of tests
    10. Initialise log
    11. Test log level 10
    12. Test log level 6
    13. Test log level 5
    14. Test log level 4 and %u
    15. Test log level 1 and %s
    16. Test log level 0
    17. Open logfile for reading
    18. Read logfile's contents
    19. Compare contents with expected
    20. Delete logfile
  */

  if ( v3LogOpen( &log, (ascii *) "test.log", 5, V3_LOG_MODE_NUMBER )
    != V3_PASS )
  {
    return( -10 );
  }

  if ( v3Log( &log, 10, V3_LOG_NOECHO,
    (ascii *) "*** This shouldn't be logged*** \n" ) != V3_PASS )
  {
    return( -11 );
  }

  if ( v3Log( &log, 6, V3_LOG_NOECHO,
    (ascii *) "*** This shouldn't be logged: %u ***\n", (u32) 123 )
    != V3_PASS )
  {
    return( -12 );
  }

  if ( v3Log( &log, 5, V3_LOG_NOECHO,
    (ascii *) "This should be logged\n" ) != V3_PASS )
  {
    return( -13 );
  }

  if ( v3Log( &log, 4, V3_LOG_NOECHO, (ascii *) "So should this: %u\n",
    (u32) 123 ) != V3_PASS )
  {
    return( -14 );
  }

  if ( v3Log( &log, 1, V3_LOG_NOECHO, (ascii *) "%.6s <- That too\n",
    (ascii *) "foobar" ) != V3_PASS )
  {
    return( -15 );
  }

  if ( v3Log( &log, 0, V3_LOG_NOECHO, (ascii *) "Yup, this too\n" )
    != V3_PASS )
  {
    return( -16 );
  }

  v3LogClose( &log );

  /* Check log */

  correct = (ascii *) "This should be logged\n\
So should this: 123\nfoobar <- That too\nYup, this too\n";

  if ( v3FileOpen( &file, (ascii *) "test.log", V3_FILE_MODE_READ,
    V3_FILE_LOCK_READ ) != V3_PASS )
  {
    return( -17 );
  }

  real_read = v3FileRead( buf, v3StrLengthA( correct ), &file );
  v3FileClose( &file );
  if ( !v3u64Eq( real_read, v3StrLengthA( correct ) ) )
  {
    return( -18 );
  }

  if ( v3StrCmpA( buf, correct, v3StrLengthA( correct ) ) != V3_PASS )
  {
    return( -19 );
  }

  if ( v3FileDelete( (ascii *) "test.log" ) != V3_PASS )
  {
    return( -20 );
  }

  /*
    Create a V3_LOG_MODE_BITS log and
    use it for the first set of tests
    21. Clear log struct
    22. Clear file struct
    23. Initialise log
    24. Test log level 0x00
    25. Test log level 0x01 & %X
    26. Test log level 0x04
    27. Test log level 0x10
    28. Test log level 0x40
    29. Test log level 0x22
    30. Test log level 0x82
    31. Open logfile for reading
    32. Read logfile's contents
    33. Compare contents with expected
    34. Delete logfile
  */

  if ( v3MemSet( &log, v3u64u32( sizeof( v3_LOG ) ), 0x00 ) != V3_PASS )
  {
    return( -21 );
  }

  if ( v3MemSet( &file, v3u64u32( sizeof( v3_FILE ) ), 0x00 ) != V3_PASS )
  {
    return( -22 );
  }

  if ( v3LogOpen( &log, (ascii *) "test.log", 0x55, V3_LOG_MODE_BITS )
    != V3_PASS )
  {
    return( -23 );
  }

  if ( v3Log( &log, 0x00, V3_LOG_NOECHO, (ascii *) "This should be logged\n" )
    != V3_PASS )
  {
    return( -24 );
  }

  if ( v3Log( &log, 0x01, V3_LOG_NOECHO,
    (ascii *) "This should also be logged: %X\n", (u32) 54321 ) != V3_PASS )
  {
    return( -25 );
  }

  if ( v3Log( &log, 0x04, V3_LOG_NOECHO, (ascii *) "This should be logged\n" )
    != V3_PASS )
  {
    return( -26 );
  }

  if ( v3Log( &log, 0x10, V3_LOG_NOECHO,
    (ascii *) "This should be logged too.\n" ) != V3_PASS )
  {
    return( -27 );
  }

  if ( v3Log( &log, 0x40, V3_LOG_NOECHO,
    (ascii *) "This should be the last line.\n" ) != V3_PASS )
  {
    return( -28 );
  }

  if ( v3Log( &log, 0x22, V3_LOG_NOECHO,
    (ascii *) "*** This shouldn't be logged ***\n" ) != V3_PASS )
  {
    return( -29 );
  }

  if ( v3Log( &log, 0x82, V3_LOG_NOECHO,
    (ascii *) "*** This shouldn't be logged ***\n" ) != V3_PASS )
  {
    return( -30 );
  }

  v3LogClose( &log );
  
  /* Check log */

  correct = (ascii *) "This should be logged\n\
This should also be logged: D431\nThis should be logged\n\
This should be logged too.\nThis should be the last line.\n";

  if ( v3FileOpen( &file, (ascii *) "test.log", V3_FILE_MODE_READ,
    V3_FILE_LOCK_READ ) != V3_PASS )
  {
    return( -31 );
  }

  real_read = v3FileRead( buf, v3StrLengthA( correct ), &file );
  v3FileClose( &file );
  if ( !v3u64Eq( real_read, v3StrLengthA( correct ) ) )
  {
    return( -32 );
  }

  if ( v3StrCmpA( buf, correct, v3StrLengthA( correct ) ) != V3_PASS )
  {
    return( -33 );
  }

  if ( v3FileDelete( (ascii *) "test.log" ) != V3_PASS )
  {
    return( -34 );
  }

  return( V3_PASS );
}
